#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include <gazebo.h>
#include <gazebo/World.hh>
#include <gazebo/WorldFile.hh>
#include <gazebo/ModelFactory.hh>
#include <gazebo/Body.hh>
#include <gazebo/BoxGeom.hh>
#include <gazebo/CylinderGeom.hh>

#include <fstream>
#include <map>
using namespace std;

#include "floorplan.h"

/////////////////////////////////////////////////////////////////////////////
// Register the model
GZ_REGISTER_PLUGIN("Floorplan", Floorplan)

//////////////////////////////////////////////////////////////////////////////
// Constructor
Floorplan::Floorplan( World *world ) : Model( world )
{
  this->body = NULL;
        
  return;
}

//////////////////////////////////////////////////////////////////////////////
// Destructor
Floorplan::~Floorplan()
{
  if (this->body)
    delete this->body;
  
  this->body = NULL;

  return;
}

//////////////////////////////////////////////////////////////////////////////
// Load the sensor
int Floorplan::Load( WorldFile *file, WorldFileNode *node )
{
  this->updatePeriod = 1.0 / (node->GetDouble("updateRate", 10) + 1e-6);
  this->updateTime = -updatePeriod;

  this->wall_color = node->GetColor("wall_color", GzColor(0.8,0.1,0.1));
	this->wall_thickness = node->GetDouble("wall_thickness", 0.05);
	this->wall_height = node->GetDouble("wall_height", 1.0);

	this->vert_input_filename = node->GetString("vertices", NULL);
	this->poly_input_filename = node->GetString("poly", NULL);
    this->texture_file = node->GetString("texture_file", NULL); //node->SearchFilename("texture_file", ".", NULL);
    fprintf(stderr, "Texture file: %s\n", this->texture_file);

	if(vert_input_filename)
		this->ReadVertInput();

	if(poly_input_filename) 
		this->ReadPolyInput();

  // Create the ODE objects
  if (this->OdeLoad(file, node) != 0)
    return -1;

  return 0;
}

int Floorplan::ReadVertInput()
{

	if(this->vert_input_filename == NULL)
		return -1;
	
	ifstream fin;

	try {
		fin.open(this->vert_input_filename);
	}
	catch (const std::exception& e)
		{
			return -1;
		} // catch
	
	double x1, y1, x2, y2, x0, y0;

	fin >> x1 >> y1;

	x0 = x1;
	y0 = y1;

	do {
		fin >> x2 >> y2;
		line_segments.push_back(LineSegment(x1,y1,x2,y2));
		x1 = x2;
		y1 = y2;
	} while (!fin.eof());

	line_segments.push_back(LineSegment(x1, y1, x0, y0));

	fin.close();

	return 0;
}

int Floorplan::ReadPolyInput()
{

	if(this->poly_input_filename == NULL)
		return -1;
	
	ifstream fin;

	try {
		fin.open(this->poly_input_filename);
	}
	catch (const std::exception& e)
		{
			return -1;
		} // catch
	
	int num_pts;
	int num_segments;
	double x0, y0;
	int id, start, end;
	char dummy[255];

	std::map<int, pair<double,double> > pts;

	fin >> num_pts;
	fin.getline(dummy, 255);

	printf("%d points\n", num_pts);

	for(int i=0; i < num_pts; ++i) {
		fin >> id >> x0 >> y0;
		fin.getline(dummy, 255);
		pts[id] = (pair<double,double>(x0, y0));
	}

	fin >> num_segments;
	fin.getline(dummy, 255);

	printf("%d segments\n", num_segments);

	for(int i=0 ; i < num_segments; ++i) {
		fin >> id >> start >> end;
		fin.getline(dummy, 255);
		line_segments.push_back(LineSegment(pts[start].first, pts[start].second, pts[end].first, pts[end].second));
	}

	fin.close();

	return 0;
}

#define MASS 1e6

//////////////////////////////////////////////////////////////////////////////
int Floorplan::OdeLoad(WorldFile *file, WorldFileNode *node)
{
  Geom *geom;

  this->body = new Body( this->world );

	list<LineSegment>::iterator iter;
	double dist, theta;
	
	pair<double,double> a, b;
	
	int i=0;
    int wall_count=0;

	uint wall_category = (uint)node->GetInt("wall_category", 4);

	for(iter = line_segments.begin(); iter != line_segments.end(); ++iter) {

		dist = iter->dist;
		theta = atan2(iter->y2-iter->y1, iter->x2-iter->x1);

		if(dist > 1e-4) {

			geom = new BoxGeom(this->body, this->modelSpaceId, dist, wall_thickness, wall_height);
			geom->SetRelativePosition(GzVectorSet(iter->x1 + 0.5*dist*cos(theta), iter->y1 + 0.5*dist*sin(theta), wall_height/2));
			geom->SetRelativeRotation(GzQuaternFromEuler(0, 0, theta));
			geom->SetColor(this->wall_color);
			geom->SetTransparency(true);
			geom->SetMass(MASS);
			geom->SetCategoryBits(wall_category);
			geom->SetCollideBits(~wall_category);

            // Apply a texture
            if (this->texture_file)
            {
              char wallTextureFile[10000];

              // If postfixed with an image file type, assume
              // that same texture used for all walls
              if (strstr(this->texture_file, "ppm"))
                sprintf(wallTextureFile, "%s", this->texture_file);
              else
                sprintf(wallTextureFile, "%s.%d.ppm", this->texture_file, wall_count);
              
              fprintf(stderr, "Applying texture %s\n", wallTextureFile);
              printf("loading texture file [%s]", wallTextureFile);  
              if (geom->SetTexture2DFile(wallTextureFile) != 0)
                printf("unable to load texture file");
              geom->SetTexture2DSize(node->GetPosition("textureSize", GzVectorSet(2.0, 2.0, 0.0)));
            } else {
              //fprintf(stderr, "Error: Loading texture file %s\n", this->texture_file);
            }
            wall_count++;
            
		}
	}

  this->AddBody( this->body, true );


  return 0;
}

//////////////////////////////////////////////////////////////////////////////
int Floorplan::Init( WorldFile *file, WorldFileNode *node )
{
  // Create position interface
  this->position_iface = gz_position_alloc();
  assert(this->position_iface);
  if (gz_position_create(this->position_iface, this->world->gz_server, this->GetId(),
                         "Floorplan", this->GetIntId(), this->GetParentIntId()) != 0)
    return -1;

  GzPose init_pose = this->body->GetPose();
  this->odomPose[0] = init_pose.pos.x;
  this->odomPose[1] = init_pose.pos.y;
  this->odomPose[2] = init_pose.pos.z;
  
  GzVector init_rot = GzQuaternToEuler(init_pose.rot);
  this->odomPose[3] = init_rot.x;
  this->odomPose[4] = init_rot.y;
  this->odomPose[5] = init_rot.z;
 
  for (int i=0;i<6;i++)
    this->odomVel[i] = 0.0;

  return 0;
}

//////////////////////////////////////////////////////////////////////////////
int Floorplan::Fini()
{
  gz_position_destroy( this->position_iface );
  gz_position_free( this->position_iface );
  this->position_iface = NULL;

  return 0;
}


//////////////////////////////////////////////////////////////////////////////
void Floorplan::Update(double step)
{
  // Do nothing if paused
  if (step == 0)
    return;

  this->UpdateOdometry(step);

  if (this->world->GetSimTime() - this->updateTime < this->updatePeriod)
    return;
  this->updateTime = this->world->GetSimTime();

  this->GetVelocityCmd();
  
  this->body->SetLinearVel(GzVectorSet(odomVel[0], odomVel[1], odomVel[2]));
  this->body->SetAngularVel(GzVectorSet(odomVel[3], odomVel[4], odomVel[5]));

  this->PutPositionData();

  return;
}

//////////////////////////////////////////////////////////////////////////////
void Floorplan::UpdateOdometry( double step )
{

	GzPose init_pose = this->body->GetPose();
  odomPose[0] = init_pose.pos.x;
  odomPose[1] = init_pose.pos.y;
  odomPose[2] = init_pose.pos.z;
  
  GzVector init_rot = GzQuaternToEuler(init_pose.rot);
  odomPose[3] = init_rot.x;
  odomPose[4] = init_rot.y;
  odomPose[5] = init_rot.z;	

	/*
  for (int i=0;i < 6; i++)
      odomPose[i] += step*odomVel[i];
	*/

	/*
	printf("Q: %4.4f %4.4f %4.4f %4.4f, E: %4.4f, %4.4f, %4.4f (%4.4f, %4.4f, %4.4f)\n", 
				 init_pose.rot.u, init_pose.rot.x, init_pose.rot.y, init_pose.rot.z,
				 currentPose[3], currentPose[4], currentPose[5], 
				 odomPose[3], odomPose[4], odomPose[5]);
	*/


  return;
}

//////////////////////////////////////////////////////////////////////////////
void Floorplan::GetVelocityCmd()
{
  for (int i=0;i < 3;i++)
    {
         odomVel[i] = this->position_iface->data->cmd_vel_pos[i];
         odomVel[3+i] = this->position_iface->data->cmd_vel_rot[i];
    }

  return;
}

//////////////////////////////////////////////////////////////////////////////
void Floorplan::PutPositionData()
{
  gz_position_lock(this->position_iface, 1);
    
  this->position_iface->data->time = this->world->GetSimTime();
  for (int i=0;i < 3;i++)
    {
      this->position_iface->data->pos[i] = this->odomPose[i];
      this->position_iface->data->rot[i] = this->odomPose[3+i];
      this->position_iface->data->vel_pos[i] = this->odomVel[i];
      this->position_iface->data->vel_rot[i] = this->odomVel[3+i];
    }

  gz_position_unlock(this->position_iface);
  
  return;
}
