import gnu.getopt.*;
import java.util.*;
import cricketdaemon.clientlib.*;       
import cricketdaemon.clientlib.data.*;  
import java.net.*;
import java.io.*;


// A Cricket callback handler class implements 
// the Callback interface
class CricketPositionServer implements Callback {
    
    // Handle to cricket
    Broker cricket;
    
    // Data structures
    String currentSpace = null;
    Position currentPosition = null;

    // Network variables
    static int CricketDPort;
    static int PositionServerListenPort;
    static int ListenerID;    
    static PrintWriter OutSock;
    static Socket ClientSock;

    public CricketPositionServer() {

	/**
	 * Established TCP connection to the CricketDaemon running on
	 * localhost.  The ServerBroker will trigger a callback
	 * whenever the CricketDaemon decides to push information out to
	 * the clients.  An application may the alternate 
	 * ServerBroker constructor:
	 *
	 * public ServerBroker(String addr, int portNum)
	 *
	 * to connect to a CricketDaemon running on <code>addr</code>
	 * (may be numeric IP or hostname), using portNum (normally,
	 * should use 5001).	 
	 */
	cricket = new ServerBroker("127.0.0.1", CricketDPort);

	/**
	 * Create a bitmask specifying the type of Cricket information
	 * that should trigger a callback on this object.  The following
	 * table shows all the valid bit types:
	 * ALL                  selects all types 
	 * SPACE                current spatial location 
	 * DEVICECOORD          current position coordinates 
	 * BEACONSHEARDLIST     set of beacons heard in history window 
	 * BEACONSTAT           distance measurement to a beacon 
	 * POSERR               exceptions in position computation
	 */
	BitSet mask = new BitSet();
	mask.set(Broker.SPACE);
	mask.set(Broker.DEVICECOORD);
	mask.set(Broker.POSERR);
	
	/**
	 * Register this object's callback handler with ServerBroker.
	 * An application can create and register multiple handlers.
	 * Here, we register only one handler.
	 */
	cricket.register(this, mask);
	cricket.start(); // fork thread, runs forever
	System.out.println("CricketPositionServer created");
    }

    /**
     * The ServerBroker object invokes the callback handler
     * with location updates contained in <code>data</code> and
     * <code>mask</code> indicates which location type()s is(are) 
     * readable in \texttt{CricketData}. All Cricket callback handler
     * classes must implement this method.  
     */
    synchronized public void callback(CricketData data, BitSet mask) {
	boolean update = false;
	
	// check for space updates
	if(mask.get(Broker.SPACE)) {
	    BeaconRecord cs = data.getCurrentSpaceObject();
	    if(cs != null && cs.uniqueName != null) {
		if(currentSpace == null || !(currentSpace.equals(cs.uniqueName))) {
		    // if space string is different from before, update it
		    update = true;
		    currentSpace = cs.uniqueName;
		}
	    }
	}

	// check for coordinate updates
	if(mask.get(Broker.DEVICECOORD) || mask.get(Broker.POSERR)) {
		//System.out.println("Coord update");
	    Position p = data.getDevicePosition();

	    if(p != null) {
	    //System.out.println("not null");
		if(currentPosition == null || !(currentPosition.equals(p))) {
		    if(p.invalid()) {
			// CricketDaemon cannot produce a position estimate
			// because it does not hear enough different beacons
		//	System.err.println("ignoring invalid position");
		    }
		    else {
	               // Check if correct algorithm is used (llsq)
	               if (data.getPosAlgor().equals("ekf-passive")) {
            	 	   System.out.println("Ignoring ekf-passive");
	       		   return;
	    		}

			// if the coordinates are different from before, 
			// update them
			update = true;
			currentPosition = p;
			OutSock.println("(" + p.x + "," + p.y + "," + p.z+")");
			System.out.println("Got a position: " + ListenerID + "," + p.x + "," + p.y + "," + p.z);
		    }
		}
	    }
	}
	
	if(update)  {
	    //printState();
        }
    }
    
    /**
     * Prints the current space and coorinates of the Cricket listener
     */
    private void printState() {
	System.out.println("space="+currentSpace+" pos="+currentPosition);
    }

    public static void main(String[] args)  {

	/* OPTIONAL: example of gnu-style getopt to parse cmd line args */
	Getopt g = new Getopt("CricketPositionServer", args, "c:p:i:");
	int c;
	try {
	    while((c = g.getopt()) != -1) {
		switch(c) {
		case 'c':
		    System.out.println("cricketd port: " + g.getOptarg());
		    CricketDPort = Integer.parseInt(g.getOptarg());
		    break;
		case 'p':
		    System.out.println("Position Server Port: " + g.getOptarg());
		    PositionServerListenPort = Integer.parseInt(g.getOptarg());
		    break;
		case 'i':
		    System.out.println("Listener ID: " + g.getOptarg());
		    ListenerID = Integer.parseInt(g.getOptarg());
		    break;
		}
	    
	    }
	}
	catch (Exception e) {
	    System.out.println("Usage: java CricketPositionServer -c cricketd_port -p position_server_port");
	}
	if (args.length != 6) {
            System.out.println("Usage: java CricketPositionServer -i listenerid -c cricketd_port -p position_server_port");
	   System.exit(-1);
	}

	// set up a socket so we can send data to the position server

        ServerSocket serverSocket = null;
        try {
            System.out.println("Waiting for connection to the java position server.");
            serverSocket = new ServerSocket(PositionServerListenPort);
        } catch (IOException e) {
            System.err.println("Could not listen on port" + PositionServerListenPort);
            System.exit(-1);
        }
	try {
           ClientSock = serverSocket.accept();
	} catch (IOException e) {
           System.err.println("Problem accepting PositionServerListen socket");
	   System.exit(-1);
        }
	System.out.println("Position Server Connected.");

	try {
        	OutSock = new PrintWriter(ClientSock.getOutputStream(), true);
	} catch (IOException e) {
           System.err.println("Problem making PrintWriter");
           System.exit(-1);
        }



	// start listening for beacons
	CricketPositionServer client = new CricketPositionServer();
    }
}

