/**********************************************************************
 **                                                                  **
 **                          cbrdoe.cc                               **
 **                                                                  **
 **  Written by:  Zsolt Kira                                         **
 **                                                                  **
 **  This file contains the main function for CBR-DOE integration.   **
 **                                                                  **
 **  The goal of this is, given a saved mission that has been run,   **
 **  to explore different parameterization of the mission via        **
 **  simulation, obtain quantitative metrics, and store the results  **
 **  in a  case library.                                             **
 **                                                                  **
 **  The decision tree & mission plan are obtained via an existing   **
 **  case library saved containing the FSA & mission (it assumes the **
 **  first mission is the one to be replicated).  It then replicates **
 **  this case, changing the input & output parameters to values     **
 **  obtained from two CSV files.                                    **
 **  The resulting case library is written to a new library file     **
 **                                                                  **
 **  Limitations/TODO:                                               **
 **  - There seem to be some problems in the way this code uses the  **
 **    CBR functions (or problems in CBR itself)                     **
 **    > The missions that are retrieved contain spurious data in    **
 **      the max_vel and base_vel entries in the CDL file.  This     **
 **      seems to occur even without the DOE stuff                   **
 **                                                                  **
 **  - Output metrics are assumed to be integers like other metrics  **
 **    that were there before (given by users such as StealthRating  ** 
 **                                                                  ** 
 **  - GUI only supports 0...1, 0....10, 0...100 etc. while some DOE **
 **    metrics can be negative.  Right now they are scaled below.    ** 
 **                                                                  ** 
 **  Copyright 2009 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/


#include <string.h>

#include "cbrdoe.h"
#include "memory_manager.h"

#define LINE_BUFFER_LENGTH 1000
#define STRIP_NEWLINE(line) (line)[strlen(line)-1] = '\0';
#define PARAMS_NUM_LINES_TO_SKIP 1

// Input parameters: These are defined in cbrplanner_protocol.h
const int numOutputFeatures = 4;
const int outputFeatures[numOutputFeatures] = {CBRPLANNER_EXTRADATA_KEY_MISSION_ACCOMPLISHED, CBRPLANNER_EXTRADATA_KEY_MISSION_TIME,CBRPLANNER_EXTRADATA_KEY_NUMBER_ROBOTS_FAILED,CBRPLANNER_EXTRADATA_KEY_TOTAL_DISTANCE_TRAVELLED};

// Output parameters:
// These come from the .mission_expertrc file in the directory from which cfgedit was started and the mission saved to the CBR library
// In the future they should probably be read or should use existing parsing code wherever that is
const int numInputFeatures = 11;
const char *inputFeatures[numInputFeatures] = {"NumberOfRobots", "CommReference", "CommCutoff", "MinObsClearance", "StealthFactor", "MaximumVelocity", "MaxAltVelocity", "MinTurnRadius", "MaxVisualRange", "MaxNavSensorRange", "MaxChemSensorRange"};

// These specify min/max of feature values, so that they can be scaled to 0...1 since
// that's what is supported in the cbr expert GUI
const double inputFeatureMin[numInputFeatures] = {1.0, -90.0, -90.0, 1.0,  5.0,  5.0,  1.0,  5.0,  5.0,   1.0,  1.0};
const double inputFeatureMax[numInputFeatures] = {5.0,   0.0, -30.0, 5.0, 75.0, 50.0, 20.0, 20.0, 50.0, 100.0, 50.0};
#define CBR_DOE_SCALE_PARAMETER(value, min, max) (value-min)/(max-min)

const char *inputCSV = "doe1.csv";
const char *outputCSV = "out1.csv";

void printFeatures(Features_t features) {
  Feature_t feature;
  Features_t::iterator fit;
  for (fit = features.begin(); fit != features.end(); fit++) {
    feature = *fit;
    gWindows->printfPlannerWindow("Feature: %s : %s  (nonIndex=%d, weight=%lf formatType=%d)\n", feature.name.c_str(), feature.value.c_str(), feature.nonIndex, feature.weight, feature.formatType);
  }
}

void printExtraData(vector<CBRPlanner_ExtraData_t> extraDataList) {
  CBRPlanner_ExtraData_t outputFeature;
  vector<CBRPlanner_ExtraData_t>::iterator ceit;
  for (ceit = extraDataList.begin(); ceit != extraDataList.end(); ceit++) {
    outputFeature = *ceit;
    gWindows->printfPlannerWindow("%s:%s ", outputFeature.key.c_str(), outputFeature.dataEntries[0].c_str());
  }
  gWindows->printfPlannerWindow("\n");
}

// CSV Filenames
bool loadDOE(CBRPlanner *planner) {
  FILE *finput = NULL;  // File containing input parameters
  FILE *foutput = NULL; // File containint output metrics 

  bool debug = false;

  // Setting this to true currently results in a crash!
  bool createNewMissionManager = true;

  DecisionTree_t *decisionTree = NULL;
  DecisionTreeNode_t *node = NULL;
  MissionMemoryManager *missionManager = NULL;
  MissionMemoryManager *newMissionManager = NULL;
  DecisionTreeData_t data;
  Features_t features;
  Features_t newFeatures;
  Feature_t feature;
  CBRPlanner_ExtraData_t outputFeature;
 
  gWindows->printfPlannerWindow("Loading DOE.\n");

  if (createNewMissionManager)
    new MissionMemoryManager(&newMissionManager);

  // Obtain mission manager
  missionManager = gCBRPlanner->missionMemoryManager();
  if (missionManager == NULL) {
    gWindows->printfPlannerWindow("Loading DOE.\n");
    return false;
  }

  // Get mission plan (contains FSA and output metrics that we are changing per DOE configuration
  // NOTE: We are assuming the one we want is the first case
  MissionPlan_t missionPlan = missionManager->getMissionPlan(0);
  if (missionPlan.type == -1) {
    gWindows->printfPlannerWindow("DOE Error: There must be a library loaded with at least one case in it.\n");
    return false;
  }

  // Obtain decision tree which contains the output metrics we are changing
  decisionTree = missionManager->getDecisionTree();
  if (decisionTree == NULL) {
    gWindows->printfPlannerWindow("DOE Error: The original decision tree is null.\n");
    return false;
  }

  // Obtain first node, which we are assuming is the one we want
  node = (DecisionTreeNode_t *) decisionTree->edges.back().nextNode;
  if (node == NULL) {
    gWindows->printfPlannerWindow("DOE Error: Original decision tree does not contain any nodes.\n");
    return false;
  }

  if (debug) {
    gWindows->printfPlannerWindow("Decision Tree has %d statuses, %d indexes, %d features\n", data.statuses.size(), data.indexes.size(), data.otherFeatures.size());
    
    //printDecisionTree(decisionTree);
  }

  // We now have everything necessary.  Go through the CSV files created by the DOE,
  // and for each one put appropriate input/output features in place, add it to the decision
  // tree
  finput = fopen(inputCSV, "r");
  if (finput == NULL) {
      gWindows->printfPlannerWindow("DOE Error: Cannot open input parameter file %s.\n", inputCSV);
      return false;
  }

  foutput = fopen(outputCSV, "r");
  if (foutput == NULL) {
      gWindows->printfPlannerWindow("DOE Error: Cannot open ouput metrics file %s.\n", inputCSV);
      return false;
  }

  int line_count = 1;
  char *cret = NULL, *token = NULL;
  char pline[LINE_BUFFER_LENGTH];
  char mline[LINE_BUFFER_LENGTH];
  while ((cret = fgets(pline, LINE_BUFFER_LENGTH-1, finput)) != NULL) {
    STRIP_NEWLINE(pline);

    // Skip first lines (contains parameter labels)
    if (line_count <= PARAMS_NUM_LINES_TO_SKIP) {
      line_count++;
      continue;
    }

    if (1||debug) gWindows->printfPlannerWindow("PARAMETER LINE: %s", pline);

    // Get corresponding output metrics from CSV file
    if ( (cret = fgets(mline, LINE_BUFFER_LENGTH-1, foutput)) == NULL) {
      gWindows->printfPlannerWindow("Warning: Have a line %d for input parameters but not output metrics!\n", line_count);
      break;
    }
    STRIP_NEWLINE(pline);

    if (1||debug) gWindows->printfPlannerWindow("Got lines:\n pline:%s\n mline:%s\n", pline, mline);

    // Change mission plan feature values (input parameters)
    int mcount = 0;
    double val;
    bool found;
    char charVal[10000];
    token = strtok(mline,",");
    vector<CBRPlanner_ExtraData_t> extraDataList = missionPlan.extraDataList;
    while (token != NULL) {
      val = atof(token);

      found = false;
      for (int i = 0; i < extraDataList.size(); i++) {
        if (strcmp(extraDataList[i].key.c_str(), CBRPLANNER_EXTRADATA_KEYS[outputFeatures[mcount]].c_str()) == 0) {
          //gWindows->printfPlannerWindow("Comparing %s to %s FOUND\n", CBRPLANNER_EXTRADATA_KEYS[outputFeatures[mcount]].c_str(),  outputFeature.key.c_str());
          //outputFeature.dataEntries.push_back("val");
 
          sprintf(charVal, "%d", (int) (val+0.5));
          extraDataList[i].dataEntries[0] = charVal;
          found = true;
          break;
        } else  {
          //gWindows->printfPlannerWindow("Comparing %s to %s\n", CBRPLANNER_EXTRADATA_KEYS[outputFeatures[mcount]].c_str(), outputFeature.key.c_str());
        }
      }
      if (!found) {
        //gWindows->printfPlannerWindow("DOE Warning: Could not find output metric %s in decision tree structure\n", CBRPLANNER_EXTRADATA_KEYS[outputFeatures[mcount]].c_str());
      }

      mcount++;
      if (mcount >= numOutputFeatures)
        break;
      token = strtok(NULL, ",");
    }
    missionPlan.extraDataList = extraDataList;  


    if (mcount < numOutputFeatures) {
      gWindows->printfPlannerWindow("DOE Error: Line %d does not contain %d output parameters as it should.\n", line_count, numOutputFeatures);
    }

    // Change decision tree node feature values (output metrics)
    int pcount = 0;
    token = strtok(pline,",");

    data = node->data;
    features = node->data.otherFeatures;

    if (debug) printFeatures(features);

    // Take care of features NOT in DOE but in the current decision tree (just copy wholesale)
    newFeatures.clear();
    Features_t::iterator fit;
    for (fit = features.begin(); fit != features.end(); fit++) {
        feature = *fit;

        found = false;
        for (int i = 0; i < numInputFeatures; i++) {
          if (strcmp(inputFeatures[i], feature.name.c_str()) == 0) {
             found = true;
          }
        }
        //gWindows->printfPlannerWindow("%s: Found is %d\n", feature.name.c_str(), found);
        if (found == false) {
          newFeatures.push_back(feature);
        }
    }

    // Now put in values from DOE for the rest of them
    while (token != NULL) {
      val = atof(token);
   
      found = false; 
      for (int i = 0; i < features.size(); i++) {
        if (strcmp(features[i].name.c_str(), inputFeatures[pcount]) == 0) {
          if (debug) gWindows->printfPlannerWindow("Comparing %s to %s FOUND\n", inputFeatures[pcount], features[i].name.c_str());

          // First input (number of robots) is int, the rest are double
          if (pcount == 0) 
            sprintf(charVal, "%d", (int) (val+0.5));
          else  {
            sprintf(charVal, "%f", CBR_DOE_SCALE_PARAMETER(val, inputFeatureMin[pcount], inputFeatureMax[pcount]));
            gWindows->printfPlannerWindow("%s: Scaled paramter: %f -> (%f,%f) -> %f\n", inputFeatures[pcount], val, inputFeatureMin[pcount], inputFeatureMax[pcount], CBR_DOE_SCALE_PARAMETER(val, inputFeatureMin[pcount], inputFeatureMax[pcount]));
          }
          
          features[i].value = charVal;
          found = true;
          newFeatures.push_back(feature);
          break;
        }
      }
      if (!found) {
        if (debug) gWindows->printfPlannerWindow("DOE Warning: Could not find input parameter %s in decision tree structure\n", inputFeatures[pcount]);
      }
  
      pcount++;
      if (pcount >= numInputFeatures)
        break;
      token = strtok(NULL, ",");
    }
    
    if (pcount < numInputFeatures) {
      gWindows->printfPlannerWindow("DOE Error: Line %d does not contain %d input parameters as it should.\n", line_count, numInputFeatures);
    }

    if (debug) {
      printFeatures(features);
      //printExtraData(missionPlan.extraDataList); 
    }

    if (createNewMissionManager) 
      newMissionManager->saveNewMissionPlan(missionPlan, features);
    else 
      missionManager->saveNewMissionPlan(missionPlan, features);

    line_count++;
  }

  fclose(finput);
  fclose(foutput);

  // Write out the final case library
  char libraryFilename[10000];
  strcpy(libraryFilename, gCBRPlanner->getCBRLibraryName().c_str());
  strcat(libraryFilename, "-DOE.cbl");

  if (createNewMissionManager) 
    newMissionManager->saveCBRLibrary(libraryFilename);
  else
    gCBRPlanner->saveCBRLibrary(libraryFilename);

  return true;
}

// This function just prints out elements of the decision tree for debug
void printDecisionTree(DecisionTree_t *decisionTree) {
    int i, numEdges;
    DecisionTreeNode_t *node;
    DecisionTreeEdge_t edge;
    vector<DecisionTreeEdge_t> edgeStack;

    gWindows->printfPlannerWindow("ROOT NODE:\n");
    gWindows->printfPlannerWindow("NAME %s\n", decisionTree->name.c_str());
    gWindows->printfPlannerWindow("LEAF %d\n", decisionTree->isLeafNode);
    gWindows->printfPlannerWindow("EDGES %d\n", decisionTree->edges.size());

    edgeStack = decisionTree->edges;

    while(!(edgeStack.empty()))
    {
        edge = edgeStack.back();
        edgeStack.pop_back();
        gWindows->printfPlannerWindow("EDGE PARENT %s\n", edge.parentName.c_str());
        gWindows->printfPlannerWindow("EDGE VALUE %s\n", edge.value.c_str());

        node = (DecisionTreeNode_t *)edge.nextNode;
        gWindows->printfPlannerWindow("NEW NODE:\n");
        gWindows->printfPlannerWindow("NAME %s\n", node->name.c_str());
        gWindows->printfPlannerWindow("LEAF %d\n", node->isLeafNode);
        gWindows->printfPlannerWindow("EDGES %d\n", node->edges.size());

        if (node->isLeafNode)
        {
            for (i = 0; i < ((int)(node->data.indexes.size())); i++)
            {
                gWindows->printfPlannerWindow("DATA %d\n", node->data.indexes[i]);
                gWindows->printfPlannerWindow("STATUS %d\n", node->data.statuses[i]);
            }
        }

        numEdges = node->edges.size();
        for (i = 0; i < numEdges; i++)
        {
            edgeStack.push_back(node->edges[i]);
        }
    }
}
