/**********************************************************************
 **                                                                  **
 **                          cbr_explore.cc                          **
 **                                                                  **
 **  Written by:  Zsolt Kira                                         **
 **                                                                  **
 **  This file contains the main function for CBRExplore.            **
 **                                                                  **
 **  The goal of this is, given a saved mission that has been run,   **
 **  to explore different parameterization of the mission via        **
 **  simulation, obtain quantitative metrics, and store the results  **
 **  in a  case library.                                             **
 **                                                                  **
 **  There are two possible modes of doing this:                     **
 **   - Reading the parameterization & resulting output metrics from **
 **     a file (i.e. someone else gives you this data), and          **
 **     outputting the resulting case library                        **
 **                                                                  **
 **  - Connecting to cbrserver, creating a new library, taking each  **
 **    parameter configuration, running the mission, gathering the   **
 **    resulting output metrics, and sending the result to the       **
 **    cbrserver so that it can create a new case.                   **
 **                                                                  **
 **  Currently the first mode is implemented in this file.           **
 **                                                                  **
 **  Copyright 2009 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: cbrexplore.cc,v 1.1 2009-05-10 21:20:12 zkira Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cbrplanner_protocol.h"

// String denoting the input parameters that show up in a case file
// TODO: This should probably be done some other way, so that changing it in the
// cbrplanner won't have to result in code changes here as well
#define INPUT_PARAMETER_CASE_STRING "NODE-DATA-OTHERFEATURE-VALUE"
#define OUTPUT_METRICS_CASE_STRING "META-DATA"
#define DECISION_TREE_CASE_STRING "DECISIONTREE"
#define DECISION_TREE_NODE_CASE_STRING "NODE"

#define LINE_BUFFER_LENGTH 1000
#define STRIP_NEWLINE(line) (line)[strlen(line)-1] = '\0';
#define PARAMS_NUM_LINES_TO_SKIP 1

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const bool DEBUG = false;

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------

FILE *open_file_or_exit(char *filename, char *mode, char *file_type) {
  FILE *out = NULL;

  out = fopen(filename, mode);
  if (out == NULL) {
    fprintf(stderr, "Error: Could not open %s file %s\n", file_type, filename);
    exit(EXIT_FAILURE);
  }

  return out;
}

//-----------------------------------------------------------------------
// Main function for cbrpserver
//-----------------------------------------------------------------------
int main(int argc, char **argv)
{
    FILE *library = NULL;
    FILE *fparams = NULL;  // File containing input parameters
    FILE *fmetrics = NULL; // File containint output metrics
    FILE *caseOut = stdout;

    char lline[LINE_BUFFER_LENGTH];
    char pline[LINE_BUFFER_LENGTH];
    char mline[LINE_BUFFER_LENGTH];
    char *cret = NULL;
    int iret = -1;

    double val;
    char sval[1000];
    char *token = NULL;
    bool wroteMetrics = false;

    /**** MISSION EXPERT PARAMETER INFORMATION *****************************************************************
    * This information comes from the .mission_expertrc file (perhaps parse that to obtain this in the future) */
    
    // Determines whether to skip a particular parameter (e.g. global ones like environment type
    // which are the same throughout)
    const bool skip_feature[]  = {false, false,  false,  false,  false,  false, false, false, false, false, false, true, true, true, true};

    // Determines feature type (int, double, or boolean) so we know how to format output
    const int feature_types[] = {CBRPLANNER_FORMAT_INT, CBRPLANNER_FORMAT_DOUBLE, CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE,CBRPLANNER_FORMAT_DOUBLE};
    /**********************************************************************************************************/

 
    int debug = 0;

    if (argc < 4) {
      fprintf(stderr, "Usage: %s <case_library_filename> <parameterization_filename> <output_metrics_filename> [output_filename]\n", argv[0]);
      return EXIT_FAILURE;
    }

    library = open_file_or_exit(argv[1], "r", "case library");
    fparams = open_file_or_exit(argv[2], "r", "parameterization");
    fmetrics = open_file_or_exit(argv[3], "r", "output metrics");

    if (argc > 4) {
      caseOut = open_file_or_exit(argv[4], "w", "case output");
    }
 
    int line_count = 1;
    while ((cret = fgets(pline, LINE_BUFFER_LENGTH-1, fparams)) != NULL) {
      STRIP_NEWLINE(pline);

      // Set some state variables for this loop
      wroteMetrics = false;

      if (line_count <= PARAMS_NUM_LINES_TO_SKIP) {
        line_count++;
        continue;
      }

      printf("PARAMETER LINE: %s\n", pline);

      if ( (cret = fgets(mline, LINE_BUFFER_LENGTH-1, fmetrics)) == NULL) {
        fprintf(stderr, "Error: Have a line %d for input parameters but not output metrics!\n", line_count);
        break;
      }
      STRIP_NEWLINE(pline);

      if (debug > 2) printf("Got lines:\n pline:%s\n mline:%s\n", pline, mline);


      // Write out all of the input parameter values 
      int pcount = 0;
      while ( (cret = fgets(lline, LINE_BUFFER_LENGTH-1, library)) != NULL) {
        STRIP_NEWLINE(lline);

printf("LINE: %s\n", lline);

        if (strstr(lline, DECISION_TREE_CASE_STRING) != NULL || 
                   strstr(lline, DECISION_TREE_NODE_CASE_STRING) != NULL) {
          // Skip decision tree features, since they only occur once at the end of the file
          pcount++;
          continue;
        }
        // Check if the line contains a parameter value
        else if (strstr(lline, INPUT_PARAMETER_CASE_STRING) == NULL) {
printf("PARAMETER\n");
          // It doesn't... Spit it back out to case file, but do not do so for 
          // existing data in the case file pertaining to output metrics, because
          // we will replace it with our own version

          if (strstr(lline, OUTPUT_METRICS_CASE_STRING) == NULL)
            fprintf(caseOut, "%s\n", lline);
          else if (wroteMetrics == false) {
            // Write out all of output metrics
            char *saveptr; // This is needed because strtok is not re-entrant, while strtok_r is
            int mcount = 0;
            int moffset = 8;
            while(1) {
              if (mcount == 0) token = strtok_r(mline, ",", &saveptr);
              else token = strtok_r(NULL, ",", &saveptr);
              if (token == NULL) break;
              if (mcount+moffset+1 >= NUM_CBRPLANNER_EXTRADATA_KEYS) {
                fprintf(stderr, "Error: Number of cbr planner extra data keys from offset (%d) is less than number of metrics in metrics file (%s: %d)\n", NUM_CBRPLANNER_EXTRADATA_KEYS-moffset-1, argv[1], (mcount+1) );
                continue;
              }
              val = atof(token);
              fprintf(caseOut, "%s START\n", OUTPUT_METRICS_CASE_STRING);
              fprintf(caseOut, "%s-KEY %s\n", OUTPUT_METRICS_CASE_STRING, CBRPLANNER_EXTRADATA_KEYS[moffset+mcount].c_str());
              fprintf(caseOut, "%s-VALUE %.2lf\n", OUTPUT_METRICS_CASE_STRING, val);
              fprintf(caseOut, "%s END\n", OUTPUT_METRICS_CASE_STRING);
              mcount++;
            }
            wroteMetrics = true;
          }
          continue;
        } else if (skip_feature[pcount]) {
          pcount++;
          fprintf(caseOut, "%s\n", lline);
          continue;
        }

        if (pcount == 0) {
          token = strtok(pline,",");
        } else {
          token = strtok(NULL,",");
        }

        if (token == NULL) {
          fprintf(stderr, "Error: There were less parameters in the param file (%s) than there are instances of parameter lines in the case library file (%s)! \n", argv[2], argv[1]);
          continue;
        }

        val = atof(token);
        if (debug > 1) printf("Line %d: Param %lf\n", line_count, val);

        if (feature_types[pcount] == CBRPLANNER_FORMAT_INT) {
          sprintf(sval, "%d", (int)val);
        } else if (feature_types[pcount] == CBRPLANNER_FORMAT_DOUBLE) {
          sprintf(sval, "%.2lf", val);
        }
        fprintf(caseOut, "%s %s\n", INPUT_PARAMETER_CASE_STRING, sval);

        pcount++;
      };
      rewind(library);
 
      if (0) {
        fprintf(stderr, "Error: Have a line %d for input parameters but not output metrics!\n", line_count);
      }
      
      line_count++;
    }

    // Write out decision tree at the end of the case library file (only one instance 
    // of it exists at the end of the file
    // TODO NOTE: It's not clear how the decision tree is used, but I think it probably
    // has to be modified, even though all of the cases we are adding have the same
    // index features
    while ( (cret = fgets(lline, LINE_BUFFER_LENGTH-1, library)) != NULL) {
      STRIP_NEWLINE(lline);
      if (strstr(lline, DECISION_TREE_CASE_STRING) != NULL || 
          strstr(lline, DECISION_TREE_NODE_CASE_STRING) != NULL) {
          fprintf(caseOut, "%s\n", lline);
      }
    }

    fclose(library); 
    fclose(fparams); 
    fclose(fmetrics); 
    if (caseOut != stdout)
      fclose(caseOut);

    return EXIT_SUCCESS;
}
