/**********************************************************************
 **                                                                  **
 **                     database_sharing.cpp                         **
 **                                                                  **
 **                                                                  **
 **  Written by:  Zsolt Kira                                         **
 **                                                                  **
 **  This file contains functions for sharing of information between **
 **  robots regarding their positions, objects, etc.  This is done   **
 **  via the database broadcast mechanism.                           **
 **                                                                  **
 **  Copyright 2009, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/*
   TODO: 
        - Make world_objects in sensor_blackboard a hashmap so that id
          can be used to retrieve elements (that way size of array won't
          be dependent on largest ID used)

        - Possibly convert to a better socket mechanism? (Downside is 
          that there is yet-another-socket-connection to go wrong)

   DONE:
        ./ Set up caching mechanism so that old objects that are not read
          in right aren't ommitted during any cycle

*/

#include "sensor_blackboard.h"
#include "database_sharing.h"

// From robot_side_com.c
extern int robot_id;

//#define DATABASE_SHARING_VIA -1
#define DATABASE_SHARING_VIA THIS_ROBOT

#define safe_strcpy(dest, src) strncpy(dest, src, DATABASE_SHARING_MAX_STRING_LENGTH-1);  dest[DATABASE_SHARING_MAX_STRING_LENGTH-1] = '\0';

int database_get_num_robot_positions() {
    int num_robots = 0;
    char *dbvalue = NULL;
    char value[DATABASE_SHARING_MAX_STRING_LENGTH];
bool debug = true;
    dbvalue = get_state("NumberOfRobots");

    if (dbvalue != NULL) {
        safe_strcpy(value, dbvalue);
    } else {
printf("NUMBER OF ROBOTS IS NULL!\n");
        safe_strcpy(value, "");
    }

    if (strcmp(value, "") != 0) {
        num_robots = atoi(value);
        if (debug) printf("NumberOfRobots database value has %d!\n", num_robots);
    } else {
        fprintf(stderr, "\n%d: No NumberOfRobots database value!\n", robot_id);
        num_robots = 2; // TODO REMOVE THIS DEFAULT
    }
    return num_robots;
}

void database_get_robot_positions() {
       // Get it from position broadcasts by other robots... if not, only
       // supported in simulation
       int i;
       int num_robots = 1;  // There is at least yourself!
       int db_size = 0;
       info_sharing_robot *object = NULL;
bool debug = true;
       char *dbvalue = NULL;
       char value[DATABASE_SHARING_MAX_STRING_LENGTH];

       char *token=NULL;
       char keyString[10000];
       double timestamp = 0.0;

       robot_position robot_loc = exec_get_position();
       Vector vec;
       double x = 0.0, y = 0.0, z = 0.0;
       int id = 0;
       bool errorFlag = false;
       num_robots = database_get_num_robot_positions();

       // Go through all of the robots (subtract one since each robot skips itself)

       // NOTE: There is some kind of memory problem with RobotReadings.  If it is allocated
       // and used throughout, a segfault results.  If values are stored and the RobotReadings
       // are allocated and filled in at the end (right before it is returned) it seems to work
       // TODO: Memory issue with RobotReadings should be looked into/fixed!!!!
       //RobotReadings rtn(num_robots);

       // Temporarily store these values, to fill in RobotReadings at the end
       int index = 0;
       for (i = 0; i < num_robots; i++) {
           sprintf(keyString, "Robot%d-Position", i+1);
           dbvalue = get_state(keyString);

           if (dbvalue != NULL) {
             safe_strcpy(value, dbvalue);
           } else {
             safe_strcpy(value, "");
           }

           errorFlag = false;
           if (strcmp(value, "") != 0) {
               if (debug) printf("Robot %d position in database: %s\n", i, value);

               // Skip yourself
               //if (i == robot_id) continue;

               id = i;

               token = strtok(value, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get timestamp\n", robot_id);
                 errorFlag = true;
               }
               else  timestamp = atof(token);

               token = strtok(NULL, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get x\n", robot_id);
                 errorFlag = true;
               }
               else  vec.x = atof(token);

               token = strtok(NULL, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get y\n", robot_id);
                 errorFlag = true;
               }
               else  vec.y = atof(token);

               token = strtok(NULL, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get z\n", robot_id);
                 errorFlag = true;
               }
               else  vec.z = atof(token);

           } else {
               fprintf(stderr, "%d: No position data for robot %d (key=%s)!\n", robot_id, i+1, keyString);
           }

           if (errorFlag == false) {
               db_size = sensor_blackboard.world_robots.size();

               if (db_size > id && sensor_blackboard.world_robots[id] != NULL) {
if (1||debug) printf("UPDATING ROBOT %d\n", id);
                 // TODO THESE SHOULD BE GLOBAL POSITIONS, MAKE RELATIVE AT BEHAVIOR
                 // Already have a record of this object, update location
                 object = sensor_blackboard.world_robots[id];
                 object->x = vec.x;
                 object->y = vec.y;

               } else {
if (1||debug) printf("CREATING NEW ROBOT %d\n", id);
                 // Create new record for this object
                 object = new info_sharing_robot();

                 object->id = id;
                 object->x = vec.x;
                 object->y = vec.y;
                 object->z = 0;

                 // Fill up vector to right before new ID number
                 for (int j = db_size; j < id; j++) {
                   sensor_blackboard.world_robots.push_back(NULL);
                 }
if (1||debug) printf("ADDING NEW ROBOT %d\n", id);
                 sensor_blackboard.world_robots.push_back(object);
if (1||debug) printf("DONE ADDING NEW ROBOT %d\n", id);
               }


               index++;
           }
       }

if (1||debug) printf("DONE GETTING ROBOTS\n");
}

int database_get_num_world_robot_positions() {
  return sensor_blackboard.world_robots.size();
}

void database_fill_robot_positions(RobotReadings &rtn, int &new_num_objects) {
       int num_objects;
       int i;
       int index = 0;
       robot_position robot_loc = exec_get_position();
       Vector vec;
       info_sharing_robot *object = NULL;

       // Fill in the values (skip partial ones)
       new_num_objects = num_objects = sensor_blackboard.world_robots.size();
       for (i = 0; i < num_objects; i++) {
printf("GETTING ROBOT %d\n", i);
         object = sensor_blackboard.world_robots[i];
         if (object == NULL)
           continue;

printf("VALID!\n");
         // Convert global coordinates to egocentric
         vec.x = object->x - robot_loc.v.x;
         vec.y = object->y - robot_loc.v.y;
         rotate_z(vec, -1.0 * robot_loc.heading);

         //printf("Filling in object i=%d id=%d x=%f, y=%f\n", i, id[i], x[i], y[i]);
         rtn.readings[index].id = 0; // object->id;
         rtn.readings[index].v.x = vec.x;
         rtn.readings[index].v.y = vec.y;
printf("DONE!\n");
       }

}

void database_put_robot_position() {
    timeval tvp;
    char positionString[100000];
    char keyString[10000];

    // Get current time for timestamp
    gettimeofday(&tvp, NULL);
    double absTime = (double)(tvp.tv_sec) + (((double)(tvp.tv_usec))/1000000.0);

    // Make key
    sprintf(keyString, "Robot%d-Position", robot_id);

    // Make data.  The format is comma-separated list containing timestamp and 6D pose
    sprintf(positionString, "%f,%f,%f,%f", absTime, sensor_blackboard.x, sensor_blackboard.y, sensor_blackboard.theta);

    if (debug) printf("Robot %d: Broadcasting %s-%s\n", robot_id, keyString, positionString);

    // Broadcast to the other robots
    exec_broadcast_value(keyString, positionString, DATABASE_SHARING_VIA);
    //put_state(keyString, positionString);

}


int database_get_num_objects(int robot_id) {
bool debug = true;
    int num_objects = 0;
    char *dbvalue = NULL;
    char value[10000];
    char keyString[10000];

    sprintf(keyString, "Robot%d-Objects", robot_id);
    dbvalue = get_state(keyString);
    if (dbvalue != NULL) {
      safe_strcpy(value, dbvalue);
    } else { 
      safe_strcpy(value, "");
    }

    if (strcmp(value, "") != 0) {
        sscanf(value, "NumObjects=%d", &num_objects);
        if (debug) printf("Robot %d has %d objects in database (%s)!\n", robot_id, num_objects, value);
    } else {
        fprintf(stderr, "%d: No Objects database value (key=%s)!\n", robot_id, keyString);
    }
    return num_objects;
}

void database_get_objects(int robot_id) {
       // Get objects from broadcasts by other robots...
       int num_objects = 0;
       int i = 0;
       int db_size = 0;
       info_sharing_object *object = NULL;
bool debug = true;
       num_objects = database_get_num_objects(robot_id);
       if (num_objects <= 0) {
           return;
       }

       int id = 0;
       char *dbvalue = NULL;
       char value[10000];
       char *token=NULL;
       char keyString[10000];

       double timestamp = 0.0;
       char color[10000];

       // Go through all of the objects

       // NOTE: There is some kind of memory problem with RobotReadings.  If it is allocated
       // and used throughout, a segfault results.  If values are stored and the RobotReadings
       // are allocated and filled in at the end (right before it is returned) it seems to work
       // TODO: Memory issue with RobotReadings should be looked into/fixed!!!!
       //RobotReadings rtn(num_objects);

       // Temporarily store these values, to fill in RobotReadings at the end
       robot_position robot_loc = exec_get_position();
       Vector vec;
       bool errorFlag = false;
       int index = 0;

       for (i = 0; i < num_objects; i++) {
           sprintf(keyString, "Robot%d-Object%d", robot_id, i);
           dbvalue = get_state(keyString);

           if (dbvalue != NULL) {
             safe_strcpy(value, dbvalue);
           } else {
             safe_strcpy(value, "");
           }

           errorFlag = false;
           if (strcmp(value, "") != 0) {
               if (debug) printf("Object %d position in database: %s (array position=%d)\n", i, value, index);

               token = strtok(value, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get timestamp\n", robot_id);
                 errorFlag = true;
               }
               else  sscanf(token, "time=%lf", &timestamp);

               token = strtok(NULL, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get x\n", robot_id);
                 errorFlag = true;
               }
               else sscanf(token, "id=%d", &id);

               token = strtok(NULL, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get x\n", robot_id);
                 errorFlag = true;
               }
               else sscanf(token, "x=%lf", &vec.x);

               token = strtok(NULL, ",");
               if (token == NULL) {
                 fprintf(stderr, "Error (%d): While detecting robots from broadcast, could not get y\n", robot_id);
                 errorFlag = true;
               }
               else sscanf(token, "y=%lf", &vec.y);

               //token = strtok(NULL, ",");
               //if (token == NULL) {
               //  fprintf(stderr, "Error: While detecting robots from broadcast, could not get color\n");
               //  errorFlag = true;
               //}
               //else {
               //  sscanf(token, "color=%s", color);
               //  objs.objects[index].objcolor = strdup(color);
               //}

           } else {
               fprintf(stderr, "Warning (%d): No object data for robot %d, object %d (key=%s)!\n", robot_id, robot_id, i, keyString);
               errorFlag = true;
           }

           if (errorFlag == false) {
               
               // Convert global coordinates to egocentric
               //vec.x = vec.x - robot_loc.v.x;
               //vec.y = vec.y - robot_loc.v.y;
               //rotate_z(vec, -1.0 * robot_loc.heading);

               db_size = sensor_blackboard.world_objects.size();

               if (db_size > id && sensor_blackboard.world_objects[id] != NULL) {
if (1||debug) printf("UPDATING OBJECT %d\n", id);
                 // TODO THESE SHOULD BE GLOBAL POSITIONS, MAKE RELATIVE AT BEHAVIOR
                 // Already have a record of this object, update location
                 object = sensor_blackboard.world_objects[id];
                 object->x = vec.x;  
                 object->y = vec.y;
  
               } else {
if (1||debug) printf("CREATING NEW OBJECT %d\n", id);
                 // Create new record for this object
                 object = new info_sharing_object();

                 object->id = id;
                 object->x = vec.x;
                 object->y = vec.y;
                 //strcpy(object->color, "red");
                 object->r = 0.01;
                 //object->objshape = CircleObject;

                 // Fill up vector to right before new ID number 
                 for (int j = db_size; j < id; j++) {
                   sensor_blackboard.world_objects.push_back(NULL);
                 } 
if (1||debug) printf("ADDING NEW OBJECT %d\n", id);
                 sensor_blackboard.world_objects.push_back(object);
if (1||debug) printf("DONE ADDING NEW OBJECT %d\n", id);
               }
               
               index++;
           }
       }
}

int database_get_num_world_objects() {
       return sensor_blackboard.world_objects.size();
}

void database_fill_objects(ObjectList &objs, int &new_num_objects, int robot_id, int objects_start_index) {
       int num_objects;
       int i;
       int index = 0;
       robot_position robot_loc = exec_get_position();
       Vector vec;
       info_sharing_object *object = NULL;

       // Fill in the values (skip partial ones)
       new_num_objects = num_objects = sensor_blackboard.world_objects.size();
       for (i = 0; i < num_objects; i++) {
printf("GETTING %d\n", i);
         object = sensor_blackboard.world_objects[i];
         if (object == NULL)
           continue;

printf("VALID!\n");
         // Convert global coordinates to egocentric
         vec.x = object->x - robot_loc.v.x;
         vec.y = object->y - robot_loc.v.y;
         rotate_z(vec, -1.0 * robot_loc.heading);

         //printf("Filling in object i=%d id=%d x=%f, y=%f\n", i, id[i], x[i], y[i]);
         objs.objects[index].id = 0; // object->id;
         objs.objects[index].x1 = vec.x; 
         objs.objects[index].y1 = vec.y; 
         objs.objects[index].x2 = vec.x; 
         objs.objects[index].y2 = vec.y;
         objs.objects[index].setColor("red");
         objs.objects[index].r = 0.01;
         objs.objects[index].objshape = CircleObject;
         objs.objects[index].ismovable = false;
         index++;
printf("DONE!\n");
       }
}

void database_put_objects(ObjectList &objs, int objects_start_index, int rid) {
    timeval tvp;
    char objectString[100000];
    char keyString[10000];
    int index = objects_start_index;
    int max_id = 0;
debug = 1;
    if (rid < 0)
      rid = robot_id;

    if (debug) printf("Robot %d: Broadcasting %d objects\n", rid, objs.count);

    // Get current time for timestamp
    gettimeofday(&tvp, NULL);
    double absTime = (double)(tvp.tv_sec) + (((double)(tvp.tv_usec))/1000000.0);

    for (int i = 0; i < objs.count; i++) {
      // Make key
      sprintf(keyString, "Robot%d-Object%d", rid, objs.objects[index].id);

      if (objs.objects[index].id > max_id)
        max_id = objs.objects[index].id;

      // Make data.  The format is comma-separated list containing timestamp and 6D pose
      sprintf(objectString, "time=%f,id=%d,x=%f,y=%f", absTime, objs.objects[index].id, objs.objects[index].x1, objs.objects[index].y1);

      if (debug) printf("Robot %d: Broadcasting %s:%s\n", rid, keyString, objectString);

      // Broadcast to the other robots
      exec_broadcast_value(keyString, objectString, DATABASE_SHARING_VIA);
      //put_state(keyString, objectString);

      index++;
    }
    sprintf(keyString, "Robot%d-Objects", rid);
    sprintf(objectString, "NumObjects=%d", max_id+1 );
    exec_broadcast_value(keyString, objectString, DATABASE_SHARING_VIA);
    //put_state(keyString, objectString);

}

