/**********************************************************************
 **                                                                  **
 **                          RComInterthread.cc                      **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains functions for the RComInterthread class.     **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: RComInterthread.cc,v 1.1 2008-07-15 22:33:59 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <netdb.h>
#include <pthread.h>
#include <fstream>
#include <iostream>
#include <string>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <pthread.h>

#include "RComInterthread.h"
#include "RComDefaultSetups.h"
#include "RComIDs.h"

using std::string;

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
extern bool gRComDebug;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constructor for RComInterthread class.
//-----------------------------------------------------------------------
RComInterthread::RComInterthread(void) :
    myID_(RCOM_USER_ID_INVALID),
    interRComPort_(DEFAULT_INTER_RCOM_PORT),
    extraCallbackFuncArg_(NULL),
    controllerInitialized_(false),
    applicationCallbackFunc_(NULL)
{
    setPeerRComIPs(DEFAULT_PEER_RCOM_IP);

    if (gRComm_Controller == NULL)
    {
        gRComm_Controller = new RComm_Controller();
    }

    rcommController_ = gRComm_Controller;
}

//-----------------------------------------------------------------------
// Distructor for RComInterthread class.
//-----------------------------------------------------------------------
RComInterthread::~RComInterthread(void)
{
    if ((myID_ > RCOM_USER_ID_INVALID) && (rcommController_ != NULL))
    {
        rcommController_->Unregister_App(myID_);
    }
}

//-----------------------------------------------------------------------
// This function will be called by RComm_Controller when data is arrived.
//-----------------------------------------------------------------------
void RComInterthread::controllerCallback_(
    int recipientID,
    int mybufferlength,
    bool reliabilityEnforced,
    void *applicationInstance)
{
    RComInterthread *rcomInterthreadInstance = NULL;
    int myID;

    if (applicationInstance == NULL)
    {
        return;
    }

    rcomInterthreadInstance = (RComInterthread *)applicationInstance;

    // Check ID.
    myID = rcomInterthreadInstance->myID_;

    if (myID != recipientID)
    {
        fprintf(
            stderr,
            "Error: RComInterthread::controllerCallback_(). Wrong recipient ID.\n");
        return;
    }

    // Call the callback function of the application.
    if (rcomInterthreadInstance->applicationCallbackFunc_ != NULL)
    {
        rcomInterthreadInstance->applicationCallbackFunc_(
            myID,
            rcomInterthreadInstance->localReadDataBuffer_,
            mybufferlength,
            reliabilityEnforced,
            rcomInterthreadInstance->extraCallbackFuncArg_);
    }
}

//-----------------------------------------------------------------------
// This function initializes the RComInterthread.
//-----------------------------------------------------------------------
bool RComInterthread::initialize(
    int myID,
    void (*applicationCallbackFunc)(int, const void *, size_t, bool, void *),
    void *extraCallbackFuncArg)
{
    // Remember my ID.
    myID_ = myID;

    // Callback function that the application code supplies and its extra
    // argument.
    applicationCallbackFunc_ = applicationCallbackFunc;
    extraCallbackFuncArg_ = extraCallbackFuncArg;

    // Initialize the controller if it has been done already.
    if (!(rcommController_->initialized()))
    {
        //Initialize the object with the IPs and a port number.
        if(rcommController_->init(peerRComIPs_, interRComPort_))
        {
            fprintf(stderr,"RCom initialized.\n");
        }
        else
        {
            fprintf (stderr,"Error: RComInterthread::initialize(). Initialization failed.\n");
            return false;
        }
    }

	// Register the callback and and the data buffer to RComm_Controller.
	rcommController_->Register_App(
        (Receiver_Callback_p)controllerCallback_,
        (unsigned int)myID_,
        (unsigned char *)localReadDataBuffer_,
        (void *)this);

    return true;
}

//-----------------------------------------------------------------------
// This function sets the peer RCom IPs.
//-----------------------------------------------------------------------
void RComInterthread::setPeerRComIPs(const char peerRComIPs[RCOM_NUM_PATHS][IP_ARRAY_SIZE])
{
    int i;

    for (i = 0; i < RCOM_NUM_PATHS; i++)
    {
        strcpy(peerRComIPs_[i], peerRComIPs[i]);
    }
}

//-----------------------------------------------------------------------
// This function will ship the data
//-----------------------------------------------------------------------
bool RComInterthread::sendData(
        int recipientID,
        const void *data,
        size_t dataSize,
        bool enfoceReliable)
{
    if (dataSize > RCOM_MAX_PAYLOAD_SIZE)
    {
        fprintf(stderr, "Error: Data size exceeds RCOM_MAX_PAYLOAD_SIZE.\n");
        return false;
    }

    if (gRComDebug)
    {
        rcommController_->Print_Stats();
    }

    if (enfoceReliable)
    {
        rcommController_->Send_Packet_Reliable(
            (byte*)data,
            (unsigned int)dataSize,
            (unsigned int)recipientID);
    }
    else
    {
        rcommController_->Send_Packet_Unreliable(
            (byte*)data,
            (unsigned int)dataSize,
            (unsigned int)recipientID);
    }

    return true;
}

//-----------------------------------------------------------------------
// This function returns the status of RCom.
//-----------------------------------------------------------------------
string RComInterthread::getRComStatus(void)
{
    string status = "";
    const char *statusPtr = NULL;

    statusPtr = rcommController_->Get_Stats();

    if (statusPtr == NULL)
    {
        return status;
    }

    status = statusPtr;

    return status;
}

/**********************************************************************
 * $Log: RComInterthread.cc,v $
 * Revision 1.1  2008-07-15 22:33:59  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.3  2006/08/09 21:59:55  endo
 * gDebug -> gRComDebug
 *
 * Revision 1.2  2006/08/09 21:46:53  endo
 * -ansi -pedantic added to Makefile
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.6  2006/04/14 11:24:35  endo
 * RComInterthread::getRComStatus() added.
 *
 * Revision 1.5  2006/04/02 22:46:29  endo
 * Unregister_App() added to Reliable_Comm.
 *
 * Revision 1.4  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.3  2006/02/25 22:10:23  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.2  2006/02/25 16:17:49  endo
 * rcomserver added
 *
 * Revision 1.1  2006/02/24 11:50:34  endo
 * RComm_Controller checked in.
 *
 **********************************************************************/
