/**********************************************************************
 **                                                                  **
 **                           RComServer.h                           **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This is a header file for RComServer.cc.                        **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef RComServer_H
#define RComServer_H

/* $Id: RComServer.h,v 1.1 2008-07-15 22:34:00 endo Exp $ */

#include <pthread.h>
#include <string>
#include <vector>

#include "RComAPI.h"

using std::string;
using std::vector;


class RComServer : public RComAPI {

    typedef struct PrivateConnectionInfo_t
    {
        int fd;
        int clientID;
        pthread_t connectionThread;
        bool connectionThreadIsUp;
        RComAPI *rcomHandler;
    };

    typedef struct PrivateConnectionThreadInput_t
    {
        int fd;
        RComServer *rcomServerInstance;
    };

    typedef struct QueuedData_t
    {
        int recipientID;
        char *data;
        int dataSize;
        bool enfoceReliable;
    };

protected:
    RComAPI *rcomHandlerMain_;
    pthread_t daemonThread_;
    vector<PrivateConnectionInfo_t> shPrivateConnectionInfoList_;
    vector<QueuedData_t> shQueuedDataList_;
    int localPort_;
    int socketFd_;
    int myID_;
    bool daemonThreadIsUp_;
    void *extraCallbackFuncArg_;

    void (*readerCallbackFunc_)(int, const void *, size_t, bool, void *);

    static pthread_mutex_t privateConnectionInfoListMutex_;
    static pthread_mutex_t queuedDataMutex_;

    static const string EMPTY_STRING_;
    static const int INVALID_SOCKET_FD_;
    static const int SOCKET_BACKLOG_SIZE_;
    static const int MAX_SYNC_ERROR_COUNT_;
    static const int MAX_QUEUED_DATA;

    pthread_t getConnectionThread_(const int fd);
    int waitForNewConnection_(int count);
    int readInteger_(const int fd);
    int getClientID_(const int fd);
    bool openSocket_(int port);
    bool readAndProcessPrivateConnectionData_(const int fd);
    bool syncToClient_(const int fd);
    bool readNBytes_(const int fd, unsigned char *buf, const int n);
    bool initClient_(const int fd, const int clientID);
    bool isClient_(int id, bool *connected);
    void startPrivateConnection_(const int fd);
    void saveConnectionThread_(const int fd, const pthread_t connectionThread);
    void daemonMainLoop_(void);
    void privateConnectionMainLoop_(const int fd);
    void setConnectionThreadIsUp_(const int fd, const bool isUp);
    void disconnectPrivateConnection_(const int fd);
    void sendAllQueuedData_(int fd);
    void sendQueuedData_(int fd, QueuedData_t queuedData);
    void writeInteger_(const int fd, int value);
    void readAndProcessClientData_(const int fd);
    void disconnectAllPrivateConnections_(void);
    void removeRComHandler_(const int fd);

    static void cbProcessRecievedDataFromRComPeer_(
        int recipientID,
        const void *data,
        size_t dataSize,
        bool reliableEnforced,
        void *extraArg);

    static void *startDaemonThread_(void *rcomServerInstance);
    static void *startPrivateConnectionThread_(void *input);

public:
    RComServer(void);
    ~RComServer(void);
    RComAPI *getMainRComHandler(void);
    bool initialize(
        int myID,
        void (*readerCallbackFunc)(int, const void *, size_t, bool, void *),
        void *extraCallbackFuncArg);
    bool sendData(
        int recipientID,
        const void *data,
        size_t dataSize,
        bool enfoceReliable);
    void setLocalPort(int port);
    //void setInterRComPort(int port);
    //void setPeerRComIPs(string peerRComIPs[]);
};

inline void RComServer::setLocalPort(int port)
{
    localPort_ = port;
}

inline RComAPI *RComServer::getMainRComHandler(void)
{
    return rcomHandlerMain_;
}

#endif
/**********************************************************************
 * $Log: RComServer.h,v $
 * Revision 1.1  2008-07-15 22:34:00  endo
 * OpenCV, Gazebo and USARSim added.
 *
 * Revision 1.2  2006/08/09 21:46:53  endo
 * -ansi -pedantic added to Makefile
 *
 * Revision 1.1  2006/08/09 21:12:37  endo
 * RCom added.
 *
 * Revision 1.6  2006/04/05 22:09:54  endo
 * Max queued data implemented.
 *
 * Revision 1.5  2006/04/02 22:46:29  endo
 * Unregister_App() added to Reliable_Comm.
 *
 * Revision 1.4  2006/02/27 20:41:54  endo
 * rcomserver bug fixed.
 *
 * Revision 1.3  2006/02/25 22:10:23  endo
 * The user can now specify ports to be used.
 *
 * Revision 1.2  2006/02/25 16:17:49  endo
 * rcomserver added
 *
 * Revision 1.1  2006/02/22 07:17:22  endo
 * RComServer and RComClient implemented.
 *
 **********************************************************************/
