/**********************************************************************
 **                                                                  **
 **                          gazebo-laser.c                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for Gazebo laser (Player/Stage)                       **
 **                                                                  **
 **  Copyright 2006, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gazebo-laser.c,v 1.4 2009-01-27 20:23:09 zkira Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "gazebo-laser.h"
#include "Watchdog.h"
#include "LogManager.h"

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
GazeboLaser *gGazeboLaser = NULL;

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t GazeboLaser::readingsMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Statics
//-----------------------------------------------------------------------
Sensor *GazeboLaser::sensorLaser_ = NULL;
vector<GazeboLaser*> GazeboLaser::laserList_;
float *GazeboLaser::allAngles_ = NULL;
float *GazeboLaser::allReadings_ = NULL;
int GazeboLaser::totalReadings_ = 0;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const struct timespec GazeboLaser::READER_SLEEPTIME_NSEC_ = {0, 100000000}; // 100 msec.
const string GazeboLaser::DEFAULT_GAZEBO_LASER_NAME_ = "laser1";
const int GazeboLaser::SKIP_STATUSBAR_UPDATE_ = 5;
const int GazeboLaser::WATCHDOG_CHECK_INTERVAL_SEC_ = 2;
const int GazeboLaser::DEFAULT_GAZEBO_SERVER_ID_ = 0;
const int GazeboLaser::DEFAULT_GAZEBO_CLIENT_ID_ = 0;

const double GazeboLaser::DEFAULT_GAZEBO_LASER_FOV_   = 180.0;
const double GazeboLaser::DEFAULT_GAZEBO_LASER_RANGE_ = 80.0;

RobotPartLaser* GazeboLaser::m_pclPartLaser = NULL;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
GazeboLaser::GazeboLaser(
    GazeboLaser **a,
    char *gazeboLaserName,
    float angleOffset, 
    float xOffset,
    float yOffset,
    float laserPitchAngle) : 
    Module((Module**)a, HS_MODULE_NAME_LASER_GAZEBO),
    gazeboClient_(NULL),
    gazeboLaser_(NULL),
    shAngles_(NULL),
    shReadings_(NULL),
    shRawAngles_(NULL),
    shRawReadings_(NULL),
    angleOffset_(angleOffset), 
    xOffset_(xOffset),
    yOffset_(yOffset),
    shPitchAngle_(laserPitchAngle),
    gazeboServerID_(DEFAULT_GAZEBO_SERVER_ID_),
    gazeboClientID_(DEFAULT_GAZEBO_CLIENT_ID_),
    shMinReadingIndex_(0),
    numReadings_(0),
    readerThreadIsUp_(false)
{

    GazeboLaserFOV_   = DEFAULT_GAZEBO_LASER_FOV_;
    GazeboLaserRange_ = DEFAULT_GAZEBO_LASER_RANGE_;
    
    // Set up the laser name.
    gazeboLaserName_ = (gazeboLaserName != NULL)?
        gazeboLaserName : DEFAULT_GAZEBO_LASER_NAME_;

    // Initialize mutexes.
    pthread_mutex_init(&readingsMutex_, NULL);

    // Set up the laser list.
    laserList_.insert(laserList_.begin(), (GazeboLaser*)NULL);
    laserList_[0] = this;

    // Update the status bar.
    updateStatusBar_();

    // Initialize the gazebo.
    if (!initializeGazebo_())
    {
        printfTextWindow("Error: GazeboLaser could not be initialized. Exiting Gazebo.\n");
        delete this;
        return;
    }

    // Initialize the buffer.
    addToOverallBuffer_();

    // Register the laser as a sensor.
    addSensor_();

    sensorLaser_->setPitchAngle(shPitchAngle_);

    // Add Laser part so that configuration data about this
    // sensor will be correctly sent to the robot executable
    m_pclPartLaser = new RobotPartLaser(numReadings_, 0, GazeboLaserFOV_, GazeboLaserRange_);    

    // Start the reader thread.
    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);

    // Start the watchdog.
    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(
            NAME_,
            WATCHDOG_CHECK_INTERVAL_SEC_,
            getpid(),
            readerThread_);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
GazeboLaser::~GazeboLaser(void)
{
    int i;

    if (readerThreadIsUp_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    if (m_pclPartLaser != NULL)
    {
        delete m_pclPartLaser;
        m_pclPartLaser = NULL;
    }

    // Close the interface and free object
    if (gazeboLaser_ != NULL)
    {
        gz_laser_close(gazeboLaser_);
        gz_laser_free(gazeboLaser_);
        gazeboLaser_ = NULL;
    }

    // Close the connection and free the client object
    if (gazeboClient_ != NULL)
    {
        gz_client_disconnect(gazeboClient_);
        gz_client_free(gazeboClient_);
        gazeboClient_ = NULL;
    }

    if (shReadings_ != NULL)
    {
        delete [] shReadings_;
        shReadings_ = NULL;
    }

    if (shAngles_ != NULL)
    {
        delete [] shAngles_;
        shAngles_ = NULL;
    }

    if (shRawReadings_ != NULL)
    {
        delete [] shRawReadings_;
        shRawReadings_ = NULL;
    }

    if (shRawAngles_ != NULL)
    {
        delete [] shRawAngles_;
        shRawAngles_ = NULL;
    }

    // Only delete the sensor if this is the LAST laser being deleted
    if ((laserList_.size() == 1) && (sensorLaser_ != NULL))
    {
        delete sensorLaser_;
        sensorLaser_ = NULL;
    }

    // Remove this laser object from the static list
    for (i = 0; i < (int)(laserList_.size()); i++)
    {
        if (laserList_[i] == this)
        {
            laserList_.erase(laserList_.begin() + i);
            break;
        }
    }

    // Clean up the buffer.
    removeFromOverallBuffer_();

    printTextWindow("GazeboLaser disconnected");

}

//-----------------------------------------------------------------------
// This function initialize gazebo
//-----------------------------------------------------------------------
bool GazeboLaser::initializeGazebo_(void)
{
    double minAngle, resAngle;
    int i;

    // Print all errors
    gz_error_init(1, 9);

    // Create a client object
    gazeboClient_ = gz_client_alloc();  

    // Connect to the server
    if (gz_client_connect_wait(gazeboClient_, gazeboServerID_, gazeboClientID_) != 0)
    {
        printfTextWindow("Error: GazeboLaser could not connect to the client.\n");
        gz_client_free(gazeboClient_);
        gazeboClient_ = NULL;
        return false;
    }

    // Create a position interface
    gazeboLaser_ = gz_laser_alloc();

    // Connect to device on server
    if (gz_laser_open(gazeboLaser_, gazeboClient_, gazeboLaserName_.c_str()) != 0)
    {
        printfTextWindow("Error: GazeboLaser could not open the position device.\n");
        gz_laser_free(gazeboLaser_);
        gazeboLaser_ = NULL;
        return false;
    }

    // Get the angles.
    gz_laser_lock(gazeboLaser_, 1);

    minAngle = GAZEBO_LASER_RAD2DEG(gazeboLaser_->data->min_angle);
    resAngle = GAZEBO_LASER_RAD2DEG(gazeboLaser_->data->res_angle);
    numReadings_ = gazeboLaser_->data->range_count;

    gz_laser_unlock(gazeboLaser_);
    if (numReadings_ > 0)
    {
        shAngles_ = new float[numReadings_];
        shReadings_ = new float[numReadings_];
        shRawAngles_ = new float[numReadings_];
        shRawReadings_ = new float[numReadings_];

        // calculate the new angles for this laser
        for (i = 0; i < numReadings_; i++)
        {
            shRawAngles_[i] = minAngle + (resAngle*((double)i));
        }

        // Initialize the rest.
        memset(shReadings_, 0x0, sizeof(float)*numReadings_);
        memset(shRawReadings_, 0x0, sizeof(float)*numReadings_);
        memcpy(shAngles_, shRawAngles_, sizeof(float)*numReadings_);
    }
    else
    {
        printfTextWindow("Error(GazeboLaser): Number of readings is zero.\n");
        return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function registers the laser readings as a sensor.
//-----------------------------------------------------------------------
void GazeboLaser::addSensor_()
{
    if (sensorLaser_ != NULL)
    {
        // only create a new sensor for the laser if one doesn't already exist
        return;
    }
    sensorLaser_ = new Sensor(
        SENSOR_SICKLMS200_1,
        sizeof(extra_laser_t), 
        (char*)&extraInfo_,
        totalReadings_,
        allAngles_, 
        totalReadings_,
        allReadings_, 
        &readingsMutex_);
}

//-----------------------------------------------------------------------
// This function assumes that the new laser has already been inserted into 
// the static laser list
//-----------------------------------------------------------------------
void GazeboLaser::addToOverallBuffer_(void)
{
    float *newReadings = NULL, *newAngles = NULL;
    float *oldReadings = NULL, *oldAngles = NULL;
    float *angleValues = NULL;
    int i;

    // save the static buffers temporarily so we can delete them later
    oldReadings = allReadings_;
    oldAngles = allAngles_;

    // make new buffers large enough to encorporate the info of the new laser
    newReadings = new float[totalReadings_ + numReadings_];
    newAngles = new float[totalReadings_ + numReadings_];

    // aquire the lock for the readings and angles -- this could cause deadlock if 
    // something tries to aquire both of these locks in an order opposite of what 
    // is here, but I don't know how to work around that. Just make sure that if 
    // these two locks are aquired at the same time, it's in this order -- and 
    // it wouldn't hurt to repeat this comment wherever that happens
    pthread_mutex_lock(&readingsMutex_);

    // Remeber the original angle values
    angleValues = new float[numReadings_];
    memcpy(angleValues, shAngles_, sizeof(float)*numReadings_);

    // give the new laser space after the old buffers. This is un-allocated memory, 
    // but it makes the upcoming transition (in the next "for" loop) easier.
    shAngles_ = oldAngles + totalReadings_;
    shReadings_ = oldReadings + totalReadings_;

    // alter all laser objects so they point to the new buffers. Offset the lasers' 
    // readings in the new buffer by the same amount that they were offset in the 
    // old buffer
    for (i = 0; i < (int)(laserList_.size()); i++)
    {
        laserList_[i]->shAngles_ = newAngles + (laserList_[i]->shAngles_ - oldAngles);
        laserList_[i]->shReadings_ = newReadings + (laserList_[i]->shReadings_ - oldReadings);
    }

    // copy the from the old buffers to the new ones
    if (totalReadings_ > 0)
    {
        memcpy(newReadings, oldReadings, sizeof(allReadings_[0]) * totalReadings_);
        memcpy(newAngles, oldAngles, sizeof(allAngles_[0] ) * totalReadings_);
    }

    // initialize the reading angles. the raw angles here will only be an approximation,
    // but it's better than initializing to 0 if the readings are needed between the time
    // that this laser is added and the time another laser reading comes in
    memcpy(shAngles_, angleValues, sizeof(float)*numReadings_);
    delete [] angleValues;
    angleValues = NULL;

    // alter the static laser data
    allReadings_ = newReadings;
    allAngles_ = newAngles;
    totalReadings_ += numReadings_;

    // if the laser sensor exists, then update it, otherwise it will be created 
    // later and initialized appropriately
    if (sensorLaser_ != NULL)
    {
        sensorLaser_->setReadingInfo(allReadings_, totalReadings_);
        sensorLaser_->setAngleInfo(allAngles_, totalReadings_);
    }

    // release the lock for the readings and angles
    pthread_mutex_unlock(&readingsMutex_);

    // get rid of the old buffers
    if (oldAngles != NULL)
    {
        delete [] oldAngles;
    }

    if (oldReadings != NULL)
    {
        delete [] oldReadings;
    }
}

//-----------------------------------------------------------------------
// This function recomputes the reading based on the offsets.
//-----------------------------------------------------------------------
void GazeboLaser::accountForXYOffset_(float &reading, float &angle)
{
    float readingXOffset, readingYOffset;
    float rad;

    rad = GAZEBO_LASER_DEG2RAD(angle); // Convert to radians
    

    // Get X and Y components of the reading
    readingXOffset = reading * cos(rad);
    readingYOffset = reading * sin(rad);

    readingXOffset += xOffset_;
    readingYOffset += yOffset_;

    reading = sqrt(readingXOffset * readingXOffset + readingYOffset * readingYOffset);

    rad = (readingXOffset == 0.0) ? 
        (M_PI / 2.0) : atan(readingYOffset / readingXOffset);

    // Quadrants 2 and 3
    if (readingXOffset < 0.0)
    {
        rad += M_PI;
    }

    // Convert back to degrees and normalize
    angle = GAZEBO_LASER_RAD2DEG(rad);
    angle += angleOffset_;
    GAZEBO_LASER_CRAMPDEG(angle, -180.0, 180.0);
}

//-----------------------------------------------------------------------
// This assumes that this laser has already been removed from the static 
// laser list
//-----------------------------------------------------------------------
void GazeboLaser::removeFromOverallBuffer_(void)
{
    float *oldReadings = NULL, *oldAngles = NULL;
    float *newReadings = NULL, *newAngles = NULL;
    int readingsInFront, readingsBehind;
    int i, offset;

    // save the static buffers temporarily so we can delete them later
    oldReadings = allReadings_;
    oldAngles = allAngles_;

    // check to see if there are laser readings other than this laser's
    if (totalReadings_ != numReadings_)
    {
        // make new, smaller buffers
        newReadings = new float[totalReadings_ - numReadings_];
        newAngles = new float[totalReadings_ - numReadings_];

        // aquire the lock for the readings and angles -- this could cause deadlock if 
        // something tries to aquire both of these locks in an order opposite of what 
        // is here, but I don't know how to work around that. Just make sure that if 
        // these two locks are aquired at the same time, it's in this order -- and 
        // it wouldn't hurt to repeat this comment wherever that happens
        pthread_mutex_lock(&readingsMutex_);

        // get the number of readings in the buffer in front of and behind this laser's
        readingsInFront = shReadings_ - laserList_[0]->shReadings_;

        // we'll get a negative number if we're killing what was the first laser in the list
        if (readingsInFront < 0)
        {
            readingsInFront = 0;
        }

        readingsBehind = totalReadings_ - readingsInFront - numReadings_;

        // copy readings and angles up to those belonging to this laser to the new buffer
        if (readingsInFront > 0)
        {
            memcpy(
                newReadings,
                oldReadings,
                readingsInFront * sizeof(oldReadings[0]));

            memcpy(
                newAngles,
                oldAngles,
                readingsInFront * sizeof(oldReadings[0]));
        }

        // copy readings and angles after those belonging to this laser to the new buffer
        if (readingsBehind > 0)
        {
            memcpy(
                newReadings + readingsInFront,
                oldReadings + readingsInFront + numReadings_,
                readingsBehind * sizeof(oldAngles[0]));

            memcpy(
                newAngles + readingsInFront, 
                oldAngles + readingsInFront + numReadings_, 
                readingsBehind * sizeof(oldAngles[0]));
        }

        // alter all laser objects so they point to the new buffers
        for (i = 0; i < (int)(laserList_.size()); i++)
        {
            // find out how far into the overall buffer the current laser's readings are
            offset = laserList_[i]->shReadings_ - oldReadings;

            // if the current laser's readings came after the readings of the laser being 
            // deleted, move it's pointer up in the buffer by the amount of readings from 
            // the laser being deleted
            if (laserList_[i]->shReadings_ > shReadings_)
            {
                offset -= numReadings_;
            }

            // place the pointer to the current laser's readings in the new buffer, using 
            // the beginning of the buffer as a reference
            laserList_[i]->shAngles_ = newAngles + offset;
            laserList_[i]->shReadings_ = newReadings + offset;
        }

        // alter the static laser data
        allReadings_ = newReadings;
        allAngles_ = newAngles;
        totalReadings_ -= numReadings_;

        // update the sensor to point to the new buffers
        if (sensorLaser_ != NULL)
        {
            sensorLaser_->setReadingInfo(allReadings_, totalReadings_);
            sensorLaser_->setAngleInfo(allAngles_, totalReadings_);
        }

        // release the lock for the readings and angles
        pthread_mutex_unlock(&readingsMutex_);
    }
    else
    {
        // alter the static laser data. It's OK to do so outside of the locks here since 
        // there are no other lasers reading or writing to the buffers, and new lasers 
        // are only created by the main thread.
        allReadings_  = NULL;
        allAngles_    = NULL;
        totalReadings_ = 0;
    }

    // get rid of the old buffers
    if (oldReadings != NULL)
    {
        delete [] oldReadings;
        oldReadings = NULL;
    }

    if (oldAngles != NULL)
    {
        delete [] oldAngles;
        oldAngles = NULL;
    }
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void GazeboLaser::updateStatusBar_(void)
{
    char buf[ 100 ];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[ 0 ] = 0;
    strcat(statusStr, "GazeboLaser [");

    sprintf(
        buf,
        "%s]: %c ",
        gazeboLaserName_.c_str(),
        statusbarSpinner_->getStatus());

    pthread_mutex_lock(&readingsMutex_);

    if ((shReadings_ != NULL) && (shAngles_ != NULL))
    {
        sprintf(
            buf,
            "%s Min Reading: %.2f m [%.1f deg]",
            buf,
            shReadings_[shMinReadingIndex_],
            shAngles_[shMinReadingIndex_]);
    }
    else
    {
        sprintf(buf, "%s Connecting...", buf);
    }

    pthread_mutex_unlock(&readingsMutex_);

    if (DISPLAY_FREQUENCY)
    {
        getFrequency_(frequency);
        sprintf(buf, "%s : %.2f Hz", buf, frequency);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function starts the reader thread.
//-----------------------------------------------------------------------
void *GazeboLaser::startReaderThread_(void *gazeboInstance)
{
    ((GazeboLaser*) gazeboInstance)->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to move GazeboLaser.
//-----------------------------------------------------------------------
void GazeboLaser::readerLoop_(void)
{
    int count = 0;

    readerThreadIsUp_ = true;
    
    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the data from the gazebo.
        readLaserData_();

        // Check for the pthread termination again.
        pthread_testcancel();

        // Compute the frequency.
        compFrequency_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (gLogManager != NULL)
        {
            manageLog_(readerThread_);
        }

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        nanosleep(&READER_SLEEPTIME_NSEC_, NULL);
    }
}

//-----------------------------------------------------------------------
// This function reads the laser data from gazebo.
//-----------------------------------------------------------------------
void GazeboLaser::readLaserData_(void)
{
    double currentGazeboReadings[GZ_LASER_MAX_RANGES];
    float minReading;
    int i, numReadings;

    memset(currentGazeboReadings, 0x0, sizeof(double)*GZ_LASER_MAX_RANGES);

    gz_laser_lock(gazeboLaser_, 1);

    memcpy(
        currentGazeboReadings,
        gazeboLaser_->data->ranges,
        sizeof(double)*GZ_LASER_MAX_RANGES);

    numReadings = gazeboLaser_->data->range_count;

    gz_laser_unlock(gazeboLaser_);

    if (numReadings != numReadings_)
    {
        printfTextWindow("Error(GazeboLaser): Invalid number of readings.\n");
        return;
    }

    pthread_mutex_lock(&readingsMutex_);

    shMinReadingIndex_ = 0;
    minReading = 0;

    for (i = 0; i < numReadings; i++)
    {
        // Copy the raw value.
        shRawReadings_[i] = (float)(currentGazeboReadings[i]);

        // Account for the offset.
        shReadings_[i] = shRawReadings_[i];
        shAngles_[i] = shRawAngles_[i];
        accountForXYOffset_(shReadings_[i], shAngles_[i]);

        // Compute the shortest reading.
        if (i == 0)
        {
            shMinReadingIndex_ = i;
            minReading = shReadings_[i];
        }
        else if (minReading > shReadings_[i])
        {
            shMinReadingIndex_ = i;
            minReading = shReadings_[i];
        }
    }

    pthread_mutex_unlock(&readingsMutex_);
}

//-----------------------------------------------------------------------
// This function sets up the user interface for the robot option.
//-----------------------------------------------------------------------
void GazeboLaser::control(void)
{
    bool done = false;
    int msgData[2];

    msgData[0] = 0; // "speed_factor" is obsolete.
    msgData[1] = 0; // "angular_speed_factor" is obsolete.

    messageDrawWindow(
        EnMessageType_GAZEBO_LASER_CONTROL,
        EnMessageErrType_NONE,
        msgData);

    do
    {
        int c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

//-----------------------------------------------------------------------
// This function logs the data.
//-----------------------------------------------------------------------
void GazeboLaser::logData_(double logTime)
{
    int i, numReadings;
    float *readings = NULL, *angles = NULL;

    if (logfile_ != NULL)
    {
        if (laserList_[0] == this)
        {
            numReadings = totalReadings_;

            fprintf(logfile_, "%f %d", logTime, numReadings);

            if (numReadings > 0)
            {
                // Copy the angles
                angles = new float[numReadings];

                pthread_mutex_lock(&readingsMutex_);
                memcpy(angles, allAngles_, sizeof(float)*numReadings); 

                // Copy the readings.
                readings = new float[numReadings];

                memcpy(readings, allReadings_, sizeof(float)*numReadings); 
                pthread_mutex_unlock(&readingsMutex_);

                for (i = 0; i < numReadings; i++)
                {
                    fprintf(logfile_, " %.1f %.2f", angles[i], readings[i]);
                }

                delete [] angles;
                angles = NULL;

                delete [] readings;
                readings = NULL;
            }

            fprintf(logfile_, "\n");
            fflush(logfile_);
        }
        else
        {
            fprintf(logfile_, "%f [See logfile for the other laser.]\n", logTime);
        }
    }
}

/**********************************************************************
# $Log: gazebo-laser.c,v $
# Revision 1.4  2009-01-27 20:23:09  zkira
# Added laser pitch angle parameter
#
# Revision 1.3  2008/11/20 02:16:20  ts2883
# Fixed a bug that resulted from adding the capability of having 4 lasers.
#
# Revision 1.2  2008/10/28 23:05:49  zkira
# Fixed PartLaser problem (sensor error in hclient)
#
# Revision 1.1  2008/07/15 22:33:48  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.11  2008/02/01 23:18:25  endo
# USARSimLaser added.
#
# Revision 1.10  2006/12/12 20:45:15  endo
# Regular update.
#
# Revision 1.9  2006/12/07 20:30:22  endo
# Regular backup.
#
# Revision 1.8  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.7  2006/11/20 22:55:51  endo
# OpenCV added.
#
# Revision 1.6  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.5  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.4  2006/11/14 01:04:01  endo
# Regular backup.
#
# Revision 1.3  2006/11/10 21:37:47  endo
# GazeboLaser bug fixed.
#
# Revision 1.2  2006/09/06 22:18:22  endo
# Regular update.
#
# Revision 1.1  2006/08/08 21:33:49  endo
# GazeboLaser added.
#
#**********************************************************************/
