/**********************************************************************
 **                                                                  **
 **                          gazebo-laser.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for Gazebo laser (Player/Stage)                       **
 **                                                                  **
 **  Copyright 2006, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gazebo-laser.h,v 1.4 2009-01-27 20:23:24 zkira Exp $ */

#ifndef GAZEBO_LASER_H
#define GAZEBO_LASER_H

#include <pthread.h>
#include <string>
#include <vector>

#include "module.h"
#include "gazebo.h"
#include "sensors.h"
#include "hserver.h"

using std::string;
using std::vector;

#define GAZEBO_LASER_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define GAZEBO_LASER_RAD2DEG(val) ((val)*180.0/M_PI)
#define GAZEBO_LASER_DEG2RAD(val) ((val)*M_PI/180.0)

class GazeboLaser : public Module {

protected:
    pthread_t readerThread_;
    gz_client_t *gazeboClient_;
    gz_laser_t *gazeboLaser_;
    extra_laser_t extraInfo_;
    string gazeboLaserName_;
    double GazeboLaserFOV_;
    double GazeboLaserRange_;

    float *shAngles_;
    float *shReadings_;
    float *shRawAngles_;
    float *shRawReadings_;
    float angleOffset_;
    float xOffset_;
    float yOffset_;
    float shPitchAngle_;
    int gazeboServerID_;
    int gazeboClientID_;
    int shMinReadingIndex_;
    int numReadings_;
    bool readerThreadIsUp_;

    static RobotPartLaser* m_pclPartLaser;
    static pthread_mutex_t readingsMutex_;
    static Sensor *sensorLaser_;
    static vector<GazeboLaser*> laserList_;
    static float *allAngles_;
    static float *allReadings_;
    static float *allRawReadings_;
    static int totalReadings_;

    static const struct timespec READER_SLEEPTIME_NSEC_;
    static const string DEFAULT_GAZEBO_LASER_NAME_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const int DEFAULT_GAZEBO_SERVER_ID_;
    static const int DEFAULT_GAZEBO_CLIENT_ID_;
    static const double DEFAULT_GAZEBO_LASER_FOV_;
    static const double DEFAULT_GAZEBO_LASER_RANGE_;

    void updateStatusBar_(void);
    void readerLoop_(void);
    void readLaserData_(void);
    void addSensor_(void);
    void addToOverallBuffer_(void);
    void removeFromOverallBuffer_(void);
    void logData_(double logTime);
    void accountForXYOffset_(float &reading, float &angle);
    bool initializeGazebo_(void);

    static void* startReaderThread_(void *gazeboInstance);

public:
    GazeboLaser(
        GazeboLaser **a,
        char *gazeboLaserName,
        float angleOffset, 
        float xOffset,
        float yOffset,
        float laserPitchAngle);
    ~GazeboLaser(void);
    void control(void);
};    

extern GazeboLaser *gGazeboLaser;

#endif

/**********************************************************************
# $Log: gazebo-laser.h,v $
# Revision 1.4  2009-01-27 20:23:24  zkira
# Added laser pitch angle parameter
#
# Revision 1.3  2008/10/29 00:12:17  ts2883
# Fixed bug: missing variables
#
# Revision 1.2  2008/10/28 23:05:49  zkira
# Fixed PartLaser problem (sensor error in hclient)
#
# Revision 1.1  2008/07/15 22:33:48  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.9  2008/02/01 23:18:25  endo
# USARSimLaser added.
#
# Revision 1.8  2006/12/07 20:30:22  endo
# Regular backup.
#
# Revision 1.7  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.6  2006/11/20 22:55:52  endo
# OpenCV added.
#
# Revision 1.5  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.4  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.3  2006/11/14 01:04:01  endo
# Regular backup.
#
# Revision 1.2  2006/09/06 22:18:22  endo
# Regular update.
#
# Revision 1.1  2006/08/08 21:33:49  endo
# GazeboLaser added.
#
#**********************************************************************/
