/**********************************************************************
 **                                                                  **
 **                          gazebo-robot.h                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for Gazebo robot (Player/Stage)                       **
 **                                                                  **
 **  Copyright 2006, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gazebo-robot.h,v 1.4 2008-09-24 00:29:40 ts2883 Exp $ */

#ifndef GAZEBO_ROBOT_H
#define GAZEBO_ROBOT_H

#include <pthread.h>
#include <string>

#include "HServerTypes.h"
#include "robot.h"
#include "sensors.h"
#include "robot_config.h"
#include "gazebo.h"

using std::string;

#define GAZEBO_ROBOT_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define GAZEBO_ROBOT_RAD2DEG(val) ((val)*180.0/M_PI)
#define GAZEBO_ROBOT_DEG2RAD(val) ((val)*M_PI/180.0)


class GazeboRobot : public Robot {

    typedef enum GazeboRobotClasses_t
    {
        GAZEBO_ROBOT_CLASS_HELICOPTER,
        GAZEBO_ROBOT_CLASS_ATRVJR,
        GAZEBO_ROBOT_CLASS_P2AT,
        GAZEBO_ROBOT_CLASS_P2DX,
        NUM_GAZEBO_ROBOT_CLASSES
    };
  
    typedef struct RobotSpec_t
    {
      int robotClass;
      string className;

      // These can be added similar to USARSim, if necessary
      //int numSonars;
      //float sonarAngles[GAZEBO_ROBOT_MAX_SONARS];
      //float sonarCoordinates[GAZEBO_ROBOT_MAX_SONARS*2];
      //float robotRadius;
      //float wheelRadius;
    };
  
protected:
    Sensor *sensorSonar_;
    Sensor *sensorVirtualBumper_;
    RobotPartXyt *partXyt_;
    RobotPartSonar *partSonar_;
    HSPose_t shGazeboPoseOffset_; // Shared
    HSPose_t shLatestRawPose_; // Shared
    pthread_t readerThread_;
    gz_client_t *gazeboClient_;
    gz_position_t *gazeboPosition_;
    gz_truth_t *gazeboTruth_;
    gz_sonar_t *gazeboSonar_;
    string gazeboRobotName_;
    string gazeboRobotClass_;
    float sonar_[GZ_SONAR_MAX_RANGES];
    float virtualBumper_[GZ_SONAR_MAX_RANGES];
    float sonarLoc_[GZ_SONAR_MAX_RANGES][3];
    float *sonarAngles_;
    float *sonarPos_;
    int gazeboServerID_;
    int gazeboClientID_;
    int numSonars_;
    bool readerThreadIsUp_;
    int robotSpecIndex_;

    static pthread_mutex_t gazeboPoseOffsetMutex_;
    static pthread_mutex_t latestRawPoseMutex_;

    static const RobotSpec_t GAZEBO_ROBOT_SPECS_[NUM_GAZEBO_ROBOT_CLASSES];

    void addSensors_(void);
    void updateStatusBar_(void);
    void readerLoop_(void);
    void controlRobot_(void);
    void readPositionData_(void);
    void readSonarData_(void);
    void setGazeboPoseOffset_(HSPose_t poseOffset);
    void getGazeboPoseOffset_(HSPose_t &poseOffset);
    void setLatestRawPose_(HSPose_t pose);
    void getLatestRawPose_(HSPose_t &pose);
    void updatePoseFromGazeboPoseOffset_(void);
    bool initializeGazebo_(void);

    static void* startReaderThread_(void *gazeboInstance);

public:

    static const string EMPTY_STRING_;
    static const string DEFAULT_GAZEBO_ROBOT_NAME_;
    static const float VIRTUAL_BUMPER_DISTANCE_;
    static const int MAIN_THREAD_USLEEP_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const string DEFAULT_GAZEBO_ROBOT_CLASS_;
    static const int DEFAULT_GAZEBO_SERVER_ID_;
    static const int DEFAULT_GAZEBO_CLIENT_ID_;
    static const bool REPORT_SONARS_;
    static const bool REPORT_VIRTUAL_BUMPER_;
    static const bool USE_GROUND_TRUTH_;

    GazeboRobot(
      Robot** a,
      const char *gazeboRobotName,
      const char *gazeboRobotClass,
      const int serverID,
      const int clientID);
    ~GazeboRobot(void);
    void control(void);
};    

#endif

/**********************************************************************
# $Log: gazebo-robot.h,v $
# Revision 1.4  2008-09-24 00:29:40  ts2883
# Added GroundTruth device
#
# Revision 1.3  2008/09/18 04:15:44  ts2883
# Fixed gazebo code to compile/run
# Still does not work with scarab, but works with multiple robots/hservers otherwise
#
# Revision 1.2  2008/09/16 16:50:41  zkira
# Added class name, client ID, and server ID as paramters/constructor
# Added advancePose() odometry simulation to helicopter
# Gazebo files are untested/uncompiled so far!
#
# Revision 1.1  2008/07/15 22:33:49  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.6  2006/12/12 20:45:15  endo
# Regular update.
#
# Revision 1.5  2006/11/16 18:20:26  endo
# The virtual bumper added.
#
# Revision 1.4  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.3  2006/11/14 01:04:01  endo
# Regular backup.
#
# Revision 1.2  2006/09/15 19:19:19  endo
# Regular backups.
#
# Revision 1.1  2006/07/31 22:01:14  endo
# Gazebo -> GazeboRobot
#
# Revision 1.1  2006/07/21 23:35:01  endo
# Gazebo integrated into HServer
#
#**********************************************************************/
