/**********************************************************************
 **                                                                  **
 **                         gazebo-video.c                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for Gazebo video (Player/Stage)                      **
 **                                                                  **
 **  Copyright 2006, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gazebo-video.c,v 1.1 2008/07/15 22:33:49 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "gazebo-video.h"
#include "Watchdog.h"
#include "LogManager.h"

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Statics
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const string GazeboVideo::DEFAULT_GAZEBO_VIDEO_NAME_ = "camera1";
const string GazeboVideo::DEFAULT_GAZEBO_TRUTH_NAME_ = "truth";
const int GazeboVideo::DEFAULT_GAZEBO_SERVER_ID_ = 0;
const int GazeboVideo::DEFAULT_GAZEBO_CLIENT_ID_ = 0;
const bool GazeboVideo::USE_GROUND_TRUTH_ = true;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
GazeboVideo::GazeboVideo(
    Video **a,
    char *gazeboVideoName,
    char *gazeboTruthName,
    bool showImage) :
    Video(a, HS_ROBOT_TYPE_GAZEBO, HS_MODULE_NAME_VIDEO_GAZEBO, showImage),
    gazeboClient_(NULL),
    gazeboVideo_(NULL),
    gazeboTruth_(NULL),
    lastTimeStamp_(0),
    gazeboServerID_(DEFAULT_GAZEBO_SERVER_ID_),
    gazeboClientID_(DEFAULT_GAZEBO_CLIENT_ID_)
{
    string displayName;
    int width, height;

    // Set up the laser name.
    gazeboVideoName_ = (gazeboVideoName != NULL)?
        gazeboVideoName : DEFAULT_GAZEBO_VIDEO_NAME_;

    // Set up the truth device name (for localization info).
    gazeboTruthName_ = (gazeboTruthName != NULL)?
        gazeboTruthName : DEFAULT_GAZEBO_TRUTH_NAME_;

    // Update the status bar.
    updateStatusBar_();

    // Initialize the gazebo.
    if (!initializeGazebo_())
    {
        printfTextWindow("Error: GazeboVideo could not be initialized. Exiting Gazebo.\n");
        delete this;
        a = NULL;
        return;
    }

    // Start the reader thread.
    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);

    // Start the watchdog.
    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(
            NAME_,
            WATCHDOG_CHECK_INTERVAL_SEC_,
            getpid(),
            readerThread_);
    }

    if (showImage_)
    {
        gz_camera_lock(gazeboVideo_, 1);
        width = gazeboVideo_->data->width;
        height = gazeboVideo_->data->height;
        gz_camera_unlock(gazeboVideo_);

        displayName = NAME_ + ": " + gazeboVideoName_;
        new Xwindow(&xwindow_, (char *)(displayName.c_str()), 0, 0, width, height);

        // Start the display thread.
        pthread_create(&displayThread_, NULL, &startDisplayThread_, (void*)this);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
GazeboVideo::~GazeboVideo(void)
{
    if (displayThreadIsUp_)
    {
        displayThreadIsUp_ = false;
        pthread_cancel(displayThread_);
        pthread_join(displayThread_, NULL);
    }

    if (readerThreadIsUp_)
    {
        readerThreadIsUp_ = false;
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    if (gazeboVideo_ != NULL)
    {
        gz_camera_lock(gazeboVideo_, 1);
        gz_camera_unlock(gazeboVideo_);
        gz_camera_close(gazeboVideo_);
        gz_camera_free(gazeboVideo_);
        gazeboVideo_ = NULL;
    }

    
    // Close the interface and free object for truth
    if (gazeboTruth_ != NULL)
    {
        gz_truth_close(gazeboTruth_);
        gz_truth_free(gazeboTruth_);
        gazeboTruth_ = NULL;
    }

    
    if (gazeboClient_ != NULL)
    {
        gz_client_disconnect(gazeboClient_);
        gz_client_free(gazeboClient_);
        gazeboClient_ = NULL;
    }

    printTextWindow("GazeboVideo disconnected");
}

//-----------------------------------------------------------------------
// This function initialize gazebo
//-----------------------------------------------------------------------
bool GazeboVideo::initializeGazebo_(void)
{
    // Print all errors
    gz_error_init(1, 9);

    // Create a client object
    gazeboClient_ = gz_client_alloc();  

    // Connect to the server
    if (gz_client_connect_wait(gazeboClient_, gazeboServerID_, gazeboClientID_) != 0)
    {
        printfTextWindow("Error: GazeboVideo could not connect to the client.\n");
        gz_client_free(gazeboClient_);
        gazeboClient_ = NULL;
        return false;
    }

    printfTextWindow("Video: %s Truth: %s\n", gazeboVideoName_.c_str(), gazeboTruthName_.c_str());

    // Create a position interface
    gazeboVideo_ = gz_camera_alloc();

    // Connect to device on server
    if (gz_camera_open(gazeboVideo_, gazeboClient_, gazeboVideoName_.c_str()) != 0)
    {
        printfTextWindow("Error: GazeboVideo could not open the video device (%s).\n", gazeboVideoName_.c_str());
        gz_camera_free(gazeboVideo_);
        gazeboVideo_ = NULL;
        return false;
    }

    // Create a truth interface (in order to be able to provide
    // a position for each frame)
    if (USE_GROUND_TRUTH_) {
      gazeboTruth_ = gz_truth_alloc();
      char name[10000];
      sprintf(name, "%s", gazeboTruthName_.c_str()); 
      printfTextWindow("Connecting to %s", name);
      if (gz_truth_open(gazeboTruth_, gazeboClient_, name ) != 0)
      {
        printfTextWindow("Warning: GazeboVideo could not open the ground truth device (named %s). \n", gazeboTruthName_.c_str());
        gz_truth_free(gazeboTruth_);
        gazeboTruth_ = NULL;
      }
    }
    
    return true;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void GazeboVideo::updateStatusBar_(void)
{
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;
    strcat(statusStr, "GazeboVideo [");

    if (frameRate_ > 0)
    {
        sprintf(
            buf,
            "%s]: %c %.2f frames/sec",
            gazeboVideoName_.c_str(),
            statusbarSpinner_->getStatus(),
            frameRate_);
    }
    else
    {
        sprintf(
            buf,
            "%s]: %c Connecting...",
            gazeboVideoName_.c_str(),
            statusbarSpinner_->getStatus());
    }

    if (DISPLAY_FREQUENCY)
    {
        getFrequency_(frequency);
        sprintf(buf, "%s : %.2f Hz", buf, frequency);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function reads the video data from gazebo.
//-----------------------------------------------------------------------
bool GazeboVideo::readVideoData_(void)
{
    double currentTime, timeStamp;
    int imageSize;

    if (gazeboVideo_ == NULL)
    {
        return false;
    }

    currentTime = getCurrentEpochTime();

    gz_camera_lock(gazeboVideo_, 1);

    timeStamp = gazeboVideo_->data->time;

    if (timeStamp != lastTimeStamp_)
    {
        imageSize = gazeboVideo_->data->image_size;

        pthread_mutex_lock(&currentImageMutex_);

        if (shCurrentImage_ != NULL)
        {
            if (shCurrentImageSize_ != imageSize)
            {
                delete [] shCurrentImage_;
                shCurrentImage_ = new unsigned char[imageSize];
            }
        }
        else
        {
            shCurrentImage_ = new unsigned char[imageSize];
        }

        shCurrentImageSize_ = imageSize;
        shCurrentImageWidth_ = gazeboVideo_->data->width;
        shCurrentImageHeight_ = gazeboVideo_->data->height;
        shCurrentImageTimeStamp_ = currentTime;

        bgr2rgb_(
            shCurrentImageWidth_,
            shCurrentImageHeight_,
            gazeboVideo_->data->image,
            &shCurrentImage_);

        pthread_mutex_unlock(&currentImageMutex_);
    }

    static int frame_counter = 0; 
    static int write_counter = 0;
    if (write_counter % 100 == 0) {
      // Get position for frame
      double x=0.0, y=0.0, z=0.0, roll=0.0, pitch=0.0, yaw=0.0;
      if (gazeboTruth_ != NULL) {
        x = gazeboTruth_->data->pos[0];
        y = gazeboTruth_->data->pos[1];
        z = gazeboTruth_->data->pos[2];

        double euler[3];
        gz_truth_euler_from_quatern(gazeboTruth_, euler, gazeboTruth_->data->rot);
        yaw = euler[2]; // - (M_PI/2.0);
        pitch = euler[1];
        roll = euler[0];
      }
      
      /* UNCOMMENT TO WRITE OUT IMAGES
      char filename[10000];
      sprintf(filename, "%s_frame_%d.ppm", gazeboTruthName_.c_str(), frame_counter++);
      FILE *out = fopen(filename, "w");
      if (out != NULL) {
        fprintf(out, "P6\n");
        fprintf(out, "# Loc: %lf,%lf,%lf,%lf,%lf,%lf\n", x, y, z, roll, pitch, yaw);
        fprintf(out, "%d %d\n255\n", shCurrentImageWidth_, shCurrentImageHeight_);
        fwrite(shCurrentImage_, sizeof(unsigned char), shCurrentImageWidth_*shCurrentImageHeight_*3,out); 
        fclose(out);
      }*/
    }
    write_counter++;

    gz_camera_unlock(gazeboVideo_);

    if (timeStamp == lastTimeStamp_)
    {
        // Image not updated yet.
        return true;
    }

    lastTimeStamp_ = timeStamp;

    if ((lastCapturedTime_ != 0) && (currentTime != lastCapturedTime_))
    {
        frameRate_ = 1.0 / (currentTime - lastCapturedTime_);
    }

    lastCapturedTime_ = currentTime;

    return true;
}

//-----------------------------------------------------------------------
// This function sets the channel.
//-----------------------------------------------------------------------
void GazeboVideo::setChannel(int channel)
{
    // Channel not supported in GazeboVideo.
}

//-----------------------------------------------------------------------
// This function sets up the user interface for the GazeboVideo.
//-----------------------------------------------------------------------
void GazeboVideo::control(void)
{
    bool done = false;
    int msgData[2];
    int c;

    msgData[0] = 0;
    msgData[1] = 0;

    messageDrawWindow(
        EnMessageType_GAZEBO_VIDEO_CONTROL,
        EnMessageErrType_NONE,
        msgData);

    do
    {
        c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

/**********************************************************************
# $Log: gazebo-video.c,v $
# Revision 1.1  2008/07/15 22:33:49  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.10  2008/01/21 19:14:15  endo
# Regular backup.
#
# Revision 1.9  2006/12/12 20:45:15  endo
# Regular update.
#
# Revision 1.8  2006/12/07 20:30:22  endo
# Regular backup.
#
# Revision 1.7  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.6  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.5  2006/11/22 22:24:28  endo
# OpenCV BlobTrack implemented.
#
# Revision 1.4  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.3  2006/11/20 22:55:52  endo
# OpenCV added.
#
# Revision 1.2  2006/11/20 16:43:37  endo
# Xwindow improved.
#
# Revision 1.1  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
#**********************************************************************/
