/**********************************************************************
 **                                                                  **
 **                             hserver.c                            **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Hardware Server                                                 **
 **                                                                  **
 **  Copyright 2000 - 2008, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: hserver.c,v 1.14 2010-05-03 22:10:08 zkira Exp $ */

#include <stdio.h>
#include <termio.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <getopt.h>
#include <curses.h>
#include <panel.h>
#include <ctype.h>
#include <semaphore.h>
#include <vector>
#include <string>

#include "hserver.h"
#include "camera.h"
#include "laser.h"
#include "terminal.h"
#include "message.h"
#include "cognachrome.h"
#include "ipt_handler.h"
#include "rcom_handler.h"
#include "RemoteControl.h"
#include "statusbar.h"
#include "video.h"
#include "frame_grabber.h"
#include "apm_module.h"
#include "robot_config.h"
#include "sensors.h"
#include "fred.h"
#ifndef NO_GAZEBO
#include "gazebo-robot.h"
#include "gazebo-laser.h"
#include "gazebo-video.h"
#endif
#include "pioneer.h"
#include "pioneer2.h"
#include "amigobot.h"
#include "report_level.h"
#include "robot.h"
#include "fred.h"
#include "gps.h"
#include "nomad.h"
#include "console.h"
#include "HSRCFile.h"
#include "HServerConfigurator.h"
#include "ControlInterfaces.h"
#include "PoseCalculatorInterface.h"
#include "PoseCalculator/PoseCalculator.h"
#include "cricket.h"
#include "jbox.h"
#include "compass.h"
#include "compass_3DM-G.h"
#include "gyro.h"
#include "Watchdog.h"
#include "LogManager.h"
#ifndef NO_OPENCV
#include "opencv_handler.h"
#endif
#include "usarsim-robot.h"
#include "usarsim-laser.h"
#include "usarsim-video.h"

#ifndef NO_ROVIO
#include "rovio-robot.h"
//#include "rovio-video.h"
#endif

#include "skybotix-robot.h"

#ifndef NO_MOBILITY
#include "urban.h"
#include "atrvjr.h"
#include "gyro_dmu-vgx.h"
#include "compass_kvh-c100.h"
#endif

// X11's Xlib.h (and may be others), which is called in xwindow.h
// defines "Status" as a macro. This name conflicts with the
// "Status" type defined in the CORBA namespace. To avoid the
// error, xwindow.h needs to be called after mobility/CORBA-related
// headers are called.
#include "xwindow.h"

const string DEFAULT_RC_FILE_NAME = ".hserverrc";
const string DEFAULT_POSECALC_RC_SECTION_NAME = "posecalc1";
const int DEFAULT_XWINDOW_X = 0;
const int DEFAULT_XWINDOW_Y = 0;
const int DEFAULT_XWINDOW_WIDTH = 640;
const int DEFAULT_XWINDOW_HEIGHT = 480;

struct SuOption
{
    ClHServerConfig::EnHServerConfigSectionType enSectionType;
    string strSectionName;

    SuOption() {}
    SuOption(const SuOption& suOriginal) { *this = suOriginal; }
    SuOption& operator=(const SuOption& suOriginal);
};

SuOption& SuOption::operator=(const SuOption& suOriginal)
{
    if (this != &suOriginal)
    {
        enSectionType = suOriginal.enSectionType;
        strSectionName = suOriginal.strSectionName;
    }

    return *this;
}

bool gWatchdogEnabled = true;

bool useApm;
bool useUrban;
bool useATRVJr;
bool useIpt;
bool useRCom;
bool useHclient;
bool useX, useFred;
bool multipleHservers;
bool noSplash = true;
bool useJbox = false;
bool useGps = false;
bool useCricket = false;
bool logData = false;
bool jboxDisableNetworkQuery = false;
int jboxID = 22;
bool UsingRovio = false;
bool UsingSkybotix = false;

//int gNumGazeboModules = 0;

// lists to hold initialization info
vector<SuLaserInitData> g_clLaserInitList;
vector<SuRobotInitData> g_clRobotInitList;
vector<SuGazeboInitData> g_clGazeboInitList;
vector<SuUSARSimInitData> g_clUSARSimInitList;
vector<SuRovioInitData> g_clRovioInitList;
vector<SuSkybotixInitData> g_clSkybotixInitList;

vector<SuOpenCVInitData> g_clOpenCVInitList;
vector<SuCameraInitData> g_clCameraInitList;
vector<SuCognachromeInitData> g_clCognachromeInitList;
vector<SuNomadInitData> g_clNomadInitList;
vector<SuPioneerInitData> g_clPioneerInitList;
vector<SuPioneer2InitData> g_clPioneer2InitList;
vector<SuAmigoBotInitData> g_clAmigoBotInitList;
vector<SuFrameGrabberInitData> g_clFrameGrabberInitList;
vector<SuGpsInitData> g_clGpsInitList;
vector<SuCricketInitData> g_clCricketInitList;
vector<SuJboxInitData> g_clJboxInitList;
vector<SuCompassInitData> g_clCompassInitList;
vector<SuGyroInitData> g_clGyroInitList;
vector<SuPoseCalcInitData> g_clPoseCalcInitList;

double gps_base_lat, gps_base_lon, gps_base_x, gps_base_y, gps_got_data, gps_mperlon, gps_mperlat;

char* ipt_server_name = NULL;

WINDOW *mainWindow = NULL;
WINDOW *backWindow,*textWindow,*helpWindow;

PANEL *backPanel,*textPanel,*helpPanel;

int statusbarDevideLine;

int mainPanelX,mainPanelY;
int screenX,screenY;

pthread_mutex_t screenMutex;	// mutex for all curses functions

char robot_name[100] = "fred";

pthread_t signal_handler;
sem_t exit_sem;

bool text_paused = false;
bool SilentMode = false;

PoseCalcInterface *gPoseCalc = NULL;
Robot *gRobot = NULL;
Gps *gps = NULL;
Cricket *cricket = NULL;
Jbox *jbox = NULL;
Compass *gCompass = NULL;
Gyro *gGyro = NULL;
RemoteControl *gRemoteControl = NULL;
LogManager *gLogManager = NULL;
Apm *apm = NULL;

// This function returns the current time (since Epoch) in second.
double getCurrentEpochTime(void)
{
    timeval tvp;
    double currentEpochTime;
    
    gettimeofday(&tvp, NULL);
    currentEpochTime = (double)tvp.tv_sec + (((double)tvp.tv_usec)/1000000.0);

    return currentEpochTime;
}


HSDateString_t getCurrentDateString(void)
{
    HSDateString_t dateString;
    time_t cur_time;
    tm local_time;
    string prefix;
    char date[100], month[100], day[100], year[100], hour[100], min[100], sec[100];

    // Get the current date and time.
    cur_time = time(NULL);
    strncpy(date,ctime(&cur_time),sizeof(date));
    local_time = *localtime(&cur_time);
    strftime(month, sizeof(month), "%m", &local_time);
    strftime(day, sizeof(day), "%d", &local_time);
    strftime(year, sizeof(year), "%Y", &local_time);
    strftime(hour, sizeof(hour), "%H", &local_time);
    strftime(min, sizeof(min), "%M", &local_time);
    strftime(sec, sizeof(sec), "%S", &local_time);

    dateString.year = year;
    dateString.month = month;
    dateString.day = day;
    dateString.hour = hour;
    dateString.min = min;
    dateString.sec = sec;

    return dateString;
}

// if the last character in the string is a '/', get rid of it
string TruncateSlash(string str)
{
    if ((str != "") && (str[str.length() - 1] == '/'))
    {
        str.erase(str.length() - 1, str.length() - 1);
    }

    return str;
}

// extract the list of colon seperated directories from an environment variable
vector<string> ExtractDirectories(string strName)
{
    vector<string> clDirList;

    // Get a copy of the environment variable
    const char* szPathList = getenv(strName.c_str());
    string strPathList = (szPathList != NULL) ? szPathList : "";

    unsigned int i;
    string strDirectory = "";
    char c;

    // iterate through the characters in the path list
    for (i = 0; i < strPathList.length(); i++)
    {
        c = strPathList[i];

        // add non-colon characters to the current path directory
        if (c != ':')
        {
            strDirectory += c;
        }
        // if we've hit a colon, we've completed a directory
        else
        {
            if ((strDirectory = TruncateSlash(strDirectory)) != "")
            {
                // add the directory to the list and reset our local variable
                clDirList.push_back(strDirectory);
                strDirectory = "";
            }
        }
    }

    // if the last character wasn't a colon, then we never added the last 
    // directory in the path list to our list. do this now.
    if (c != ':')
    {
        if ((strDirectory = TruncateSlash(strDirectory)) != "")
        {
            clDirList.push_back(strDirectory);
        }
    }

    return clDirList;
}

void redrawWindows(void)
{
    pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);
    update_panels();
    doupdate();
    pthread_cleanup_pop(1);
}

void refreshScreen(void)
{
    pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);
    clear();
    pthread_cleanup_pop(1);
    redrawWindows();
}    

/* Print a formatted text string to the text window */
void printfTextWindow(const char* szFormat, ...)
{
    va_list args;
    char szMessage[512];

    va_start(args, szFormat);
    vsprintf(szMessage, szFormat, args);
    va_end(args);

    printTextWindow(szMessage);
}

/* Print string to text window. Filter will remove all non-alphanumerics.
   nl will scroll window. clear will clear window and return. */
void printTextWindow(char *str, int nl, int filter, int clear)
{
    static int y = 0;
    static int x = 0;
    static int firsttime = 0;
    static int ff = false;
    int i, len;
    char buf[HS_BUF_SIZE];
    
    if (text_paused || SilentMode)
    {
        return;
    }

    if (clear)
    {
        y = 0;
        x = 0;
        firsttime = 0;
        ff = false;
        werase(textWindow);
        redrawWindows();
        return;
    }

    if (filter)
    {
        len = strlen(str);
        for (i = 0; i < len; i++)
        {
            if (isalnum(str[i]))
            {
                buf[i] = str[i];
            }
            else
            {
                buf[i] = '^';
            }
        }
        buf[len] = 0;
    }
  
    pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);
    if (ff)
    {
        scroll(textWindow);
        ff = false;
    }

    if ((y == (screenY - 14)) && firsttime && (x == 0))
    {
        scroll(textWindow);
        firsttime = true;
    }
    wmove(textWindow, y, x);
    if (!filter)
    {
        waddnstr(textWindow, str, screenX - x - 1);
    }
    else
    {
        waddnstr(textWindow, buf, screenX - x - 1);
    }
    if (nl)
    {
        if (y < (screenY - 13))
        {
            y++;
        }
        else
        {
            ff = true;
        }
        x = 0;
    }
    else
    {
        getyx(textWindow, y, x);
    }
    pthread_cleanup_pop(1);
    redrawWindows();
}

/* Initialize serial port */
int setupSerial(
    int* fd,
    const char* port,
    long baud,
    int report_level,
    char* caller,
    int timeout,
    int nonblock)
{
    struct termios oldtp;
    struct termios tp;

    if ((*fd = open(port, O_RDWR | O_NOCTTY)) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printfTextWindow("%s Error opening serial port %s", caller, port);
        }
        return -1;
    }
    if (tcgetattr(*fd, &oldtp) < 0)
    {
        close(*fd);
        if (report_level >= HS_REP_LEV_ERR)
        {
            printfTextWindow("%s Error Couldn't get term attributes.", caller);
        }
        return -1;
    }
    bzero(&tp, sizeof(tp)); 
    tp.c_cflag = CS8 | CLOCAL | CREAD | baud;
    tp.c_oflag = 0;
    tp.c_iflag = IGNBRK | IGNPAR;
    tp.c_lflag = 0;
    tp.c_cc[VTIME] = timeout;
    tp.c_cc[VMIN] = 0;
    if (nonblock)
    {
        fcntl(*fd, F_SETFL ,O_NONBLOCK);
    }
    tcflush(*fd, TCIFLUSH);
    if (tcsetattr(*fd, TCSANOW, &tp) < 0)
    {
        if (report_level >= HS_REP_LEV_ERR)
        {
            printfTextWindow("%s Error Couldn't set term attributes.", caller);
        }
        return -1;
    }
    if (report_level >= HS_REP_LEV_DEBUG)
    {
        printfTextWindow("%s serial opened on %s", caller, port);
    }
    return 0;
}

/* Shutdown curses */
void closeCurses()
{
  curs_set(1);
  endwin();
}

void drawHelpWindow(void)
{
    int line = 1;
    pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);
    wmove(helpWindow, 0, 0);
    wborder(helpWindow, 0, 0, 0, 0, 0, 0, 0, 0);
    mvwaddstr(helpWindow, line++, 1, "HServer Hardware Control");
    mvwaddstr(helpWindow, line++, 1, "h help");
    mvwaddstr(helpWindow, line++, 1, "r refresh screen");
    mvwaddstr(helpWindow, line++, 1, "c clear text window");
    mvwaddstr(helpWindow, line++, 1, "p stop text window");
    mvwaddstr(helpWindow, line++, 1, "Q quit");
    mvwaddstr(helpWindow, line++, 1, "R robot");
    mvwaddstr(helpWindow, line++, 1, "C camera");
    mvwaddstr(helpWindow, line++, 1, "E sensors");
    mvwaddstr(helpWindow, line++, 1, "B battery usage");
    mvwaddstr(helpWindow, line++, 1, "L laser");
    mvwaddstr(helpWindow, line++, 1, "M cognachrome");
    mvwaddstr(helpWindow, line++, 1, "F remote control");
    mvwaddstr(helpWindow, line++, 1, "T terminal");
    mvwaddstr(helpWindow, line++, 1, "V video");
    mvwaddstr(helpWindow, line++, 1, "X xwindow");    
    mvwaddstr(helpWindow, line++, 1, "I ipt");
    mvwaddstr(helpWindow, line++, 1, "D report level");
    mvwaddstr(helpWindow, line++, 1, "W robot config");
    mvwaddstr(helpWindow, line++, 1, "G gps");
    mvwaddstr(helpWindow, line++, 1, "Y gyroscope");
    mvwaddstr(helpWindow, line++, 1, "K compass");
    mvwaddstr(helpWindow, line++, 1, "J Jbox");
    pthread_cleanup_pop(1);
}


/* Draw borders and lines */
void drawBackWindow(void)
{
    char buf[100];
    pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);
    wclear(backWindow);
    wmove(backWindow, 0, 0);
    waddch(backWindow, ACS_ULCORNER);
    whline(backWindow, ACS_HLINE, screenX - 2);
    wmove(backWindow, 0, screenX - 1);
    waddch(backWindow, ACS_URCORNER);
    wmove(backWindow, 1, 0);
    wvline(backWindow, ACS_VLINE, screenY - 2);
    wmove(backWindow, 1, screenX - 1);
    wvline(backWindow, ACS_VLINE, screenY - 2);
    wmove(backWindow, screenY - 1, 0);
    waddch(backWindow, ACS_LLCORNER);
    whline(backWindow, ACS_HLINE, screenX - 2);
    wmove(backWindow, screenY - 1, screenX - 1);
    waddch(backWindow, ACS_LRCORNER);
    wmove(backWindow, statusbarDevideLine, 0);
    waddch(backWindow, ACS_LTEE);
    whline(backWindow, ACS_HLINE, screenX - 2);
    wmove(backWindow, statusbarDevideLine, screenX - 1);
    waddch(backWindow, ACS_RTEE);
    wmove(backWindow, 1, 1);
    wattron(backWindow, A_UNDERLINE);
    sprintf(buf, "HServer %s", robot_name);
    waddstr(backWindow, buf);  
    wattroff(backWindow, A_UNDERLINE);
    pthread_mutex_unlock(&screenMutex);
    pthread_cleanup_pop(1);
}

void initializeCurses(void)
{
    statusbarDevideLine = 4;

    if (mainWindow == NULL)
    {
        mainWindow = (WINDOW*)malloc (sizeof (WINDOW));
    }
    mainWindow = initscr();	// initialize the curses library
    nonl();			        // tell curses not to do NL->CR/NL on output
    cbreak();			    // take input chars one at a time, no wait for \n
    noecho();			    // don't echo input
    keypad(stdscr, TRUE);

    if (has_colors())
    {
        start_color();
        init_pair(COLOR_BLACK, COLOR_BLACK, COLOR_BLACK);
        init_pair(COLOR_GREEN, COLOR_GREEN, COLOR_BLACK);
        init_pair(COLOR_RED, COLOR_RED, COLOR_BLACK);
        init_pair(COLOR_CYAN, COLOR_CYAN, COLOR_BLACK);
        init_pair(COLOR_WHITE, COLOR_WHITE, COLOR_BLACK);
        init_pair(COLOR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
        init_pair(COLOR_BLUE, COLOR_BLUE, COLOR_BLACK);
        init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLACK);
    }
  
    getmaxyx(mainWindow, screenY, screenX);
    backWindow = newwin(screenY, screenX, 0, 0);
    backPanel = new_panel(backWindow);
    drawBackWindow();
    show_panel(backPanel);
    mainPanelX = screenX;
    mainPanelY = screenY;

    textWindow = newwin(
        screenY - statusbarDevideLine - 2,
        screenX - 2,
        statusbarDevideLine + 1,
        1);
    textPanel = new_panel(textWindow);
    scrollok(textWindow, TRUE);
    show_panel(textPanel);

    helpWindow = newwin(HS_HELP_ROW, HS_HELP_COL, HS_HELP_Y, HS_HELP_X);
    helpPanel = new_panel(helpWindow);
    drawHelpWindow();
    hide_panel(helpPanel);

    statusbar = new Statusbar(statusbarDevideLine - 3);

    terminalInitialize();
    messageInitialize();

    displayTerm = false;
    curs_set(0);
}

void exit_hserver(void)
{
    refreshScreen();
    printfTextWindow("Exiting HSerever...");
    sleep(1);
    SilentMode = true;
    sleep(1);

    // delete all lasers
    while (Laser::m_clLaserList.size() > 0)
    {
        if (Laser::m_clLaserList[0] != NULL)
        {
            // just delete the laser -- the laser being deleted
            // will remove itself from the list
            delete Laser::m_clLaserList[0];
        }
        else
        {
            Laser::m_clLaserList.erase(Laser::m_clLaserList.begin());
        }
    }

    if (gLogManager != NULL)
    {
        delete gLogManager;
        gLogManager = NULL;
    }

    if (gRemoteControl != NULL)
    {
        delete gRemoteControl;
        gRemoteControl = NULL;
    }

    if (gGyro != NULL)
    {
        delete gGyro;
    }

    if (camera != NULL)
    {
        delete camera;
    }
    if (cognachrome != NULL)
    {
        delete cognachrome;
    }
    if (gps != NULL)
    {
        delete gps;
    }

    if (iptHandler != NULL)
    {
        delete iptHandler;
    }

    if (gRComHandler != NULL)
    {
        delete gRComHandler;
    }

    if (jbox != NULL)
    {
        delete jbox;
    }

    if (gVideo != NULL)
    {
        delete gVideo;
        gVideo = NULL;
    }

    if (apm != NULL)
    {
        delete apm;
    }

    if (gRobot != NULL)
    {
        delete gRobot;
    }

    if (gCompass != NULL)
    {
        delete gCompass;
    }

    if (gXwindow != NULL)
    {
        delete gXwindow;
    }

    if (gPoseCalc != NULL)
    {
        delete gPoseCalc;
        gPoseCalc = NULL;
    }

    closeCurses();
    sem_post(&exit_sem);
    printf("\n");
}

void perform_laser_action()
{
    char cUserChoice;

    if (Laser::m_clLaserList.size() == 0)
    {
        Laser::ConnectLaser();
    }
    else
    {
        cUserChoice = messageGetChar(EnMessageFilterType_LASER_USER_ACTION,
                                      EnMessageType_LASER_USER_ACTION,
                                      EnMessageErrType_GENERAL_INVALID);
        if (cUserChoice != 'x')
        {
            if (cUserChoice == '1')
            {
                Laser::ConnectLaser();
            }
            else
            {
                Laser* pLaser = Laser::GetLaserByName();
                if (pLaser != NULL)
                {
                    pLaser->control();
                }
            }
        }
    }
}

// **** KEYBOARD READER
void keyboard_reader_thread(void)
{
    int c = 0;
    bool displayHelp = false;
    while(true)
    {
        pthread_testcancel();
        c = getch();
        switch (c) {

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        case 'B':
            if (apm != NULL)
            {
                apm->control();
            }
            else
            {
                new Apm(&apm);
            }
            break;

        case 'c':
            printTextWindow(NULL, false, false, true);
            break;

        case 'C':
            if (camera != NULL)
            {
                camera->control();
            }
            else
            {
                cameraConnect();
            }
            break;

        case 'D':
            reportlevel->show();
            break;

        case 'E':
            sensors->show();
            break;

        case 'F':
            if (gRemoteControl != NULL)
            {
                gRemoteControl->runRemoteControl();
            }
            break;

        case 'G':
            if (gps != NULL)
            {
                controlGps(gps);
            }
            else
            {
                gpsConnect(
                    gps,
                    gps_base_lat,
                    gps_base_lon,
                    gps_base_x,
                    gps_base_y, 
                    gps_mperlat,
                    gps_mperlon);
            }
            break;

        case 'h':
            displayHelp = !displayHelp;
            if (displayHelp)
            {
                show_panel(helpPanel);
            }
            else
            {
                hide_panel(helpPanel);
            }
            redrawWindows();
            break;

        case 'I':
            if (iptHandler != NULL)
            {
                iptHandler->control();
            }
            else
            {

                new IptHandler(
                    &iptHandler,
                    ipt_server_name, 
                    robot_name,
                    &gps,
                    multipleHservers);

                sleep(1);

                if (useHclient)
                {
                    iptHandler->listen_hclient();
                }
            }
            break;

        case 'J':
            if (jbox != NULL)
            {
                controlJbox(jbox);
            }
            else
            {
                Jbox::connectToJbox(jbox);
            }
            break;

        case 'K':
            if (gCompass != NULL)
            {
                gCompass->control();
            }
            else
            {
                compassConnect(gCompass);
            }
            break;

        case 'L':
            perform_laser_action();
            break;

        case 'M':
            if (cognachrome != NULL)
            {
                cognachrome->control();
            }
            else
            {
                cognachromeConnect();
            }
            break;

        case 'O':
#ifndef NO_OPENCV
            if (gOpenCVHandler != NULL)
            {
                gOpenCVHandler->control();
            }
            else
            {
                // opencvConnect() to be implemented.
            }
#endif
            break;

        case 'p':
            text_paused = !text_paused;
            break;

        case 'P':
            if (gPoseCalc != NULL)
            {
                gPoseCalc->control();
            }
            else
            {
                printfTextWindow("Error: PoseCalc not running.\n");
            }
            break;

        case 'Q':
            exit_hserver();
            break;

        case 'r':
            refreshScreen();
            break;

        case 'S':
            SilentMode = !SilentMode;
            break;

        case 'T':
            if (terminalConnected)
            {
                terminalShow();
            }
            else
            {
                terminalReconnect();
            }
            break;

        case 'V':
            if (gVideo != NULL)
            {
                gVideo->control();
            }
            else
            {
                videoConnect(gVideo);
            }
            break;

        case 'X':
            if (gXwindow != NULL)
            {
                gXwindow->control();
            }
            else
            {
                new Xwindow(
                    &gXwindow,
                    NULL,
                    DEFAULT_XWINDOW_X,
                    DEFAULT_XWINDOW_Y,
                    DEFAULT_XWINDOW_WIDTH,
                    DEFAULT_XWINDOW_HEIGHT);
            }
            break;

        case 'W':
            robot_config->show();
            break;

        case 'R':
            if (gRobot != NULL)
            {
                gRobot->control();
            }
            else
            {
                robotConnect(gRobot);
            }
            break;

        case 'Y':
            if (gGyro != NULL)
            {
                gGyro->control();
            }
            else
            {
                gyroConnect(gGyro);
            }
            break;

        case KEY_ESC:
        case 'x':
            messageHide();
            break;
        }
    }
}


void resizeWindows(bool erase)
{
    struct winsize win;
    
    if ((ioctl(1, TIOCGWINSZ, &win) >= 0) ||       // 1 is stdout
         (ioctl(0, TIOCGWINSZ, &win) >= 0))        // 0 is stdin
    {
        screenX = win.ws_col;
        screenY = win.ws_row;
        mainPanelX = screenX;
        mainPanelY = screenY;
        pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &screenMutex);
        pthread_mutex_lock(&screenMutex);
        resizeterm(screenY, screenX);
        mvwin(textWindow, statusbarDevideLine + 1, 1);
        wresize(textWindow, screenY - statusbarDevideLine - 2, screenX - 2);
        statusbar->resize();
        messageMove();
        pthread_cleanup_pop(1);
        drawBackWindow();
        if (erase)
        {
            printTextWindow(NULL, true, false, true);
        }
    }
    if (erase)
    {
        printfTextWindow("Window resized %dx%d", win.ws_row, win.ws_col);
    }
}

char* signalName(int signal)
{
    static char name[10];

    switch (signal) {
    case 1: return "SIGHUP";
    case 2: return "SIGINT";
    case 3: return "SIGQUIT";
    case 4: return "SIGILL";
    case 6: return "SIGABRT";
    case 8: return "SIGFPE";
    case 9: return "SIGKILL";
    case 11: return "SIGSEGV";
    case 13: return "SIGPIPE";
    case 14: return "SIGALRM";
    case 15: return "SIGTERM";
    case 10: return "SIGUSR1";
    case 12: return "SIGUSR2";
    case 17: return "SIGCHLD";
    case 18: return "SIGCONT";
    case 19: return "SIGSTOP";
    case 20: return "SIGTSTP";
    case 21: return "SIGTTIN";
    case 22: return "SIGTTOU";
    case 23: return "SIGURG";
    case 24: return "SIGXCPU";
    case 25: return "SIGXFSZ";
    case 26: return "SIGVTALRM";
    case 27: return "SIGPROF";
    case 28: return "SIGWINCH";
    case 29: return "SIGPOLL";
    case 30: return "SIGPWR";
    case 31: return "SIGSYS";
    default:
        sprintf(name, "signal %d", signal);
        return name;
    }
}

/* Thread to wait for signals. All other threads ignore all signals */
void signal_handler_thread(void)
{
    sigset_t sigset;
    int signal;

    sigfillset(&sigset);
    while(true)
    {
        sigwait(&sigset, &signal);
        switch(signal)
        {
        case SIGINT:
            exit_hserver();
            break;
        case SIGWINCH:
            resizeWindows();
            break;
        default:
            printfTextWindow("Signal handler recieved %s", signalName(signal));
        }
    }
}

int startHServer(void)
{
    PoseCalcInterface *poseCalc = NULL;
    PoseCalcConstants_t poseCalcConstants;
    vector<unsigned int> usarsimInitListLaserIndexes;
    pthread_t keyboard_reader;
    sigset_t sigset;
    int poseCalcID;
    unsigned int i, index;

    pthread_mutex_init(&screenMutex, NULL);
    
    sem_init(&exit_sem, 0, 0);
    
    initializeCurses();
    if (!noSplash)
    {
        messageShow();
    }
    sensors = new Sensors(10);
    robot_config = new RobotConfiguration(10, 1, robot_name);

    reportlevel = new ReportLevel;

    printfTextWindow("HServer Version %s", HSERVER_VERSION);
    printTextWindow("Generic Robot Contoller");
    printTextWindow("");
    printTextWindow("Press h for help");
    
    pthread_create(&signal_handler, NULL, (void * (*)(void *)) &signal_handler_thread, NULL);
    sigfillset(&sigset);
    pthread_sigmask(SIG_BLOCK, &sigset, NULL);

    printTextWindow("I know how to connect to:");
    printTextWindow("   Pioneer AT robot via serial");
    printTextWindow("   Pioneer2 DXE robot via serial");
    printTextWindow("   AmigoBot robot via serial");
    printTextWindow("   Sony evi-d30 camera via serial");
    printTextWindow("   SICK LMS200 laser scanner via serial");
    printTextWindow("   Missionlab console via tcp [ipt]");
    printTextWindow("   iRobot Urban robot via tcp [mobility]");
    printTextWindow("   iRobot ATRV-Jr robot via tcp [mobility]");
    printTextWindow("   Nomadic technologies Nomad 150/200 robot via serial/tcp");
    printTextWindow("   Congnachrome vision system via serial");
    printTextWindow("   Bt848 Video Framegrabber via /dev/video");
    printTextWindow("   X11 server via local or tcp");
    printTextWindow("   USC Gazebo");
    printTextWindow("   NIST USARSim");
    printTextWindow("   WowWee Rovio");
    printTextWindow("   Skybotix UAV");
    printTextWindow("   Intel OpenCV");
    printTextWindow("");
    printfTextWindow("HServer name is %s", robot_name);

    if (logData)
    {
        gLogManager = new LogManager();
    }

    if (useIpt)
    {

        new IptHandler(
            &iptHandler,
            ipt_server_name,
            robot_name,
            &gps, 
            multipleHservers);

        sleep(2);

        // TODO: This is where hserver hangs with "bytes 0" showing but not console: [down].
        // Not sure how it could be hanging on a sleep, but it is.  Changed sleep(2) to sleep(3) 
        // to try to alleviate the problem, but this should be looked into
        sleep(3);

        printfTextWindow("HSERVER HANGING DEBUG: Created new handler, sleeping for a bit");
        if (useHclient)
        {
            iptHandler->listen_hclient();
            sleep(2);
        }
        printfTextWindow("HSERVER HANGING DEBUG: Done with everything");
       
        // TODO REMOVE: This file is outputted b/c hserver hangs... bug should be fixed
        char dfile[10000];
        sprintf(dfile, "%s_conf", robot_name);
        FILE *dout = fopen(dfile, "w");
        fprintf(dout, "Start confirmed!\n");
        fclose(dout);
    }

    if (useRCom)
    {

        new RComHandler(&gRComHandler);
        sleep(2);
    }

    if (g_clCognachromeInitList.size() > 0)
    {
        new Cognachrome(&cognachrome, g_clCognachromeInitList[0].strPortString);
        sleep(4);
    }

    if (g_clCameraInitList.size() > 0)
    {
        new Camera(&camera, g_clCameraInitList[0].strPortString);
        sleep(2);
    }

    if (useATRVJr)
    {
#ifndef NO_MOBILITY

        new ATRVJr(&gRobot);

        sleep(2);
#else 
        printTextWindow("No Mobility support");
#endif
    }

    bool bUseStream = false;
    for (i = 0; i < g_clLaserInitList.size(); i++)
    {
        Laser::EnConnectType enConnectType = Laser::EnConnectType_DIRECT;
        if (g_clLaserInitList[i].bListen)
        {
            enConnectType = Laser::EnConnectType_LISTEN_IPT;
        }
        else if (g_clLaserInitList[i].bSend)
        {
            enConnectType = Laser::EnConnectType_DIRECT_AND_STREAM_IPT;
        }

        new Laser(
            enConnectType,
            g_clLaserInitList[i].strPort, 
            g_clLaserInitList[i].strName,
            g_clLaserInitList[i].fAngleOffset, 
            g_clLaserInitList[i].fXOffset,
            g_clLaserInitList[i].fYOffset);


        if (g_clLaserInitList[i].strStreamAddress != "")
        {
            bUseStream = true;
            Laser::SetStreamAddress(g_clLaserInitList[i].strStreamAddress);
        }

        sleep(2);
    }

    if ((g_clGazeboInitList.size()) > 0)
    {
#ifndef NO_GAZEBO
        for (i = 0; i < g_clGazeboInitList.size(); i++)
        {
            switch (g_clGazeboInitList[i].type) {

            case HS_GAZEBO_TYPE_ROBOT:
                if (gRobot == NULL)
                {
                  new GazeboRobot(
                    &gRobot,
                    (char *)(g_clGazeboInitList[i].name.c_str()),
                    (char *)(g_clGazeboInitList[i].robot_class.c_str()),
                    g_clGazeboInitList[i].clientID,
                    g_clGazeboInitList[i].serverID);
                }
                else
                {
                    printfTextWindow("HServer cannot connect to more than one robot module.");
                }
                break;

            case HS_GAZEBO_TYPE_LASER:
                new GazeboLaser(
                    &gGazeboLaser,
                    (char *)(g_clGazeboInitList[i].name.c_str()),
                    g_clGazeboInitList[i].angleOffset, 
                    g_clGazeboInitList[i].xOffset,
                    g_clGazeboInitList[i].yOffset,
                    g_clGazeboInitList[i].laserPitchAngle);
                break;

            case HS_GAZEBO_TYPE_VIDEO:
                printfTextWindow("HS Video: %s Truth %s\n", (char *)(g_clGazeboInitList[i].name.c_str()), (char *)(g_clGazeboInitList[i].truthName.c_str()));
                new GazeboVideo(
                    &gVideo,
                    (char *)(g_clGazeboInitList[i].name.c_str()),
                    (char *)(g_clGazeboInitList[i].truthName.c_str()),
                    g_clGazeboInitList[i].showImage); 
                break;

            default:
                printfTextWindow(
                    "Unknown Gazebo type '%d'.",
                    g_clGazeboInitList[i].type);
                break;
            }

            sleep(2);
        }
#else
        printTextWindow("Gazebo disabled.");
#endif
    }

    for (i = 0; i < g_clUSARSimInitList.size(); i++)
    {
        switch (g_clUSARSimInitList[i].type) {

        case HS_USARSIM_TYPE_ROBOT:
            if (gRobot == NULL)
            {
                new USARSimRobot(
                    &gRobot,
                    (char *)(g_clUSARSimInitList[i].host.c_str()),
                    (char *)(g_clUSARSimInitList[i].name.c_str()),
                    (char *)(g_clUSARSimInitList[i].robot_class.c_str()),
                    g_clUSARSimInitList[i].port,
                    g_clUSARSimInitList[i].pos_x,
                    g_clUSARSimInitList[i].pos_y,
                    g_clUSARSimInitList[i].pos_z,
                    g_clUSARSimInitList[i].rot_yaw,
                    g_clUSARSimInitList[i].rot_pitch,
                    g_clUSARSimInitList[i].rot_roll);
            }
            else
            {
                printfTextWindow("HServer cannot connect to more than one robot module.");
            }
            break;

        case HS_USARSIM_TYPE_LASER:
            usarsimInitListLaserIndexes.push_back(i);
            break;

        case HS_USARSIM_TYPE_VIDEO:
            new USARSimVideo(
                &gVideo,
                (char *)(g_clUSARSimInitList[i].host.c_str()),
                (char *)(g_clUSARSimInitList[i].name.c_str()),
                g_clUSARSimInitList[i].port,
                g_clUSARSimInitList[i].showImage);
            break;

        default:
            printfTextWindow(
                "Unknown USARSim type '%d'.",
                g_clUSARSimInitList[i].type);
            break;
        }

        sleep(2);
    }

    // USARSimLaser has to be instantiated after USARSimRobot.
    for (i = 0; i < usarsimInitListLaserIndexes.size(); i++)
    {
        index = usarsimInitListLaserIndexes[i];
        new USARSimLaser(
            &gUSARSimLaser,
            (char *)(g_clUSARSimInitList[index].name.c_str()),
            g_clUSARSimInitList[index].angleOffset, 
            g_clUSARSimInitList[index].xOffset,
            g_clUSARSimInitList[index].yOffset,
            g_clUSARSimInitList[index].laserPitchAngle);
    }

    for (i = 0; i < g_clRovioInitList.size(); i++)
    {
#ifndef NO_ROVIO
        switch (g_clRovioInitList[i].type) {

        case HS_ROVIO_TYPE_ROBOT:
            if (gRobot == NULL)
            {
                new RovioRobot(
                    &gRobot,
                    (char *)(g_clRovioInitList[i].host.c_str()),
                    (char *)(g_clRovioInitList[i].name.c_str()),
                    (char *)(g_clRovioInitList[i].robot_class.c_str()),
                    g_clRovioInitList[i].port,
                    g_clRovioInitList[i].pos_x,
                    g_clRovioInitList[i].pos_y,
                    g_clRovioInitList[i].pos_z,
                    g_clRovioInitList[i].rot_yaw,
                    g_clRovioInitList[i].rot_pitch,
                    g_clRovioInitList[i].rot_roll);
		UsingRovio = true;
            }
            else
            {
                printfTextWindow("HServer cannot connect to more than one robot module.");
            }
            break;
/*
        case HS_ROVIO_TYPE_VIDEO:
            new RovioVideo(
                &gVideo,
                (char *)(g_clRovioInitList[i].host.c_str()),
                (char *)(g_clRovioInitList[i].name.c_str()),
                g_clRovioInitList[i].port,
                g_clRovioInitList[i].showImage);
            break;
*/
        default:
            printfTextWindow(
                "Unknown Rovio type '%d'.",
                g_clRovioInitList[i].type);
            break;
        }

        sleep(2);
#else
        printTextWindow("Rovio disabled.");
#endif
    }


    for (i = 0; i < g_clSkybotixInitList.size(); i++) {
      switch (g_clSkybotixInitList[i].type) {

        case HS_SKYBOTIX_TYPE_ROBOT:
	  if (gRobot == NULL) {
	    new SkybotixRobot(
			      &gRobot,
			      (char *)(g_clSkybotixInitList[i].host.c_str()),
			      (char *)(g_clSkybotixInitList[i].name.c_str()),
			      (char *)(g_clSkybotixInitList[i].robot_class.c_str()),
			      g_clSkybotixInitList[i].port,
			      g_clSkybotixInitList[i].pos_x,
			      g_clSkybotixInitList[i].pos_y,
			      g_clSkybotixInitList[i].pos_z,
			      g_clSkybotixInitList[i].rot_yaw,
			      g_clSkybotixInitList[i].rot_pitch,
			      g_clSkybotixInitList[i].rot_roll);
	    UsingSkybotix = true;
	  }
	  else
            {
	      printfTextWindow("HServer cannot connect to more than one robot module.");
            }
	  break;

      default:
	printfTextWindow(
                "Unknown Skybotix type '%d'.", g_clSkybotixInitList[i].type);
	break;
      }

      sleep(2);

    }



    if (bUseStream)
    {
        Laser::streamStart();
        sleep(2);
    }

    if (g_clFrameGrabberInitList.size() > 0)  
    {
        new FrameGrabber(
            &gVideo,
            g_clFrameGrabberInitList[0].strPortString,
            g_clFrameGrabberInitList[0].showImage);
        sleep(2);
    }
    
    if (useUrban) {
#ifndef NO_MOBILITY
        new Urban(&gRobot);
        sleep(2);
#else 
        printTextWindow("No Mobility support");
#endif
    }

    if (g_clNomadInitList.size() > 0)
    {
        new Nomad(
            &gRobot,
            g_clNomadInitList[0].enType,
            g_clNomadInitList[0].strPortString, 
            g_clNomadInitList[0].strHost);

        sleep(2);
    }

    if (useApm)
    {
        new Apm(&apm);
        sleep(2);
    }
    
    if (useX)
    {
        new Xwindow(
            &gXwindow,
            NULL,
            DEFAULT_XWINDOW_X,
            DEFAULT_XWINDOW_Y,
            DEFAULT_XWINDOW_WIDTH,
            DEFAULT_XWINDOW_HEIGHT);

        sleep(2);
    }

    if (g_clPioneerInitList.size() > 0)
    {
        new Pioneer(&gRobot, g_clPioneerInitList[0].strPortString);
        sleep(2);
    }

    if (g_clPioneer2InitList.size() > 0)
    {
        new Pioneer2(
            &gRobot,
            g_clPioneer2InitList[0].strPortString);
        sleep(2);
    }

    if (g_clAmigoBotInitList.size() > 0)
    {
        new AmigoBot(&gRobot, g_clAmigoBotInitList[0].strPortString);
        sleep(2);
    }

    if (useFred)
    {
        new Fred(&gRobot);
        sleep(2);
    }

    if (useJbox)
    {
        Jbox::connectToJbox(
            jbox,
            jboxID,
            jboxDisableNetworkQuery,
            gps_base_lat,
            gps_base_lon,
            gps_base_x,
            gps_base_y,
            gps_mperlat,
            gps_mperlon);

        sleep (2);
    }

    if ((g_clOpenCVInitList.size()) > 0)
    {
#ifndef NO_OPENCV
        for (i = 0; i < g_clOpenCVInitList.size(); i++)
        {
            new OpenCVHandler(
                &gOpenCVHandler,
                g_clOpenCVInitList[i].trackBlobs,
                g_clOpenCVInitList[i].detectFace,
                g_clOpenCVInitList[i].detectObjects,
                g_clOpenCVInitList[i].directGrab,
                g_clOpenCVInitList[i].showImage);

            sleep(2);
        }
#else
        printTextWindow("OpenCV disabled.");
#endif
    }

    if (g_clGyroInitList.size() > 0)
    {
        switch (g_clGyroInitList[0].type) {

        case (int)(Gyro::TYPE_DMUVGX):
#ifndef NO_MOBILITY
            new GyroDMUVGX(&gGyro);
            sleep(2);
#else 
            printTextWindow("No Mobility support");
#endif
            break;

        default:
            break;
        }
    }

    if (g_clCompassInitList.size() > 0)
    {
        switch (g_clCompassInitList[0].type) {

        case (int)(Compass::TYPE_KVHC100):
#ifndef NO_MOBILITY
            new CompassKVHC100(&gCompass);
            sleep(2);
#else 
            printTextWindow("No Mobility support");
#endif
            break;

        case (int)(Compass::TYPE_3DMG):
            new Compass3DMG(&gCompass, g_clCompassInitList[0].portString);
            sleep(2);
            break;
        }
    }

    if (g_clCricketInitList.size() > 0)
    {

            new Cricket(&cricket, g_clCricketInitList[0].IP, g_clCricketInitList[0].ID);
            sleep(2);
    }


    //if (g_clGpsInitList.size() > 0)
    if (useGps)
    {
        new Gps(
            &gps,
            g_clGpsInitList[0].strPortString,
            g_clGpsInitList[0].dBaseLat, 
            g_clGpsInitList[0].dBaseLong,
            g_clGpsInitList[0].dXDiff, 
            g_clGpsInitList[0].dYDiff,
            g_clGpsInitList[0].dMPerLat, 
            g_clGpsInitList[0].dMPerLong,
            g_clGpsInitList[0].bUseBase);

        sleep(2);
    }

    if (g_clPoseCalcInitList.size() > 0)
    {
        poseCalcConstants.compass2MlabHeadingFacor = 
            g_clPoseCalcInitList[0].compass2MlabHeadingFacor;
        poseCalcConstants.compass2MlabHeadingOffset =
            g_clPoseCalcInitList[0].compass2MlabHeadingOffset;
        poseCalcConstants.compassMaxValidAngSpeed4Heading =
            g_clPoseCalcInitList[0].compassMaxValidAngSpeed4Heading;
        poseCalcConstants.gps2MlabHeadingFacor = 
            g_clPoseCalcInitList[0].gps2MlabHeadingFacor;
        poseCalcConstants.gps2MlabHeadingOffset = 
            g_clPoseCalcInitList[0].gps2MlabHeadingOffset;
        poseCalcConstants.gpsMinValidTransSpeed4Heading = 
            g_clPoseCalcInitList[0].gpsMinValidTransSpeed4Heading;
        poseCalcConstants.gpsMaxValidAngSpeed4Heading = 
            g_clPoseCalcInitList[0].gpsMaxValidAngSpeed4Heading;

        if ((poseCalc = new PoseCalc(
                 poseCalcConstants,
                 g_clPoseCalcInitList[0].fuserType,
                 g_clPoseCalcInitList[0].varianceScheme)) == NULL)
        {
            printTextWindow("Fatal error: Couldn't allocate memory for PoseCalculator.");
            exit(-1);
        }

        gPoseCalc = poseCalc;
    }
    else
    {
        printTextWindow(
            "Fatal error: Section name for PoseCalculator not found in the RC file.");
        exit(-1);
    }

    refreshScreen();

    if (gRobot == NULL)
    {
        // Connect to a robot if not done yet.
        robotConnect(gRobot);
    }

    // Configure general robot parameters
    if (gRobot != NULL) {
        if (g_clRobotInitList.size() > 1) {
	   fprintf(stderr, "Warning: There are multiple robot configuration sections!\n");
        }

	for (i = 0; i < g_clRobotInitList.size(); i++) {
	   gRobot->setMinimumPointTurnAngle(g_clRobotInitList[i].minPointTurnAngle);
	   gRobot->setMaxSpeed(g_clRobotInitList[i].maxSpeed);
	   gRobot->setMaxSteerSpeed(g_clRobotInitList[i].maxSteerSpeed);
	   gRobot->setMaxSteerSpeedAngle(g_clRobotInitList[i].maxSteerSpeedAngle);
	   gRobot->setMaxAltitudeSpeed(g_clRobotInitList[i].maxAltitudeSpeed);
	   gRobot->setMaxNoCommandTimeSec(g_clRobotInitList[i].maxNoCommandTimeSec);
	   gRobot->setMinSteerSpeed(g_clRobotInitList[i].minSteerSpeed);
	   gRobot->setStoppingSpeedThreshold(g_clRobotInitList[i].stoppingSpeedThreshold);
	   gRobot->setNumVelocitySamples(g_clRobotInitList[i].numVelocitySamples);
    	}
    }

    if (gRobot != NULL)
    {
        poseCalcID = gPoseCalc->addRobot(gRobot);
        gRobot->savePoseCalcID(poseCalcID);
    }
    else
    {
        printTextWindow("\n");
        printTextWindow("Error: HServer has to be running with a robot.");
        exit(-1);
    }


    if (gps != NULL)
    {
        poseCalcID = gPoseCalc->addGps(gps);
        gps->savePoseCalcID(poseCalcID);
    }

    // PU TODO
    if (cricket != NULL)
    {
        printfTextWindow("Adding cricket to posecalc");
        poseCalcID = gPoseCalc->addCricket(cricket);
        cricket->savePoseCalcID(poseCalcID);
    }


    if (gGyro != NULL)
    {
        poseCalcID = gPoseCalc->addGyro(gGyro);
        gGyro->savePoseCalcID(poseCalcID);
    }

    if (gCompass != NULL)
    {
        poseCalcID = gPoseCalc->addCompass(gCompass);
        gCompass->savePoseCalcID(poseCalcID);
    }

    if (gRemoteControl == NULL)
    {
        gRemoteControl = new RemoteControl();
    }

    gPoseCalc->resetPose();

    pthread_create(&keyboard_reader, NULL, (void* (*)(void*)) &keyboard_reader_thread, NULL);

    sem_wait(&exit_sem);
    return 0;
} 



string FindRCFile(string strOriginalFileName)
{
    unsigned int i;
    vector<string> clDirList;
    FILE* pfFile;
    string strCurrentFileName;
    string strRCFileName = "";

    // first, try the raw file name that's passed in to check the current directory
    if ((pfFile = fopen(strOriginalFileName.c_str(), "r")) != NULL)
    {
        fclose(pfFile);
        strRCFileName = strOriginalFileName;
    }

    // if the file hasn't been found yet, check the home directory
    if (strRCFileName == "")
    {
        const char* szHomeDir = getenv("HOME");
        if (szHomeDir != NULL)
        {
            strCurrentFileName = TruncateSlash(szHomeDir) + '/' + strOriginalFileName;
            if ((pfFile = fopen(strCurrentFileName.c_str(), "r")) != NULL)
            {
                fclose(pfFile);
                strRCFileName = strCurrentFileName;
            }
        }
    }

    // if the file hasn't been found yet, check the directories in the PATH
    // environment variable
    if (strRCFileName == "")
    {
        clDirList = ExtractDirectories("PATH");
        for (i = 0; i < clDirList.size(); i++)
        {
            strCurrentFileName = clDirList[i] + '/' + strOriginalFileName;
            if ((pfFile = fopen(strCurrentFileName.c_str(), "r")) != NULL)
            {
                fclose(pfFile);
                strRCFileName = strCurrentFileName;
                break;
            }
        }
    }

    return strRCFileName;
}

void usage(void)
{
    printf("If ran with no arguments hserver starts up with no hardware connections.\n");
    printf("Hardware can be then connected via keyboard command.\n");
    printf("\n");
    printf("Hardware can be started up automatically with command line arguments.\n");
    printf("\n");
    printf("-C config_file_name Use this configuration file\n");
    printf("-c camera_section   Connect to a Sony evi-d20 described in the configuration\n"); 
    printf("                    file under camera section <start camera ...>\n");
    printf("-l laser_section    Connect to a SICK LMS200 laser scanner using settings\n");
    printf("                    described in the configuration file under laser\n");
    printf("                    section <start laser ...>\n");
    printf("-m cognachrome_section\n");
    printf("                    Connect to cognachrome vision system described in the\n");
    printf("                    configuration file under cognachrome section\n");
    printf("                    <start cognachrome ...>\n");
    printf("-n nomad_section    Connect to a Nomad 150 or Nomad 200 described in the\n");
    printf("                    configuration file under nomad section <nomad_section>\n");
    printf("-p pioneer_section  Connect a Pioneer described in the configuration\n");
    printf("                    file under pioneer section <start pioneer ...>\n");
    printf("-2 pioneer2_section Connect a Pioneer2 described in the configuration\n");
    printf("                    file under pioneer2 section <start pioneer2 ...>\n");
    printf("-b amigobot_section Connect an AmigoBot described in the configuration\n");
    printf("                    file under amigobot section <start amigobot ...>\n");
    printf("-u                  Connect urban via mobility\n");
    printf("-j                  Connect ATRV-Jr via mobility\n");
    printf("-v framegrabber_section\n");
    printf("                    Connect to framegrabber described in the configuration\n");
    printf("                    file under framegrabber section <start framegrabber ...>\n");
    printf("-g gps_section      Connect to the GPS unit described in the configuration file\n");
    printf("                    under gps section <start gps ...>\n");
    printf("-A cricket_section  Connect to the cricket position server described in configuration file\n");
    printf("-y gyro_section     Connect to a gyroscope.\n");
    printf("-k compass_section  Connect a compass described in the configuration\n");
    printf("                    file under compass section <start compass ...>\n");
    printf("\n");
    printf("Other options\n");
    printf("\n");
    printf("-r                  hserver ipt name  defaults to fred\n");
    printf("-B                  Display battery information\n");
    printf("-f                  Connect Fred\n");
    printf("-i hostname         Start ipt. Set ipthost to [hostname] (optional)\n");
    printf("-a                  Start ipt and register hclient messages\n");
    printf("-o                  Setup ipt handler for multiple hservers\n");
    printf("-x                  Open X window\n");
    printf("-S                  no spash screen\n");
    printf("-J                  Connect to JBox\n");
    printf("-P posecalc_section Pose Calculator configuration\n");
    printf("-R                  Enable the RCom connection.\n");
    printf("-G gazebo_section   Enable Gazebo.\n");
    printf("-O opencv_section   Enable OpenCV.\n");
    printf("-U usarsim_section  Enable USARSim.\n");
    printf("-W rovio_section    Enable WowWee Rovio.\n");
    printf("-X skybotix_section Enable Skybotix UAV.\n");
    printf("\n");
    printf("-h                  this help message\n");
}

int main(int argc, char* argv[])
{
    int c;
    unsigned int i, j;
    char* szEnvironment;
    string strRCFileName = DEFAULT_RC_FILE_NAME;
    vector<SuOption> clOptionList;
    SuOption suOption;
    bool posecalcSectionSpecified = false;

    useTerminal = false;
    useUrban = false;
    useATRVJr = false;
    useIpt = false;
    useRCom = false;
    useHclient = false;
    useX = false;
    useFred = false;
    multipleHservers = false;
    
    szEnvironment = getenv("HSERVER_USE_IPT");
    if (szEnvironment != NULL)
    {
        useIpt = true;
    }

    while((c = getopt (argc, argv, "2:ab:A:Bc:C:efg:G:hi:jJ:k:l:Lm:Mn:O:p:P:r:RSt:uU:W:v:X:xy:z:c:")) != -1)
    {
        switch (c) {

        case 'A':
            useCricket = true;
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_CRICKET;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
	    break;

        case 'a':
            useIpt = true;
            useHclient = true;
            break;

        case 'b':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_AMIGOBOT;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'B':
            useApm = true;
            break;

        case 'c':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_CAMERA;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'C':
            strRCFileName = optarg;
            break;
            
        case '2':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_PIONEER2;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'f':
            useFred = true;
            break;

        case 'g':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_GPS;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            useGps = true;
            break;

        case 'G':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_GAZEBO;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'h':
            usage();
            return 0;

        case 'i':
            useIpt = true;
            if (optarg)
            {
                ipt_server_name = strdup(optarg);
            }
            break;

        case 'j':
            useATRVJr = true;
            break;

        case 'J':
            useJbox = true;
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_JBOX;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'k':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_COMPASS;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'l':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_LASER;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'L':
            logData = true;
            break;
            
        case 'm':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_COGNACHROME;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'M':
            multipleHservers = true;
            break;

        case 'n':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_NOMAD;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'p':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_PIONEER;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'P':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_POSECALC;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            posecalcSectionSpecified = true;
            break;

        case 'O':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_OPENCV;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'r':
            strcpy(robot_name, optarg);
            break;

        case 'R':
            useRCom = true;
            break;

        case 'S':
            noSplash = true;
            break;

        case 'u':
            useUrban = true;
            break;

        case 'U':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_USARSIM;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'W':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_ROVIO;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

	case 'X':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_SKYBOTIX;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'z':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_ROBOT;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
	    break;

        case 'v':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_FRAMEGRABBER;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        case 'x':
            useX = true;
            break;

        case 'y':
            suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_GYRO;
            suOption.strSectionName = optarg;
            clOptionList.push_back(suOption);
            break;

        default:
            usage();
            exit(-1);
        }
    }


    if (!posecalcSectionSpecified)
    {
        suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_POSECALC;
        suOption.strSectionName = DEFAULT_POSECALC_RC_SECTION_NAME;
        clOptionList.push_back(suOption);
        posecalcSectionSpecified = true;
    }

    string strOriginalRCFileName = strRCFileName; // save this for error msg
    strRCFileName = FindRCFile(strRCFileName);  // try to find the RC file using 
    //  PATH and HOME environment vars
    // make sure we have a valid RC file name
    if (strRCFileName == "")
    {
        fprintf(stderr, "hserver error: couldn't find RC file: %s\n", 
                strOriginalRCFileName.c_str());
        exit(-1);
    }

    // parse the RC file into a ClHServerConfig object
    ClHSRCFile clRCFile(strRCFileName);
    ClHServerConfig clConfig;
    if (!clRCFile.CreateConfiguration(clConfig))
    {
        fprintf(stderr, "hserver error: error with configuration file\n");
        exit(-1);
    }

    ClHServerConfigurator clConfigurator(clConfig);

    bool bFoundSection = false;

    // look at all options from the command line that could specify sections from the config file
    for (i = 0; i < clOptionList.size(); i++)
    {
        bFoundSection = false;
        for (j = 0; j < clConfig.GetSectionCount(); j++)
        {
            string strSectionName;

            // check to see if the current option matches any sections in the config file
            if ((clConfig.GetSectionType(j) == clOptionList[i].enSectionType) &&
                clConfig.GetSectionName(j, strSectionName) && 
                (clOptionList[i].strSectionName == strSectionName))
            {
                SuRobotInitData suRobotData;
                SuCameraInitData suCameraData;
                SuLaserInitData suLaserData;
                SuGazeboInitData suGazeboData;
                SuUSARSimInitData suUSARSimData;
                SuRovioInitData suRovioData;
		SuSkybotixInitData suSkybotixData;
                SuOpenCVInitData suOpenCVData;
                SuCognachromeInitData suCognachromeData;
                SuNomadInitData suNomadData;
                SuPioneerInitData suPioneerData;
                SuPioneer2InitData suPioneer2Data;
                SuAmigoBotInitData suAmigoBotData;
                SuFrameGrabberInitData suFrameGrabberData;
                SuGpsInitData suGpsData;
                SuCricketInitData suCricketData;
                SuJboxInitData suJboxData;
                SuCompassInitData suCompassData;
                SuGyroInitData suGyroData;
                SuPoseCalcInitData suPoseCalcData;

                // get device-specific info from a general option
                switch (clOptionList[i].enSectionType) {

                case ClHServerConfig::EnHServerConfigSectionType_CAMERA:
                    suCameraData = clConfigurator.GetCameraOptions((int) j);
                    g_clCameraInitList.push_back(suCameraData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_LASER:
                    suLaserData = clConfigurator.GetLaserOptions((int) j);
                    g_clLaserInitList.push_back(suLaserData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_ROBOT:
                    suRobotData = clConfigurator.GetRobotOptions((int) j);
                    g_clRobotInitList.push_back(suRobotData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_GAZEBO:
                    suGazeboData = clConfigurator.GetGazeboOptions((int) j);
                    g_clGazeboInitList.push_back(suGazeboData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_USARSIM:
                    suUSARSimData = clConfigurator.GetUSARSimOptions((int) j);
                    g_clUSARSimInitList.push_back(suUSARSimData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_ROVIO:
                    suRovioData = clConfigurator.GetRovioOptions((int) j);
                    g_clRovioInitList.push_back(suRovioData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_SKYBOTIX:
                    suSkybotixData = clConfigurator.GetSkybotixOptions((int) j);
                    g_clSkybotixInitList.push_back(suSkybotixData);
                    break;
		    

                case ClHServerConfig::EnHServerConfigSectionType_OPENCV:
                    suOpenCVData = clConfigurator.GetOpenCVOptions((int) j);
                    g_clOpenCVInitList.push_back(suOpenCVData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_COGNACHROME:
                    suCognachromeData = clConfigurator.GetCognachromeOptions((int) j);
                    g_clCognachromeInitList.push_back(suCognachromeData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_NOMAD:
                    suNomadData = clConfigurator.GetNomadOptions((int) j);
                    g_clNomadInitList.push_back(suNomadData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_PIONEER:
                    suPioneerData = clConfigurator.GetPioneerOptions((int) j);
                    g_clPioneerInitList.push_back(suPioneerData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_PIONEER2:
                    suPioneer2Data = clConfigurator.GetPioneer2Options((int) j);
                    g_clPioneer2InitList.push_back(suPioneer2Data);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_AMIGOBOT:
                    suAmigoBotData = clConfigurator.GetAmigoBotOptions((int) j);
                    g_clAmigoBotInitList.push_back(suAmigoBotData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_FRAMEGRABBER:
                    suFrameGrabberData = clConfigurator.GetFrameGrabberOptions((int) j);
                    g_clFrameGrabberInitList.push_back(suFrameGrabberData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_GPS:
                    suGpsData = clConfigurator.GetGpsOptions((int) j);
                    g_clGpsInitList.push_back(suGpsData);
                    if (!suGpsData.bUseBase) // this needs to go away some time, along w/ the 
                        // global vars, but for time's sake now... jbl
                    {
                        gps_base_lat = suGpsData.dBaseLat;
                        gps_base_lon = suGpsData.dBaseLong;
                        gps_base_x = suGpsData.dXDiff;
                        gps_base_y = suGpsData.dYDiff;
                        gps_mperlat = suGpsData.dMPerLat;
                        gps_mperlon = suGpsData.dMPerLong;
                    }
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_CRICKET:
                    suCricketData = clConfigurator.GetCricketOptions((int) j);
                    g_clCricketInitList.push_back(suCricketData);
                    break;
		

                case ClHServerConfig::EnHServerConfigSectionType_JBOX:
                    suJboxData = clConfigurator.GetJboxOptions((int) j);
                    g_clJboxInitList.push_back(suJboxData);
                    jboxID = suJboxData.jboxID;
                    jboxDisableNetworkQuery = suJboxData.disableNetworkQuery;
                    if (!useGps)
                    {
                        suOption.enSectionType = ClHServerConfig::EnHServerConfigSectionType_GPS;
                        suOption.strSectionName = suJboxData.gpsSection;
                        clOptionList.push_back(suOption);
                    }
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_COMPASS:
                    suCompassData = clConfigurator.GetCompassOptions((int) j);
                    g_clCompassInitList.push_back(suCompassData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_GYRO:
                    suGyroData = clConfigurator.GetGyroOptions((int) j);
                    g_clGyroInitList.push_back(suGyroData);
                    break;

                case ClHServerConfig::EnHServerConfigSectionType_POSECALC:
                    suPoseCalcData = clConfigurator.GetPoseCalcOptions((int) j);
                    g_clPoseCalcInitList.push_back(suPoseCalcData);
                    break;

                default:
                    fprintf(stderr, "hserver error: encountered an invalid option type.\n");
                    exit(-1);
                }

                bFoundSection = true;
                break;
            }
        }
        if (!bFoundSection)
        {
            fprintf(stderr, "hserver error: option with name \"%s\" could not be found in\n", 
                    clOptionList[i].strSectionName.c_str());
            fprintf(stderr, "   the configuration file.\n");
            exit(-1);
        }
    }

    startHServer();
}

/**********************************************************************
# $Log: hserver.c,v $
# Revision 1.14  2010-05-03 22:10:08  zkira
# Cricket sensor, blob tracking, rovio (inc. IR sensor), skybotix interface (unfinished)
#
# Revision 1.10  2009/05/10 20:07:28  zkira
# Rovio additions
#
# Revision 1.9  2009/04/17 17:51:48  zkira
# Added maxSpeed in robot (clamps speed to this max value), and ability to speed up usarsim data acquisition so that speeding up the USARSim simulator becomes possible
#
# Revision 1.8  2009/04/13 02:58:43  zkira
# RFID code
#
# Revision 1.7  2009/03/02 18:54:33  zkira
# Added Rovio and "Robot" (options for lower level robot movement parameters)
#
# Revision 1.6  2009/01/27 20:21:57  zkira
# Added laserPitchAngle to Gazebo configuration
#
# Revision 1.5  2008/11/20 02:17:20  ts2883
# Finished adding the pitch information for the laser.
#
# Revision 1.4  2008/10/30 22:50:31  zkira
# Fixed bug: accidental break; that shouldn't have been there
#
# Revision 1.3  2008/09/16 16:51:15  zkira
# Added gazebo parameters (robot class, client id, server id) to constructor
#
# Revision 1.2  2008/07/15 22:33:50  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.28  2008/06/18 19:23:14  endo
# USARSim experiment final.
#
# Revision 1.27  2008/04/14 21:51:27  endo
# Regular backup.
#
# Revision 1.26  2008/04/13 00:33:42  endo
# Regular backup.
#
# Revision 1.25  2008/04/12 20:17:24  endo
# OpenCV face detection implemented.
#
# Revision 1.24  2008/02/01 23:18:25  endo
# USARSimLaser added.
#
# Revision 1.23  2008/01/21 19:14:16  endo
# Regular backup.
#
# Revision 1.22  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.21  2008/01/11 15:51:18  endo
# Regular backup.
#
# Revision 1.20  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
# Revision 1.19  2006/12/22 18:49:36  endo
# Regular backup.
#
# Revision 1.18  2006/12/19 20:12:43  endo
# Regular backup.
#
# Revision 1.17  2006/12/17 04:42:53  endo
# SENSOR_OPENCV_BLOB_COLOR added.
#
# Revision 1.16  2006/12/12 20:45:15  endo
# Regular update.
#
# Revision 1.15  2006/12/01 21:14:54  endo
# Regular backup.
#
# Revision 1.14  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.13  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.12  2006/11/22 22:24:28  endo
# OpenCV BlobTrack implemented.
#
# Revision 1.11  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.10  2006/11/20 22:55:52  endo
# OpenCV added.
#
# Revision 1.9  2006/11/20 16:43:37  endo
# Xwindow improved.
#
# Revision 1.8  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.7  2006/11/15 23:45:58  endo
# GazeboLaser can now accept the second laser.
#
# Revision 1.6  2006/08/14 21:12:13  endo
# RComHandler implemented.
#
# Revision 1.5  2006/08/09 22:22:16  endo
# RComHandler added.
#
# Revision 1.4  2006/08/08 21:33:49  endo
# GazeboLaser added.
#
# Revision 1.3  2006/07/31 22:01:14  endo
# Gazebo -> GazeboRobot
#
# Revision 1.2  2006/07/21 23:35:01  endo
# Gazebo integrated into HServer
#
# Revision 1.1.1.1  2006/07/21 17:42:09  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/24 21:51:20  endo
# gcc 4.1 upgrade
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.11  2004/11/22 00:18:37  endo
# Network query of JBox can now be disabled.
#
# Revision 1.10  2004/09/22 03:59:34  endo
# *** empty log message ***
#
# Revision 1.9  2004/09/10 19:41:06  endo
# New PoseCalc integrated.
#
# Revision 1.8  2004/04/24 11:20:53  endo
# 3DM-G added.
#
# Revision 1.7  2004/04/19 08:49:45  endo
# Particle filter fixed. Visualization option added.
#
# Revision 1.6  2004/04/17 12:20:35  endo
# Intalling JBox Sensor Update.
#
# Revision 1.5  2004/04/13 19:25:33  yangchen
# JBox module integrated.
#
# Revision 1.4  2004/04/12 01:14:36  endo
# noSplash meant to be true.
#
# Revision 1.3  2004/04/12 01:12:19  endo
# Default noSplash is now "false".
#
# Revision 1.2  2004/04/06 02:07:39  pulam
# Added code for Silent Mode
#
# Revision 1.1.1.1  2004/02/27 22:30:50  endo
# Initial MARS 2020 revision
#
# Revision 1.14  2003/04/26 06:48:31  ebeowulf
# Added the Pioneer2-dxe controller.
#
# Revision 1.13  2003/04/06 13:32:37  endo
# gcc 3.1.1
#
# Revision 1.12  2002/04/01 22:20:44  endo
# sound DSP removed.
#
# Revision 1.11  2002/02/18 13:48:21  endo
# AmigoBot added.
#
# Revision 1.10  2002/01/16 21:46:27  ernest
# Added WatchDog functionality
#
# Revision 1.9  2001/05/29 22:36:30  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.8  2001/03/23 21:19:44  blee
# massive modifications to use a config file
#
# Revision 1.7  2000/12/12 22:54:18  blee
# Updates were added to support multiple lasers.
# Added printfTextWindow.
# Fixed usage() and added the -h command line option for help.
#
# Revision 1.6  2000/11/13 20:09:23  endo
# ARTV-Jr class added to HServer.
#
# Revision 1.5  2000/09/19 09:06:31  endo
# robot name -> hserver name.
#
# Revision 1.4  2000/09/19 03:14:21  endo
# RCS log added.
#
#
#**********************************************************************/
