/**********************************************************************
 **                                                                  **
 **                              robot.c                             **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Generic robot class for HServer                                 **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: robot.c,v 1.11 2010-05-03 22:10:09 zkira Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <pthread.h>
#include <math.h>
#include <string>
#include <algorithm>
#include <numeric>

#include "message.h"
#include "robot.h"
#include "hserver.h"
#include "PoseCalculatorInterface.h"

using std::string;

//-----------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t Robot::poseMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::velocityMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::controlSpeedMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::desiredRotationMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::desiredSpeedMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::travelDistMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::minPointTurnAngleMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::reversableDriveMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t Robot::batteryVoltageMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------

// Defaults for various mobility and other parameters
const double Robot::DEFAULT_MIN_POINT_TURN_ANGLE_ = 30.0;
const double Robot::DEFAULT_MAX_SPEED_ = 15.0;
const double Robot::DEFAULT_MAX_STEER_SPEED_ = 15.0;
const double Robot::DEFAULT_MAX_STEER_SPEED_ANGLE_ = 15.0;
const double Robot::DEFAULT_MAX_ALTITUDE_SPEED_ = 0.25;
double Robot::DEFAULT_MAX_NO_COMMAND_TIME_SEC_ = 2.0;
const double Robot::DEFAULT_MIN_STEER_SPEED_ = 6.0;
const double Robot::DEFAULT_STOPPING_SPEED_THRESHOLD_ = 0.1;
const int    Robot::DEFAULT_NUM_VELOCITY_SAMPLES_ = 10;

float desiredang = 0.0f;
bool startrot = true;

//-----------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------
Robot::Robot(Robot** a, int robotType, string name) : 
   Module((Module**) a, name),
   shBatteryVoltage_(0),
   shMinPointTurnAngle_(DEFAULT_MIN_POINT_TURN_ANGLE_),
   shTravelDist_(0),
   maxSpeed_(DEFAULT_MAX_SPEED_),
   maxSteerSpeed_(DEFAULT_MAX_STEER_SPEED_),
   maxSteerSpeedAngle_(DEFAULT_MAX_STEER_SPEED_ANGLE_),
   maxAltitudeSpeed_(DEFAULT_MAX_ALTITUDE_SPEED_),
   maxNoCommandTimeSec_(DEFAULT_MAX_NO_COMMAND_TIME_SEC_),
   minSteerSpeed_(DEFAULT_MIN_STEER_SPEED_),
   stoppingSpeedThreshold_(DEFAULT_STOPPING_SPEED_THRESHOLD_),
   numVelocitySamples_(DEFAULT_NUM_VELOCITY_SAMPLES_),
   robotType_(robotType),
   poseCalcID_(-1),
   velIndex_(0),
   shReversableDrive_(false),
   velBufferFilled_(false)
{

    if (UsingRovio) {
      DEFAULT_MAX_NO_COMMAND_TIME_SEC_ = 0.001;
    }
    battery_level = -1;
    memset((HSPose_t *)&shPose_, 0x0, sizeof(HSPose_t));
    memset((HSPose_t *)&shLastPose_, 0x0, sizeof(HSPose_t));
    memset((HSVelocity_t *)&shVelocity_, 0x0, sizeof(HSVelocity_t));
    memset((HSSpeed_t *)&shControlSpeed_, 0x0, sizeof(HSSpeed_t));
    memset((HSSpeed_t *)&shDesiredSpeed_, 0x0, sizeof(HSSpeed_t));
    memset((HSRotation_t *)&shDesiredRotation_, 0x0, sizeof(HSRotation_t));
    memset((VelocityBuffers_t *)&velBuffers_, 0x0, sizeof(VelocityBuffers_t));
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
Robot::~Robot(void)
{
    stop();

    if ((gPoseCalc != NULL) && (poseCalcID_ >= 0))
    {
        gPoseCalc->disconnectModule(poseCalcID_);
        poseCalcID_ = -1;
    }
}

//-----------------------------------------------------------------------
// This function initializes the robot. It is a virtual function which
// can be overwritten by a derived robot.
//-----------------------------------------------------------------------
void Robot::initRobot(void)
{
    //resetPose_();
}

//-----------------------------------------------------------------------
// This function returns the current robot pose.
//-----------------------------------------------------------------------
void Robot::getPose(HSPose_t &pose)
{
    pthread_mutex_lock(&poseMutex_);
    pose = shPose_;
    pthread_mutex_unlock(&poseMutex_);
}

//-----------------------------------------------------------------------
// This function sets the pose to be the specified value.
//-----------------------------------------------------------------------
void Robot::setPose_(HSPose_t pose)
{
    pthread_mutex_lock(&poseMutex_);
    shLastPose_ = shPose_;
    shPose_ = pose;
    pthread_mutex_unlock(&poseMutex_);

    if ((gPoseCalc != NULL) && (poseCalcID_ >= 0))
    {
        gPoseCalc->updateModulePose(poseCalcID_, pose);
    }
}

//-----------------------------------------------------------------------
// This function returns the traveling distance.
//-----------------------------------------------------------------------
void Robot::getTravelDist(double &travelDist)
{
    pthread_mutex_lock(&travelDistMutex_);
    travelDist = shTravelDist_;
    pthread_mutex_unlock(&travelDistMutex_);
}

//-----------------------------------------------------------------------
// This function sets the traveling distance.
//-----------------------------------------------------------------------
void Robot::addTravelDist_(double travelDist)
{
    pthread_mutex_lock(&travelDistMutex_);
    shTravelDist_ += travelDist;
    pthread_mutex_unlock(&travelDistMutex_);
}

//-----------------------------------------------------------------------
// This function updates the pose based on the specified increments.
//-----------------------------------------------------------------------
void Robot::advancePose_(HSPose_t deltaPose)
{
    HSPose_t pose;
    double curTime;

    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&poseMutex_);

    shLastPose_ = shPose_;
    pose = shPose_;
    pose.loc.x += deltaPose.loc.x;
    pose.loc.y += deltaPose.loc.y;
    pose.loc.z += deltaPose.loc.z;
    pose.rot.yaw += deltaPose.rot.yaw;
    ROBOT_CRAMPDEG(pose.rot.yaw, 0.0, 360);
    pose.rot.pitch += deltaPose.rot.pitch;
    ROBOT_CRAMPDEG(pose.rot.pitch, 0.0, 360);
    pose.rot.roll += deltaPose.rot.roll;
    ROBOT_CRAMPDEG(pose.rot.roll, 0.0, 360);
    pose.loc.extra.time = curTime;
    pose.rot.extra.time = curTime;
    pose.extra.time = curTime;
    shPose_ = pose;

    pthread_mutex_unlock(&poseMutex_);


    if ((gPoseCalc != NULL) && (poseCalcID_ >= 0))
    {
        gPoseCalc->updateModulePose(poseCalcID_, pose);
    }
}

//-----------------------------------------------------------------------
// This function updates the pose based on the specified speed and delta
// time (considered in 2D).
//-----------------------------------------------------------------------
void Robot::advancePose2D_(HSSpeed_t speed, double deltaTimeSec)
{
    HSPose_t pose;
    double dist, angle, altDist, dx, dy, dz;
    double curTime;

    dist  = speed.driveSpeed.value*deltaTimeSec;
    angle = speed.steerSpeed.value*deltaTimeSec;
    altDist = speed.altitudeSpeed.value*deltaTimeSec;
    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&poseMutex_);
    shLastPose_ = shPose_;
    pose = shPose_;

    dy = sin(ROBOT_DEG2RAD(pose.rot.yaw+(angle/2.0)))*dist;
    dx = cos(ROBOT_DEG2RAD(pose.rot.yaw+(angle/2.0)))*dist;
    dz = sin(ROBOT_DEG2RAD(pose.rot.pitch))*dist + altDist;

    pose.loc.x += dx;
    pose.loc.y += dy;
    pose.loc.z += dz;
    pose.rot.yaw += angle;
    ROBOT_CRAMPDEG(pose.rot.yaw, 0.0, 360.0);
    pose.loc.extra.time = curTime;
    pose.rot.extra.time = curTime;
    pose.extra.time = curTime;

    shPose_ = pose;

    pthread_mutex_unlock(&poseMutex_);

    if ((gPoseCalc != NULL) && (poseCalcID_ >= 0))
    {
        gPoseCalc->updateModulePose(poseCalcID_, pose);
    }
}

//-----------------------------------------------------------------------
// This function updates the pose based on the specified speed and delta
// time .
//-----------------------------------------------------------------------
void Robot::advancePoseAltitude_(HSPose_t pose, HSSpeed_t speed, double deltaTimeSec)
{
    double dist, angle, altDist, dx, dy, dz;
    double curTime;

    dist  = speed.driveSpeed.value*deltaTimeSec;
    angle = speed.steerSpeed.value*deltaTimeSec;
    altDist = speed.altitudeSpeed.value*deltaTimeSec;
    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&poseMutex_);

    dz = sin(ROBOT_DEG2RAD(pose.rot.pitch))*dist + altDist;
    pose.loc.z += dz;
    pose.loc.extra.time = curTime;
    pose.extra.time = curTime;

    shPose_ = pose;

    pthread_mutex_unlock(&poseMutex_);

    if ((gPoseCalc != NULL) && (poseCalcID_ >= 0))
    {
        gPoseCalc->updateModulePose(poseCalcID_, pose);
    }
}

//-----------------------------------------------------------------------
// This function sets the pose to be zero.
//-----------------------------------------------------------------------
void Robot::resetPose_(void)
{
    HSPose_t pose;
    double curTime;
    
    curTime = getCurrentEpochTime();

    memset((HSPose_t *)&pose, 0x0, sizeof(HSPose_t));
    pose.extra.time = curTime;
    pose.loc.extra.time = curTime;
    pose.loc.extra.time = curTime;

    setPose_(pose);
}

//-----------------------------------------------------------------------
// This function sets the desired rotation of the robot with respect to
// the world coordinate (Note: East = 0).
//-----------------------------------------------------------------------
void Robot::setDesiredRotation(HSRotation_t rot)
{
    pthread_mutex_lock(&desiredRotationMutex_);
    shDesiredRotation_ = rot;
    pthread_mutex_unlock(&desiredRotationMutex_);
}

//-----------------------------------------------------------------------
// This function returns the desired rotation of the robot with respect to
// the world coordinate (Note: East = 0).
//-----------------------------------------------------------------------
void Robot::getDesiredRotation_(HSRotation_t &rot)
{
    pthread_mutex_lock(&desiredRotationMutex_);
    rot = shDesiredRotation_;
    pthread_mutex_unlock(&desiredRotationMutex_);
}

//-----------------------------------------------------------------------
// This function sets the desired speed. This speed may not be the exact
// speed that would be used to control the robot, depending on
// compControlSpeed_() below.
//-----------------------------------------------------------------------
void Robot::setDesiredSpeed(HSSpeed_t speed)
{
    double curTime;

    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&desiredSpeedMutex_);
    shDesiredSpeed_ = speed;
    pthread_mutex_unlock(&desiredSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function sets the desired drive speed. This speed may not be
// the exact speed that would be used to control the robot, depending
// on compControlSpeed_() below.
//-----------------------------------------------------------------------
void Robot::setDesiredDriveSpeed(double driveSpeed)
{
    double curTime;

    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&desiredSpeedMutex_);
    shDesiredSpeed_.driveSpeed.value = driveSpeed;
    shDesiredSpeed_.driveSpeed.extra.time = curTime;
    shDesiredSpeed_.extra.time = curTime;
    pthread_mutex_unlock(&desiredSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function sets the desired steer speed. This speed may not be
// the exact speed that would be used to control the robot, depending
// on compControlSpeed_() below.
//-----------------------------------------------------------------------
void Robot::setDesiredSteerSpeed(double steerSpeed)
{
    double curTime;

    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&desiredSpeedMutex_);
    shDesiredSpeed_.steerSpeed.value = steerSpeed;
    shDesiredSpeed_.steerSpeed.extra.time = curTime;
    shDesiredSpeed_.extra.time = curTime;
    pthread_mutex_unlock(&desiredSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function sets the desired altitude speed. This speed may not be
// the exact speed that would be used to control the robot, depending
// on compControlSpeed_() below.
//-----------------------------------------------------------------------
void Robot::setDesiredAltitudeSpeed(double altitudeSpeed)
{
    double curTime;

    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&desiredSpeedMutex_);
    shDesiredSpeed_.altitudeSpeed.value = altitudeSpeed;
    shDesiredSpeed_.altitudeSpeed.extra.time = curTime;
    shDesiredSpeed_.extra.time = curTime;
    pthread_mutex_unlock(&desiredSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function sets the desired arm speed. This speed may not be
// the exact speed that would be used to control the robot, depending
// on compControlSpeed_() below.
//-----------------------------------------------------------------------
void Robot::setDesiredArmSpeed(double armSpeed)
{
    double curTime;

    curTime = getCurrentEpochTime();

    pthread_mutex_lock(&desiredSpeedMutex_);
    shDesiredSpeed_.armSpeed.value = armSpeed;
    shDesiredSpeed_.armSpeed.extra.time = curTime;
    shDesiredSpeed_.extra.time = curTime;
    pthread_mutex_unlock(&desiredSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function returns the desired speed. This speed may not be
// the exact speed that would be used to control the robot, depending
// on compControlSpeed_() below.
//-----------------------------------------------------------------------
void Robot::getDesiredSpeed_(HSSpeed_t &desiredSpeed)
{
    pthread_mutex_lock(&desiredSpeedMutex_);
    desiredSpeed = shDesiredSpeed_;
    pthread_mutex_unlock(&desiredSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function computes the input speed of the robot that would be
// commanded.
//-----------------------------------------------------------------------
void Robot::compControlSpeed_(void)
{
    static bool first = true;
    HSSpeed_t desiredSpeed;
    HSRotation_t currentRot;
    HSRotation_t desiredRot;
    double curTime;
    double minPointTurnAngle, desiredHeading, angle, factor, sign;
    bool reversable = false;

    if (gPoseCalc == NULL)
    {
        return;
    }

    curTime = getCurrentEpochTime();
    getReversableDrive_(reversable);
    getMinimumPointTurnAngle_(minPointTurnAngle);
    getDesiredSpeed_(desiredSpeed);
    getDesiredRotation_(desiredRot);

    gPoseCalc->getRotation(currentRot);
    desiredHeading = desiredRot.yaw - currentRot.yaw;
    ROBOT_CRAMPDEG(desiredHeading, -180.0, 180.0);

//printfTextWindow("Desired heading: %lf (minpoint=%lf)\n", desiredHeading, minPointTurnAngle);

    if (UsingRovio) {
	if (desiredSpeed.steerSpeed.value > 0) {
           //printfTextWindow("Rovio: Ang=33");  
	   desiredang = 33;
        }
        else if (desiredSpeed.steerSpeed.value < 0) {
           //printfTextWindow("Rovio: Ang=-33");  
           desiredang = -33;
        }
        else {

            if (donerotate) {
	      desiredang = 0;
            }
            else {
              desiredang = desiredHeading;
            }
            //printfTextWindow("Rovio: Ang=%f (donrotate=%d)", desiredang, donerotate);  
	}
    }

    // If both desired rotation and desired speed was specified long
    // time ago. Set the control speed (both driving and steering) to
    // be zero. If the desired rotation is requested after desired 
    // steering speed is specified, compute the control steering speed 
    // based on the rotation.

    // rovio special case cause it it handles motor commands differently
    if ((UsingRovio) && (desiredRot.extra.time < (curTime - maxNoCommandTimeSec_))) {
      if (desiredRot.extra.time < (curTime - maxNoCommandTimeSec_)) {

        if (fabs(desiredSpeed.steerSpeed.value) > 0.1) {
          desiredSpeed.steerSpeed.value = 0;
          desiredSpeed.driveSpeed.value = 0;
          desiredSpeed.altitudeSpeed.value = 0;
        }
        else {
          if (desiredSpeed.extra.time < (curTime - 2)) {
            desiredSpeed.steerSpeed.value = 0;
            desiredSpeed.driveSpeed.value = 0;
            desiredSpeed.altitudeSpeed.value = 0;
          } 
        }     
      }
    }
    else if ((!UsingRovio) && (desiredRot.extra.time < (curTime - maxNoCommandTimeSec_)) &&
        (desiredSpeed.extra.time < (curTime - maxNoCommandTimeSec_)))
    {
        desiredSpeed.driveSpeed.value = 0;
        desiredSpeed.steerSpeed.value = 0;
        desiredSpeed.altitudeSpeed.value = 0;
    }
    else if (desiredRot.extra.time > (desiredSpeed.steerSpeed.extra.time))
    {
        gPoseCalc->getRotation(currentRot);
        desiredHeading = desiredRot.yaw - currentRot.yaw;
        ROBOT_CRAMPDEG(desiredHeading, -180.0, 180.0);
	desiredang = desiredHeading;

        if (reversable)
        {
            if (fabs(desiredHeading) > 90.0)
            {
                desiredHeading += 180.0;
                desiredSpeed.driveSpeed.value *= -1.0;
            }
        }

        ROBOT_CRAMPDEG(desiredHeading, -180.0, 180.0);

        if (fabs(desiredHeading) >= minPointTurnAngle)
        {
            desiredSpeed.driveSpeed.value = 0;
        }

        // Calculate altitude speed
        // Note: This assumes altitude movement is independent of movement on 2D plane (e.g. for a rotary wing craft)
        if (desiredSpeed.altitudeSpeed.value == 0.0 && desiredSpeed.steerSpeed.value != 0.0)
          desiredSpeed.altitudeSpeed.value = sin(ROBOT_RAD2DEG(desiredRot.pitch) / desiredSpeed.steerSpeed.value);
	if (fabs(desiredSpeed.altitudeSpeed.value) > maxAltitudeSpeed_) {
          sign = (desiredSpeed.altitudeSpeed.value < 0 ? -1 : 1);
          desiredSpeed.altitudeSpeed.value = sign * maxAltitudeSpeed_;
        }

        // Determine the steering speed based on the sine function of the
        // angle from the desired heading.
        sign = (desiredHeading >= 0)? 1.0 : -1.0;
        angle = fabs(desiredHeading);
        angle = (angle > maxSteerSpeedAngle_)? maxSteerSpeedAngle_ : angle;
        factor = sin(0.5*M_PI*angle/maxSteerSpeedAngle_);
        desiredSpeed.steerSpeed.value = sign * factor* maxSteerSpeed_;

        if (fabs(desiredSpeed.driveSpeed.value) > maxSpeed_) {
          desiredSpeed.driveSpeed.value = desiredSpeed.driveSpeed.value<0?-maxSpeed_:maxSpeed_;
        }

	if (fabs(desiredSpeed.steerSpeed.value) < minSteerSpeed_)
	{
	    if (fabs(angle) >= minPointTurnAngle)
	    {
	        desiredSpeed.driveSpeed.value = 0;
		printfTextWindow("Warning: Minimum point-turn angle (%f) too small.\n", minPointTurnAngle);
	    }
	    else
	    {
	        desiredSpeed.steerSpeed.value = 0.0;
	    }
	}
    }
    else
    {
        // Use the desired speed as the control speed.

        // Clamp the speed to a maximum
        if (fabs(desiredSpeed.driveSpeed.value) > maxSpeed_) {
           desiredSpeed.driveSpeed.value = desiredSpeed.driveSpeed.value<0?-maxSpeed_:maxSpeed_;
        }

        // Calculate altitude speed
        // Note: This assumes altitude movement is independent of movement on 2D plane (e.g. for a rotary wing craft)
        if (desiredSpeed.altitudeSpeed.value == 0.0 && desiredSpeed.steerSpeed.value != 0.0)
          desiredSpeed.altitudeSpeed.value = sin(ROBOT_RAD2DEG(desiredRot.pitch) / desiredSpeed.steerSpeed.value);
	if (fabs(desiredSpeed.altitudeSpeed.value) > maxAltitudeSpeed_) {
          sign = (desiredSpeed.altitudeSpeed.value < 0 ? -1 : 1);
          desiredSpeed.altitudeSpeed.value = sign * maxAltitudeSpeed_;
        }

    }

    pthread_mutex_lock(&controlSpeedMutex_);
    shControlSpeed_ = desiredSpeed;
    pthread_mutex_unlock(&controlSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function returns the input speed of the robot that would be
// commanded.
//-----------------------------------------------------------------------
void Robot::getControlSpeed_(HSSpeed_t &speed)
{
    pthread_mutex_lock(&controlSpeedMutex_);
    speed = shControlSpeed_;
    pthread_mutex_unlock(&controlSpeedMutex_);
}

//-----------------------------------------------------------------------
// This function enables (or disables) the revesable mode of the robot.
//-----------------------------------------------------------------------
void Robot::setReversableDrive(bool reversable)
{
    pthread_mutex_lock(&reversableDriveMutex_);
    shReversableDrive_ = reversable;
    pthread_mutex_unlock(&reversableDriveMutex_);
}

//-----------------------------------------------------------------------
// This function returns the status of the revesable mode of the robot.
//-----------------------------------------------------------------------
void Robot::getReversableDrive_(bool &reversable)
{
    pthread_mutex_lock(&reversableDriveMutex_);
    reversable = shReversableDrive_;
    pthread_mutex_unlock(&reversableDriveMutex_);
}

//-----------------------------------------------------------------------
// This function sets the minimum angle that forces the robot to make
// a point turn.
//-----------------------------------------------------------------------
void Robot::setMinimumPointTurnAngle(double angle)
{
    pthread_mutex_lock(&minPointTurnAngleMutex_);
    shMinPointTurnAngle_ = angle;
    pthread_mutex_unlock(&minPointTurnAngleMutex_);
}

//-----------------------------------------------------------------------
// This function returns the minimum angle that forces the robot to make
// a point turn.
//-----------------------------------------------------------------------
void Robot::getMinimumPointTurnAngle_(double &angle)
{
    pthread_mutex_lock(&minPointTurnAngleMutex_);
    angle = shMinPointTurnAngle_;
    pthread_mutex_unlock(&minPointTurnAngleMutex_);
}

//-----------------------------------------------------------------------
// This function stops the robot by setting the drive and steer speeds
// to be zero.
//-----------------------------------------------------------------------
void Robot::stop(void)
{
    setDesiredDriveSpeed(0);
    setDesiredSteerSpeed(0);
    setDesiredAltitudeSpeed(0);
    setDesiredArmSpeed(0);
}

//-----------------------------------------------------------------------
// This function computes the velocity of the robot.
//-----------------------------------------------------------------------
void Robot::updateVelocity_(void)
{
    HSPose_t pose, lastPose, delta;
    HSVelocity_t aveVel;
    double angle, diffAngle;
    double frequency;
    double curTime;
    double sum;
    double transSpeed;

    curTime = getCurrentEpochTime();

    memset((HSVelocity_t *)&aveVel, 0x0, sizeof(HSVelocity_t));
    aveVel.extra.time = curTime;
    aveVel.transVel.extra.time = curTime;
    aveVel.angVel.extra.time = curTime;

    // Get the frequency.
    getFrequency_(frequency);

    if (frequency == 0)
    {
        // The velocity not yet ready to be computed.
        return;
    }

    if (!(velBuffers_.initialized))
    {
        velBuffers_.xBuffer.resize(numVelocitySamples_);
        velBuffers_.yBuffer.resize(numVelocitySamples_);
        velBuffers_.zBuffer.resize(numVelocitySamples_);
        velBuffers_.yawBuffer.resize(numVelocitySamples_);
        velBuffers_.pitchBuffer.resize(numVelocitySamples_);
        velBuffers_.rollBuffer.resize(numVelocitySamples_);
        velBuffers_.initialized = true;
        velIndex_ = 0;
        velBufferFilled_ = false;
    }

    // Copy the current and previous poses.
    pthread_mutex_lock(&poseMutex_);
    pose = shPose_;
    lastPose = shLastPose_;
    pthread_mutex_unlock(&poseMutex_);

    // Get the differences.
    delta.loc.x = pose.loc.x - lastPose.loc.x;
    delta.loc.y = pose.loc.y - lastPose.loc.y;
    delta.loc.z = pose.loc.z - lastPose.loc.z;
    delta.rot.yaw = pose.rot.yaw - lastPose.rot.yaw;
    delta.rot.pitch = pose.rot.pitch - lastPose.rot.pitch;
    delta.rot.roll = pose.rot.roll - lastPose.rot.roll;

    // Assumption: No robot will turn more than 180 degree per cycle.
    // This assumption is necessary as there is no way to tell whether
    // the robot turned 270 degree or - 90 degree by just comparing
    // the yaw value.
    ROBOT_CRAMPDEG(delta.rot.yaw, -180.0, 180.0);
    ROBOT_CRAMPDEG(delta.rot.pitch, -180.0, 180.0);
    ROBOT_CRAMPDEG(delta.rot.roll, -180.0, 180.0);

    // Compute the current velocity.
    velBuffers_.xBuffer[velIndex_] = (delta.loc.x)*frequency;
    velBuffers_.yBuffer[velIndex_] = (delta.loc.y)*frequency;
    velBuffers_.zBuffer[velIndex_] = (delta.loc.z)*frequency;
    velBuffers_.yawBuffer[velIndex_] = (delta.rot.yaw)*frequency;
    velBuffers_.pitchBuffer[velIndex_] = (delta.rot.pitch)*frequency;
    velBuffers_.rollBuffer[velIndex_] = (delta.rot.roll)*frequency;
    velIndex_++;

    if (velIndex_ >= numVelocitySamples_)
    {
        velBufferFilled_ = true;
        velIndex_ = 0;
    }

    // Compute the average velocity over the buffers.
    if (velBufferFilled_)
    {
        sum = accumulate(
            velBuffers_.xBuffer.begin(),
            velBuffers_.xBuffer.end(),
            0.0);
        aveVel.transVel.x = sum/(double)numVelocitySamples_;

        sum = accumulate(
            velBuffers_.yBuffer.begin(),
            velBuffers_.yBuffer.end(),
            0.0);
        aveVel.transVel.y = sum/(double)numVelocitySamples_;

        sum = accumulate(
            velBuffers_.zBuffer.begin(),
            velBuffers_.zBuffer.end(),
            0.0);
        aveVel.transVel.z = sum/(double)numVelocitySamples_;

        sum = accumulate(
            velBuffers_.yawBuffer.begin(),
            velBuffers_.yawBuffer.end(),
            0.0);
        aveVel.angVel.yaw = sum/(double)numVelocitySamples_;

        sum = accumulate(
            velBuffers_.pitchBuffer.begin(),
            velBuffers_.pitchBuffer.end(),
            0.0);
        aveVel.angVel.pitch = sum/(double)numVelocitySamples_;

        sum = accumulate(
            velBuffers_.rollBuffer.begin(),
            velBuffers_.rollBuffer.end(),
            0.0);
        aveVel.angVel.roll = sum/(double)numVelocitySamples_;
    }
    else
    {
        sum = accumulate(
            velBuffers_.xBuffer.begin(),
            (velBuffers_.xBuffer.begin()) + (velIndex_-1),
            0.0);
        aveVel.transVel.x = sum/(double)velIndex_;

        sum = accumulate(
            velBuffers_.yBuffer.begin(),
            (velBuffers_.yBuffer.begin()) + (velIndex_-1),
            0.0);
        aveVel.transVel.y = sum/(double)velIndex_;

        sum = accumulate(
            velBuffers_.zBuffer.begin(),
            (velBuffers_.zBuffer.begin()) + (velIndex_-1),
            0.0);
        aveVel.transVel.z = sum/(double)velIndex_;

        sum = accumulate(
            velBuffers_.yawBuffer.begin(),
            (velBuffers_.yawBuffer.begin()) + (velIndex_-1),
            0.0);
        aveVel.angVel.yaw = sum/(double)velIndex_;

        sum = accumulate(
            velBuffers_.pitchBuffer.begin(),
            (velBuffers_.pitchBuffer.begin()) + (velIndex_-1),
            0.0);
        aveVel.angVel.pitch = sum/(double)velIndex_;

        sum = accumulate(
            velBuffers_.rollBuffer.begin(),
            (velBuffers_.rollBuffer.begin()) + (velIndex_-1),
            0.0);
        aveVel.angVel.roll = sum/(double)velIndex_;
    }

    // Now, check for the backwardness. First, get the angle that the
    // system moved.
    transSpeed = sqrt(pow(aveVel.transVel.x, 2)+pow(aveVel.transVel.y, 2));
    if (transSpeed > stoppingSpeedThreshold_)
    {
        angle = ROBOT_RAD2DEG(atan2(aveVel.transVel.y, aveVel.transVel.x));

        // Get the difference between those two.
        ROBOT_CRAMPDEG(pose.rot.yaw, 0.0, 360.0);
        ROBOT_CRAMPDEG(angle, 0.0, 360.0);
        diffAngle =  angle - pose.rot.yaw;
        ROBOT_CRAMPDEG(diffAngle, -180.0, 180.0);

        // It is moving backward if the difference is more than 90 degrees.
        aveVel.extra.backward = (fabs(diffAngle) > 90.0)? true : false;
    }
    
    setVelocity(aveVel);
}

//-----------------------------------------------------------------------
// This function updates the velocity info.
//-----------------------------------------------------------------------
void Robot::setVelocity(HSVelocity_t vel)
{
    pthread_mutex_lock(&velocityMutex_);
    shVelocity_ = vel;
    pthread_mutex_unlock(&velocityMutex_);
}

void Robot::setMaxSpeed(double val) {
  maxSpeed_ = val;
}
void Robot::setMaxSteerSpeed(double val) {
  maxSteerSpeed_ = val;
}
void Robot::setMaxSteerSpeedAngle(double val) {
  maxSteerSpeedAngle_ = val;
}
void Robot::setMaxAltitudeSpeed(double val) {
  maxAltitudeSpeed_ = val;
}
void Robot::setMaxNoCommandTimeSec(double val) {
  maxNoCommandTimeSec_ = val;
}
void Robot::setMinSteerSpeed(double val) {
  minSteerSpeed_ = val;
}
void Robot::setStoppingSpeedThreshold(double val) {
  stoppingSpeedThreshold_ = val;
}
void Robot::setNumVelocitySamples(double val) {
  numVelocitySamples_ = val;
}

//-----------------------------------------------------------------------
// This function retrieves the velocity info.
//-----------------------------------------------------------------------
void Robot::getVelocity(HSVelocity_t &vel)
{
    pthread_mutex_lock(&velocityMutex_);
    vel = shVelocity_;
    pthread_mutex_unlock(&velocityMutex_);
}

//-----------------------------------------------------------------------
// This function sets the battery voltage.
//-----------------------------------------------------------------------
void Robot::setBatteryVoltage_(double batteryVoltage)
{
    pthread_mutex_lock(&batteryVoltageMutex_);
    shBatteryVoltage_ = batteryVoltage;
    pthread_mutex_unlock(&batteryVoltageMutex_);
}

//-----------------------------------------------------------------------
// This function retrieves the battery voltage.
//-----------------------------------------------------------------------
void Robot::getBatteryVoltage_(double &batteryVoltage)
{
    pthread_mutex_lock(&batteryVoltageMutex_);
    batteryVoltage = shBatteryVoltage_;
    pthread_mutex_unlock(&batteryVoltageMutex_);
}

//-----------------------------------------------------------------------
// This function logs the data.
//-----------------------------------------------------------------------
void Robot::logData_(double logTime)
{
    HSPose_t pose;
    HSSpeed_t speed;
    double batteryVoltage, dist;

    if (logfile_ != NULL)
    {
        getPose(pose);
        getControlSpeed_(speed);
        getBatteryVoltage_(batteryVoltage);
        getTravelDist(dist);

        fprintf(
            logfile_,
            "%f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %f %f %f %.2f %.2f %.2f %f %.2f\n",
            logTime,
            batteryVoltage,
            pose.loc.x,
            pose.loc.y,
            pose.loc.z,
            pose.rot.yaw,
            pose.rot.pitch,
            pose.rot.roll,
            pose.extra.time,
            pose.loc.extra.time,
            pose.rot.extra.time,
            speed.driveSpeed.value,
            speed.steerSpeed.value,
            speed.armSpeed.value,
            speed.extra.time,
            dist);

        fflush(logfile_);
    }
}

/**********************************************************************
# $Log: robot.c,v $
# Revision 1.11  2010-05-03 22:10:09  zkira
# Cricket sensor, blob tracking, rovio (inc. IR sensor), skybotix interface (unfinished)
#
# Revision 1.9  2009/04/17 17:51:48  zkira
# Added maxSpeed in robot (clamps speed to this max value), and ability to speed up usarsim data acquisition so that speeding up the USARSim simulator becomes possible
#
# Revision 1.8  2009/04/13 02:58:43  zkira
# RFID code
#
# Revision 1.7  2009/03/02 18:52:47  zkira
# Made lower level robot movement parameters variables so that they can be modified externally
#
# Revision 1.6  2008/12/25 11:13:01  ts2883
# Yet another set of new parameters for movement; by decreasing the speed and
# angle of the rotational movement of the robot, the spinning problem when
# doing the 3D obstacle avoidance behavior is minimized.
#
# Revision 1.5  2008/10/29 02:24:52  ts2883
# New parameters for movement
#
# Revision 1.4  2008/10/27 21:44:47  zkira
# Changed movement parameters to allow movement in a building
# Added advancePoseAltitude to allow for control-signal based crude odometry (not used currently)
#
# Revision 1.3  2008/09/09 20:20:07  zkira
# Crude altitude control: Added altitudeSpeed & z component
#
# Revision 1.2  2008/07/15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.2  2007/04/09 17:36:10  endo
# TravelDist and its logging added.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2004/09/10 19:41:07  endo
# New PoseCalc integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.9  2003/04/02 22:25:20  zkira
# Added battery level variable to class
#
# Revision 1.8  2002/01/16 22:29:42  ernest
# Added flag  m_bposition_set.
#
# Revision 1.7  2001/12/22 16:35:24  endo
# RH 7.1 porting.
#
# $Log: robot.c,v $
# Revision 1.11  2010-05-03 22:10:09  zkira
# Cricket sensor, blob tracking, rovio (inc. IR sensor), skybotix interface (unfinished)
#
# Revision 1.9  2009/04/17 17:51:48  zkira
# Added maxSpeed in robot (clamps speed to this max value), and ability to speed up usarsim data acquisition so that speeding up the USARSim simulator becomes possible
#
# Revision 1.8  2009/04/13 02:58:43  zkira
# RFID code
#
# Revision 1.7  2009/03/02 18:52:47  zkira
# Made lower level robot movement parameters variables so that they can be modified externally
#
# Revision 1.6  2008/12/25 11:13:01  ts2883
# Yet another set of new parameters for movement; by decreasing the speed and
# angle of the rotational movement of the robot, the spinning problem when
# doing the 3D obstacle avoidance behavior is minimized.
#
# Revision 1.5  2008/10/29 02:24:52  ts2883
# New parameters for movement
#
# Revision 1.4  2008/10/27 21:44:47  zkira
# Changed movement parameters to allow movement in a building
# Added advancePoseAltitude to allow for control-signal based crude odometry (not used currently)
#
# Revision 1.3  2008/09/09 20:20:07  zkira
# Crude altitude control: Added altitudeSpeed & z component
#
# Revision 1.2  2008/07/15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.2  2007/04/09 17:36:10  endo
# TravelDist and its logging added.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2004/09/10 19:41:07  endo
# New PoseCalc integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.9  2003/04/02 22:25:20  zkira
# Added battery level variable to class
#
# Revision 1.8  2002/01/16 22:29:42  ernest
# Added flag  m_bposition_set.
#
# Revision 1.6  2001/05/29 22:35:15  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.5  2001/03/23 21:32:03  blee
# altered to use a config file
#
# Revision 1.4  2000/12/12 23:29:29  blee
# Changed #defines for the user interface to enumerated types.
#
# Revision 1.3  2000/11/13 20:09:23  endo
# ARTV-Jr class added to HServer.
#
# Revision 1.2  2000/09/19 03:49:46  endo
# RCS log added.
#
#
#**********************************************************************/
