/**********************************************************************
 **                                                                  **
 **                         rovio-robot.c                            **
 **                                                                  **
 **                                                                  **
 **  Written by: Zsolt Kira                                          **
 **  Based on code written by: Yoichiro Endo                         **
 **                                                                  **
 **  Interface for Rovio robot  .                                    **
 **                                                                  **
 **  Copyright 2009, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: rovio-robot.c,v 1.4 2010-05-03 22:10:10 zkira Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <netdb.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "rovio-robot.h"
#include "Watchdog.h"
#include "LogManager.h"

#include "PoseCalculatorTypes.h"
#include "PoseCalculator.h"
#include "Kalman/mxn_matrix.h"
#include "Kalman/ekf_localize.h"

//-----------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------
extern pthread_mutex_t cricketPositionMutex;
extern HSPose_t cricketPosition;
extern bool cricketGotValidYaw;
extern bool cricketEnabled;

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t RovioRobot::latestRawPoseMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t RovioRobot::sensorDataTableMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t RovioRobot::socketMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t RovioRobot::rovioPoseOffsetMutex_ = PTHREAD_MUTEX_INITIALIZER;

const string RovioRobot::DEFAULT_ROVIO_ROBOT_NAME_ = "rovio_bot";
const string RovioRobot::DEFAULT_ROVIO_ROBOT_CLASS_ = "Rovio";
const string RovioRobot::DEFAULT_ROVIO_SERVER_HOST_MACHINE_ = "127.0.0.1";
const string RovioRobot::EMPTY_STRING_ = "";
const int RovioRobot::DEFAULT_ROVIO_SERVER_PORT_ = 3000;
const int RovioRobot::SKIP_STATUSBAR_UPDATE_ = 10;
const int RovioRobot::WATCHDOG_CHECK_INTERVAL_SEC_ = 2;
const int RovioRobot::MAIN_THREAD_USLEEP_ = 0; //500/4;
const int RovioRobot::READER_THREAD_USLEEP_ = 10;

const float rovio_range_angle_IR[] = { 0.0, 0.0 };
bool useIR = true;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const RovioRobot::RobotSpec_t RovioRobot::ROBOT_SPECS_[NUM_ROVIO_ROBOT_CLASSES] =
{
    // Note: The robot's Z-axis in Rovio points opposite of Z-axis in MissionLab.
    {
        ROVIO_ROBOT_CLASS_DEFAULT,
        "Rovio", // className
        0.5, // robotRadius
        0.1 // wheelRadius
    }
};

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
RovioRobot::RovioRobot(
    Robot** a,
    char *serverHostMachine,
    char *rovioRobotName,
    char *rovioRobotClass,
    int serverPort,
    float initX,
    float initY,
    float initZ,
    float initYaw,
    float initPitch,
    float initRoll) :
    Robot(a, HS_ROBOT_TYPE_ROVIO, HS_MODULE_NAME_ROBOT_ROVIO),
    partXyt_(NULL),
    partIr(NULL),
    sensorIr(NULL),
    initX_(initX),
    initY_(initY),
    initZ_(initZ),
    initYaw_(initYaw),
    initPitch_(initPitch),
    initRoll_(initRoll),
    robotSpecIndex_(-1),
    serverPort_(DEFAULT_ROVIO_SERVER_PORT_),
    mainThreadIsUp_(false),
    readerThreadIsUp_(false)
{
    static int rovio_number = 0;

    rovio_number++;
    rovioNumber = rovio_number;

    if (strstr( (char *) serverHostMachine_.c_str(), "130") == NULL)  {
      rovioNumber = 2;
    }

    numIR = ROVIO_NUM_IR;

    serverHostMachine_ = (serverHostMachine != NULL)?
        serverHostMachine : DEFAULT_ROVIO_SERVER_HOST_MACHINE_;

    rovioRobotName_ = (rovioRobotName != NULL)?
        rovioRobotName : DEFAULT_ROVIO_ROBOT_NAME_;

    rovioRobotClass_ = (rovioRobotClass != NULL)?
        rovioRobotClass : DEFAULT_ROVIO_ROBOT_CLASS_;

    serverPort_ = (serverPort_ >= 0)?
        serverPort : DEFAULT_ROVIO_SERVER_PORT_;

    pthread_mutex_init(&sensorDataTableMutex_, NULL);
    pthread_mutex_init(&socketMutex_, NULL);
    pthread_mutex_init(&rovioPoseOffsetMutex_, NULL);
    pthread_mutex_init(&latestRawPoseMutex_, NULL);

    resetPose_();
    updateStatusBar_();

    if (!initializeRovio_())
    {
        printfTextWindow("Error: RovioRobot could not be initialized. Exiting Rovio.\n");
        delete this;
        return;
    }

    addSensors_();

    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);
    pthread_create(&mainThread_, NULL, &startMainThread_, (void*)this);

    if (0&&gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(NAME_, WATCHDOG_CHECK_INTERVAL_SEC_, getpid(), mainThread_);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
RovioRobot::~RovioRobot(void)
{
    if (mainThreadIsUp_)
    {
        pthread_cancel(mainThread_);
        pthread_join(mainThread_, NULL);
    }

    if (partIr != NULL)
    {
        delete partIr;
        partIr = NULL;
    }

    if (readerThreadIsUp_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    delete partXyt_;
}

//-----------------------------------------------------------------------
// This function initialize rovio
//-----------------------------------------------------------------------
bool RovioRobot::initializeRovio_(void)
{
    int i;

    rovioInterface = new Rovio( strdup(serverHostMachine_.c_str()) );

    for (i = 0; i < NUM_ROVIO_ROBOT_CLASSES; i++)
    {
        if (ROBOT_SPECS_[i].className == rovioRobotClass_)
        {
            robotSpecIndex_ = i;
            break;
        }
    }

    if (robotSpecIndex_ < 0)
    {
        robotSpecIndex_ = 0;

        printfTextWindow(
            "Warning: RovioRobot::%s(). Robot spect not found. Using %s.\n",
            __FUNCTION__,
            ROBOT_SPECS_[robotSpecIndex_].className.c_str());
    }


    return true;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void RovioRobot::updateStatusBar_(void)
{
    HSPose_t pose;
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;

    if (mainThreadIsUp_)
    {
        strcat(statusStr, "RovioRobot [");

        getPose(pose);

        sprintf(
            buf,
            "%s]: %c x: %.2f  y: %.2f z: %.2f t: %.2f pitch:%.2f",
            rovioRobotName_.c_str(),
            statusbarSpinner_->getStatus(),
            pose.loc.x,
            pose.loc.y,
            pose.loc.z,
            pose.rot.yaw,
            pose.rot.pitch);

        if (DISPLAY_FREQUENCY)
        {
            getFrequency_(frequency);
            sprintf(buf, "%s : %.2f Hz", buf, frequency);
        }
    }
    else
    {
        sprintf(
            buf,
            "RovioRobot: Connecting to the server on %s [port %d]...",
            serverHostMachine_.c_str(),
            serverPort_);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function starts the main thread.
//-----------------------------------------------------------------------
void *RovioRobot::startMainThread_(void *rovioInstance)
{
    ((RovioRobot*) rovioInstance)->mainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function starts the reader thread.
//-----------------------------------------------------------------------
void *RovioRobot::startReaderThread_(void *rovioInstance)
{
    ((RovioRobot*) rovioInstance)->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to move RovioRobot.
//-----------------------------------------------------------------------
void RovioRobot::mainLoop_(void)
{
    int count = 0;
    
    mainThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();
//printfTextWindow("%lf: 1\n", getCurrentEpochTime());
        // Read the pose data from the rovio.
        updatePose_();
//printfTextWindow("%lf: 2\n", getCurrentEpochTime());

        // Check for the pthread termination again.
        pthread_testcancel();

        // Compute the control speed.
        compControlSpeed_();
//printfTextWindow("%lf: 3\n", getCurrentEpochTime());

        // Control the robot.
        controlRobot_();
//printfTextWindow("%lf: 4\n", getCurrentEpochTime());

	//rovioInterface->GetImage();

        // Compute the frequency.
        compFrequency_();

//printfTextWindow("%lf: 4\n", getCurrentEpochTime());
        // Compute the velocity.
        updateVelocity_();

	//	CurFrame = rovioInterface->GetImage();

//printfTextWindow("%lf: 4\n", getCurrentEpochTime());

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (gLogManager != NULL)
        {
            manageLog_(readerThread_);
        }

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        // Ease the computational load.
        usleep(MAIN_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function loops around to receive the sensor from the server.
//-----------------------------------------------------------------------
void RovioRobot::readerLoop_(void)
{
    readerThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();
	//printfTextWindow("Readerloop");

        // Read the sensor data.
        if (rovioInterface != NULL) {
		rovioInterface->PeriodicUpdate();

	}

        // Check for the pthread termination again.
        pthread_testcancel();
	//printfTextWindow("Done Readerloop");

        // Ease the computational load.
        usleep(READER_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function waits for the sensor data to arrive.
//-----------------------------------------------------------------------
void RovioRobot::waitForSensorData_(void)
{
    bool arrived = false;
    const struct timespec WAITER_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep

    while (true)
    {
        pthread_mutex_lock(&sensorDataTableMutex_);

        if (shSensorDataTable_.size() > 0)
        {
            arrived = true;
        }

        pthread_mutex_unlock(&sensorDataTableMutex_);

        if (arrived)
        {
            break;
        }

        nanosleep(&WAITER_SLEEP_TIME_NSEC, NULL);
    }
}

double normalizeAngle(double angle) {
  while (angle > 360) {
    angle -= 360;
  }

  while (angle < 0) {
    angle += 360;
  }

  return angle;

}

void RovioRobot::runKalman(double ox, double oy, double ot,  double &newx, double &newy, double &newtheta) {
  static EKF_Localize *localizeEKF = new EKF_Localize();
  static bool EKFInit = false;

  static PoseCalcFusedPose_t fusedPose;
  static PoseCalcPoseInfo_t poseData;
  static PoseCalcFusedPose_t LastPose;

  static bool gotValidYaw = false;
  static bool useCricket = false;

  static double LastOdomX = 0;
  static double LastOdomY = 0;
  static double LastOdomTheta = 0;

  static double OdomOffsetX = 0.0;
  static double OdomOffsetY = 0.0;
  static double OdomOffsetTheta = 0.0;

  static bool firstTime = true;
  double timestamp = getCurrentEpochTime(); 

  mxn_Matrix state;
  mxn_Matrix rot(2,2);
  mxn_Matrix od(2,1);
  mxn_Matrix roted(2,1);

  char debug_filename[10000];
  sprintf(debug_filename, "kalman%d-%f.log", rovioNumber, timestamp);
  static FILE *kdebug = fopen(debug_filename, "a");
 
  fprintf(kdebug, "%lf: Odometry: \t\t%.2lf, %.2lf, %.2lf\n", timestamp, ox, oy, ot);

  //printfTextWindow("Pose: \t%.2lf, %.2lf, %.2lf\t(Offset: %.2lf => %.2lf)\n", ox, oy, ot, OdomOffsetTheta, (ot+OdomOffsetTheta));
  ot = ot + OdomOffsetTheta;
  ot = normalizeAngle(ot);

  fprintf(kdebug, "%lf: OdometryOffset: \t%.2lf, %.2lf, %.2lf\n", timestamp, OdomOffsetX, OdomOffsetY, OdomOffsetTheta);

  // Initialize data.
  memset((PoseCalcFusedPose_t *)&fusedPose, 0x0, sizeof(PoseCalcFusedPose_t));
  fusedPose.fuserType = 0;  // FUSER_TYPE_;
  fusedPose.extra.validPose = false;
  fusedPose.extra.time = getCurrentEpochTime();

  // Get first odometry?
  if (firstTime) {
    fusedPose.pose.loc.x = ox;    
    fusedPose.pose.loc.y = oy;    
    fusedPose.pose.rot.yaw = ot;    

    // Initialize Kalman
    localizeEKF->Init(fusedPose);
    firstTime = false;
  }


  // Get first odometry readings?
  double dx, dy, dtheta, varianceloc, variancerot;

  // 02-23-10 CHANGED
  dx = ox - LastOdomX;
  dy = oy - LastOdomY;
  dtheta = ot - LastOdomTheta;
  //dtheta = normalizeAngle(dtheta);

  varianceloc = 0.1; //0.55; //0.1
  variancerot = 2.0;

  fprintf(kdebug, "%lf: OdometryDiff: \t%.2lf, %.2lf, %.2lf  \t(last: %.2lf,%.2lf,%.2lf) \t(variance: %.2lf, %.2lf)\n", timestamp, dx, dy, dtheta, LastOdomX, LastOdomY, LastOdomTheta, varianceloc, variancerot);

  // Update motion model
  localizeEKF->Update_MM(
      dx,
      dy,
      dtheta,
      varianceloc,
      POSECALC_DEG2RAD(variancerot));

  state = localizeEKF->Get_State();
  fprintf(kdebug, "%lf: State: \t\t%.2lf, %.2lf, %.2lf\n", timestamp, state[0][0], state[1][0], state[2][0]);

  // Get Cricket information
  pthread_mutex_lock(&cricketPositionMutex);
  poseData.pose.loc.x = cricketPosition.loc.x;
  poseData.pose.loc.y = cricketPosition.loc.y;
  poseData.pose.loc.z = cricketPosition.loc.z;
  poseData.pose.rot.yaw = cricketPosition.rot.yaw;
  gotValidYaw = cricketGotValidYaw;
  useCricket = cricketEnabled;
  pthread_mutex_unlock(&cricketPositionMutex);
  //printfTextWindow("R: Cricket yaw valid %d (%lf)\n", gotValidYaw, cricketPosition.rot.yaw);

  poseData.variance.locVariance.x = 1.0; //01; //0.3;
  poseData.variance.locVariance.y = 1.0; //01; //0.3;
  //poseData.variance.locVariance.z = 0.01; //100;

  // REMOVE 
  //gotValidYaw = false;
  //poseData.pose.rot.yaw = 0;

  // REMOVE Fixed values
  //gotValidYaw = true;
  //static double fixedyaw = 0.0;
  //poseData.variance.rotVariance.yaw = fixedyaw;
  //fixedyaw += 0.1;

  // Vary the variances depending one whether a yaw was calculated succesfully
  if (gotValidYaw) {
    poseData.variance.rotVariance.roll = 10000000;//10 //100; 
    poseData.variance.rotVariance.pitch = 1000000; //10//100;
    poseData.variance.rotVariance.yaw = 1;//10//1.0;
  } else {
    poseData.variance.rotVariance.roll = 100000; 
    poseData.variance.rotVariance.pitch = 100000;
    poseData.variance.rotVariance.yaw = 100000;
    poseData.pose.rot.yaw = 0.0;
  }
  poseData.sensorType = POSECALC_SENSOR_TYPE_CRICKET;

  poseData.pose.loc.extra.time = getCurrentEpochTime();
  poseData.pose.rot.extra.time = getCurrentEpochTime();

  // REMOVE DONT USE CRICKET
  //useCricket = false;
  if (1||rovioNumber == 1) {
    useCricket = true;
  } else {
    useCricket = false;
  }

  // Update sensor model
  if (useCricket) {
    fprintf(kdebug, "%lf: Cricket: \t\t%.2lf, %.2lf, %.2lf \t(variance: %.2lf,%.2lf,%.2lf)\t%d\n", timestamp, poseData.pose.loc.x, poseData.pose.loc.y, poseData.pose.rot.yaw, poseData.variance.locVariance.x, poseData.variance.locVariance.y, poseData.variance.rotVariance.yaw, gotValidYaw);
    //fprintf(kdebug, "{ %lf, %lf, %lf }, \n", poseData.pose.loc.x, poseData.pose.loc.y, poseData.pose.rot.yaw);

    // Update sensor model
    localizeEKF->Update_S(poseData, poseData.variance);
  } else {
    fprintf(kdebug, "Cricket: disabled\n");
  }

  // Get state
  fusedPose.extra.validPose = true;
  state = localizeEKF->Get_State();
  newx = fusedPose.pose.loc.x = state[0][0]; //state[0][0];
  newy = fusedPose.pose.loc.y = state[1][0]; //state[1][0];
  newtheta = fusedPose.pose.rot.yaw = state[2][0]; //state[2][0];
  newtheta += 90.0;
  newtheta = normalizeAngle(newtheta);

  LastPose = fusedPose;

  fprintf(kdebug, "%lf: State: \t\t%.2lf, %.2lf, %.2lf\n", timestamp, fusedPose.pose.loc.x, fusedPose.pose.loc.y, fusedPose.pose.rot.yaw);
  //state.Print_Matrix(kdebug);

  // REMOVE Ignore kalman, just set to current values!!
  //newx = fusedPose.pose.loc.x = cricketPosition.loc.x;
  //newy = fusedPose.pose.loc.x = cricketPosition.loc.y;
  //if (gotValidYaw) {
  //  newtheta = cricketPosition.rot.yaw;
  //  OdomOffsetTheta = -ot + newtheta;
  //  ot = ot + OdomOffsetTheta;
  //  ot = normalizeAngle(ot);
  //} else {
  //  // Use old one
  //  //newtheta = fusedPose.pose.rot.yaw = LastPose.pose.rot.yaw;
  //
  //  // Use odometry
  //  newtheta = ot;
  //}

  fprintf(kdebug, "%lf: FINAL: \t\t%.2lf, %.2lf, %.2lf\n\n", timestamp, newx, newy, newtheta);

  LastOdomX = ox;
  LastOdomY = oy;
  LastOdomTheta = ot;

  fflush(kdebug);

}

//-----------------------------------------------------------------------
// This function updates the current pose.
//-----------------------------------------------------------------------
void RovioRobot::updatePose_(void)
{
    HSPose_t pose, lastPose;
    HSSpeed_t controlSpeed;
    DataLine_t odoDataLine;
    map<string, DataLine_t>::iterator itr;
    string poseString = EMPTY_STRING_;
    vector<string> poseList;
    double dist;
    double dt, curTime;
    static double lastTime = 0;
    double x,y,theta;
    double kx,ky,ktheta;
    memset((HSPose_t *)&pose, 0x0, sizeof(HSPose_t));
    getLatestRawPose_(lastPose);

    switch (robotSpecIndex_) {

    case ROVIO_ROBOT_CLASS_DEFAULT:
        if (rovioInterface != NULL) rovioInterface->GetPose(x, y, theta);
        else { 
          x = lastPose.loc.x + 0.01;
          y = lastPose.loc.y + 0.02;
          theta = lastPose.rot.yaw + 0.04;
	  printfTextWindow("No rovio interface: Made-up data! (%f,%f,%f)\n", x, y, theta);
        }
        break;
      default:
	printfTextWindow("Error: Invalid robot type %d\n", robotSpecIndex_);
    }
 

    //printfTextWindow("Odometry: \t\t\t\t\t\t%.2lf, %.2lf, %.2lf\n", x, y, theta);

    runKalman(x, y, theta, kx, ky, ktheta);
    x = kx;
    y = ky;
    theta = ktheta;

    pose.loc.x = x;
    pose.loc.y = y;
    pose.rot.yaw = theta;

    dist = sqrt(pow((pose.loc.x - lastPose.loc.x), 2.0) + pow((pose.loc.y - lastPose.loc.y), 2.0) + pow((pose.loc.z - lastPose.loc.z), 2.0));
    addTravelDist_(dist);
    setLatestRawPose_(pose);

    updatePoseFromRovioPoseOffset_();

    // Update IR reading
    ir[0] = rovioInterface->getIR() ? 0.3 : 10000.0;  // 0.457 meters == 1.5 feet
    ir_loc[0][0] =  pose.loc.x;
    ir_loc[0][1] =  pose.loc.y;
    ir_loc[0][2] =  pose.rot.yaw;
}

//-----------------------------------------------------------------------
// This function sets the offset between the raw rovio data and the
// current robot pose.
//-----------------------------------------------------------------------
void RovioRobot::setRovioPoseOffset_(HSPose_t poseOffset)
{
    pthread_mutex_lock(&rovioPoseOffsetMutex_);
    shRovioPoseOffset_ = poseOffset;
    pthread_mutex_unlock(&rovioPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function returns the offset between the raw rovio data and the
// current robot pose.
//-----------------------------------------------------------------------
void RovioRobot::getRovioPoseOffset_(HSPose_t &poseOffset)
{
    pthread_mutex_lock(&rovioPoseOffsetMutex_);
    poseOffset = shRovioPoseOffset_;
    pthread_mutex_unlock(&rovioPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function saves the latest raw pose data.
//-----------------------------------------------------------------------
void RovioRobot::setLatestRawPose_(HSPose_t pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    shLatestRawPose_ = pose;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function returns the saved latest raw pose data.
//-----------------------------------------------------------------------
void RovioRobot::getLatestRawPose_(HSPose_t &pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    pose = shLatestRawPose_;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function updates the robot pose from the offset.
//-----------------------------------------------------------------------
void RovioRobot::updatePoseFromRovioPoseOffset_(void)
{
    HSPose_t pose;
    HSPose_t poseOffset;
    double x, y, z, angle, altAngle;

    getLatestRawPose_(pose);
    getRovioPoseOffset_(poseOffset);

    x = pose.loc.x;
    y = pose.loc.y;
    z = pose.loc.z;

    // REMOVED
    //angle = poseOffset.rot.yaw;
    //ROBOT_CRAMPDEG(angle, -180.0, 180.0);
    //angle = ROBOT_DEG2RAD(angle);

    //pose.loc.x = (x*cos(angle)) - (y*sin(angle)) + poseOffset.loc.x;
    //pose.loc.y = (x*sin(angle)) + (y*cos(angle)) + poseOffset.loc.y;
    //pose.rot.yaw += poseOffset.rot.yaw;
    //ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);
    //pose.rot.pitch += poseOffset.rot.pitch;
    //ROBOT_CRAMPDEG(pose.rot.pitch, -180.0, 180.0);

    setPose_(pose);
}

//-----------------------------------------------------------------------
// This function controls the robot with the control speed.
//-----------------------------------------------------------------------
void RovioRobot::controlRobot_(void)
{
    HSSpeed_t controlSpeed;
    float vt, va, rr, rt, leftSpeed, rightSpeed;
    float altVelocity, linearVelocity, latVelocity, rotVelocity;
    bool normalized = false;

    getControlSpeed_(controlSpeed);

    switch (robotSpecIndex_) {

    case ROVIO_ROBOT_CLASS_DEFAULT:
        vt = controlSpeed.driveSpeed.value;
        va = ROBOT_DEG2RAD(controlSpeed.steerSpeed.value);
        rr = ROBOT_SPECS_[robotSpecIndex_].robotRadius;
        rt = ROBOT_SPECS_[robotSpecIndex_].wheelRadius;

        if ((rr <= 0) || (rt <= 0))
        {
            printfTextWindow("Error:RovioRobot::controlRobot_(). Invalid radius.\n");
            return;
        }

        leftSpeed = (vt - rr*va)/rt;
        rightSpeed = (vt + rr*va)/rt;

        if (rovioInterface != NULL) rovioInterface->SetControl(vt, 0, va);
        break;
      default:
	printfTextWindow("Error: Invalid robot type %d\n", robotSpecIndex_);
    }

}

//-----------------------------------------------------------------------
// This function adds the sensors.
//-----------------------------------------------------------------------
void RovioRobot::addSensors_(void)
{

    float x, y, dist;

    partXyt_ = new RobotPartXyt();

    // Treat IR as sonar because obstacle avoidance only considers sonar & laser
    if (useIR) {
      partIr = new RobotPartSonar(ROVIO_NUM_IR, (float *) rovio_range_angle_IR);
      sensorIr = new Sensor(
         SENSOR_SONAR,
         0,
         NULL,
         numIR,
         (float *) rovio_range_angle_IR,
         numIR,
         ir,
         NULL);
    }



}

//-----------------------------------------------------------------------
// This function sets up the user interface for the robot option.
//-----------------------------------------------------------------------
void RovioRobot::control(void)
{
    bool done = false;
    int msgData[2];

    msgData[0] = 0; // "speed_factor" is obsolete.
    msgData[1] = 0; // "angular_speed_factor" is obsolete.

    /*
      messageDrawWindow(
      EnMessageType_ROVIO_ROBOT_CONTROL,
      EnMessageErrType_NONE,
      msgData);
    */

    do
    {
        int c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

/**********************************************************************
# $Log: rovio-robot.c,v $
# Revision 1.4  2010-05-03 22:10:10  zkira
# Cricket sensor, blob tracking, rovio (inc. IR sensor), skybotix interface (unfinished)
#
# Revision 1.2  2009/05/10 20:07:29  zkira
# Rovio additions
#
# Revision 1.1  2009/03/02 18:39:22  zkira
# Rovio robot (currently only skeleton copied from USARSim)
#
# Revision 1.9  2008/10/30 22:50:51  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.8  2008/10/28 23:12:18  ts2883
# Use ground truth always (even for ground vehicles)
#
# Revision 1.7  2008/10/28 23:05:17  zkira
# Fixed LaserPart problem
#
# Revision 1.6  2008/10/27 21:39:25  zkira
# Added separate dragonfly model
#
# Revision 1.5  2008/09/18 07:36:01  zkira
# Added ground truth
# BUG fix: Changed MAIN_LOOP_SLEEP because it was too slow (only every 10th or so message
# is an odometry reading!).  It may have to be changed back if it causes network troubles though.
#
# Revision 1.4  2008/09/09 20:18:53  zkira
# Removed debug print
#
# Revision 1.3  2008/09/09 20:16:23  zkira
# Crude altitude control
#
# Revision 1.2  2008/08/26 19:40:50  endo
# Improving the Rovio module.
#
# Revision 1.5  2008/02/01 23:18:26  endo
# RovioLaser added.
#
# Revision 1.4  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.3  2008/01/11 15:51:18  endo
# Regular backup.
#
# Revision 1.2  2008/01/10 23:02:54  endo
# Regular backup.
#
# Revision 1.1  2008/01/09 22:54:03  endo
# RovioRobot added.
#
#**********************************************************************/
