/**********************************************************************
 **                                                                  **
 **                          rovio-robot.h                         **
 **                                                                  **
 **                                                                  **
 **  Written by:  Zsolt Kira                                         **
 **                                                                  **
 **  Interface for Rovio robot.                                      **
 **                                                                  **
 **  Copyright 2009, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: rovio-robot.h,v 1.4 2010-05-03 22:10:10 zkira Exp $ */

#ifndef ROVIO_ROBOT_H
#define ROVIO_ROBOT_H

#include <curl/curl.h>
#include <pthread.h>
#include <string>
#include <vector>
#include <map>

#include "HServerTypes.h"
#include "robot.h"
#include "sensors.h"
#include "robot_config.h"
#include "RovioOdometry/rovio.h"

using std::string;
using std::vector;
using std::map;

#define ROVIO_ROBOT_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}
#define ROVIO_ROBOT_RAD2DEG(val) ((val)*180.0/M_PI)
#define ROVIO_ROBOT_DEG2RAD(val) ((val)*M_PI/180.0)
#define ROVIO_ROBOT_MAX_SONARS 100
#define ROVIO_ROBOT_SONAR_LOC_DIMENSIONS 3
#define ROVIO_NUM_IR 2

class RovioRobot : public Robot {

    typedef enum InitiConnectionStatus_t
    {
        INIT_CONNECTION_STATUS_FAILURE,
        INIT_CONNECTION_STATUS_NOT_READY,
        INIT_CONNECTION_STATUS_SUCCESS
    };

    typedef enum RovioRobotClasses_t
    {
        ROVIO_ROBOT_CLASS_DEFAULT,
        NUM_ROVIO_ROBOT_CLASSES
    };

    typedef struct RobotSpec_t
    {
        int robotClass;
        string className;
        float robotRadius;
        float wheelRadius;
    };

    typedef enum ERSLowSpeedGatePhase_t
    {
        ERS_LOWSPEED_GATE_PHASE_0,
        ERS_LOWSPEED_GATE_PHASE_1,
        ERS_LOWSPEED_GATE_PHASE_2,
        ERS_LOWSPEED_GATE_PHASE_3,
        ERS_LOWSPEED_GATE_PHASE_4,
        ERS_LOWSPEED_GATE_PHASE_5,
        ERS_LOWSPEED_GATE_PHASE_6,
        ERS_LOWSPEED_GATE_PHASE_7,
        ERS_LOWSPEED_GATE_PHASE_8,
        ERS_LOWSPEED_GATE_PHASE_9,
        NUM_ERS_LOWSPEED_GATE_PHASES
    };

    typedef vector<string> DataLine_t;


protected:
    Sensor *sensorIr;
    HSPose_t shRovioPoseOffset_; // Shared
    HSPose_t shLatestRawPose_; // Shared
    RobotPartXyt *partXyt_;
    RobotPartSonar *partIr;
    pthread_t mainThread_;
    pthread_t readerThread_;
    string serverHostMachine_;
    string rovioRobotName_;
    string rovioRobotClass_;
    map<string, DataLine_t> shSensorDataTable_;
    float initX_, initY_, initZ_;
    float initYaw_, initPitch_, initRoll_;
    int robotSpecIndex_;
    int shSocketFd_;
    int serverPort_;
    bool mainThreadIsUp_;
    bool readerThreadIsUp_;
    Rovio *rovioInterface;
    float ir[ROVIO_NUM_IR];
    float ir_loc[ROVIO_NUM_IR][3];
    int numIR;

#ifndef NO_OPENCV
    IplImage *CurFrame;
#endif

    static pthread_mutex_t latestRawPoseMutex_;
    static pthread_mutex_t sensorDataTableMutex_;
    static pthread_mutex_t socketMutex_;
    static pthread_mutex_t rovioPoseOffsetMutex_;

    static const RobotSpec_t ROBOT_SPECS_[NUM_ROVIO_ROBOT_CLASSES];
    static const string DEFAULT_ROVIO_ROBOT_CLASS_;
    static const string DEFAULT_ROVIO_ROBOT_NAME_;
    static const string DEFAULT_ROVIO_SERVER_HOST_MACHINE_;
    static const int DEFAULT_ROVIO_SERVER_PORT_;
    static const int MAIN_THREAD_USLEEP_;
    static const int READER_THREAD_USLEEP_;
    static const int WATCHDOG_CHECK_INTERVAL_SEC_;
    static const int SKIP_STATUSBAR_UPDATE_;
    static const string EMPTY_STRING_;
    int rovioNumber;
    DataLine_t readDataLine_(void);
    vector<string> extractCommaSepratedValues_(string values);
    string readString_(void);
    string generateERSGateCommand_(HSSpeed_t controlSpeed);
    void addSensors_(void);
    void controlRobot_(void);
    void getLatestRawPose_(HSPose_t &pose);
    void getRovioPoseOffset_(HSPose_t &poseOffset);
    void mainLoop_(void);
    void parsePositionData_(HSPose_t &pose);
    void readerLoop_(void);
    void setLatestRawPose_(HSPose_t pose);
    void setRovioPoseOffset_(HSPose_t poseOffset);
    void updatePoseFromRovioPoseOffset_(void);
    void updatePose_(void);
    void updateStatusBar_(void);
    void waitForSensorData_(void);
    void writeString_(string value);
    bool initializeRovio_(void);
    bool readNBytes_(const int fd, unsigned char *buf, const int n);

    static void* startMainThread_(void *rovioInstance);
    static void* startReaderThread_(void *rovioInstance);

public:
    RovioRobot(
        Robot** a,
        char *serverHostMahine,
        char *rovioRobotName,
        char *rovioRobotClass,
        int serverPort,
        float initX,
        float initY,
        float initZ,
        float initYaw,
        float initPitch,
        float initRoll);
    ~RovioRobot(void);
    int getLaserData(float **laserReadings, float **laserAngles, const char *laserName);
    void control(void);
    void runKalman(double ox, double oy, double ot,  double &newx, double &newy, double &newtheta);


    FrameData GetFrame(void) { printfTextWindow("getting frame from rovio"); return rovioInterface->GetCurFrame(); };
};    

#endif

/**********************************************************************
# $Log: rovio-robot.h,v $
# Revision 1.4  2010-05-03 22:10:10  zkira
# Cricket sensor, blob tracking, rovio (inc. IR sensor), skybotix interface (unfinished)
#
# Revision 1.3  2009/09/29 17:00:17  zkira
# Blob detection stuff
#
# Revision 1.2  2009/05/10 20:07:29  zkira
# Rovio additions
#
# Revision 1.1  2009/03/02 18:39:22  zkira
# Rovio robot (currently only skeleton copied from USARSim)
#
# Revision 1.5  2008/10/30 22:50:51  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.4  2008/10/27 21:39:26  zkira
# Added separate dragonfly model
#
# Revision 1.3  2008/09/18 07:36:02  zkira
# Added ground truth
# BUG fix: Changed MAIN_LOOP_SLEEP because it was too slow (only every 10th or so message
# is an odometry reading!).  It may have to be changed back if it causes network troubles though.
#
# Revision 1.2  2008/08/26 19:40:50  endo
# Improving the Rovio module.
#
# Revision 1.5  2008/02/01 23:18:26  endo
# RovioLaser added.
#
# Revision 1.4  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.3  2008/01/11 15:51:18  endo
# Regular backup.
#
# Revision 1.2  2008/01/10 23:02:54  endo
# Regular backup.
#
# Revision 1.1  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
#**********************************************************************/
