/**********************************************************************
 **                                                                  **
 **                             sensor.h                             **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Sensor routines for HServer                                     **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: sensors.h,v 1.4 2008-11-20 03:51:46 ts2883 Exp $ */

#ifndef SENSORS_H
#define SENSORS_H
 
#define SENSORS_WIN_X      8
#define SENSORS_WIN_Y      5
#define SENSORS_WIN_ROW    40
#define SENSORS_WIN_COL    60

#include <panel.h>
#include <pthread.h>
#include "hserver.h"

//-----------------------------------------------------------------------
// Class Sensors
//-----------------------------------------------------------------------
class Sensor {

    friend class Sensors;

protected:
    // These are for locking the data so buffers aren't being written and 
    // read at the same time. Originally, there were only the mutexes, not 
    // the pointers. The pointers were added so that a mutex could be passed
    // in via the constructor to protect the buffers. The actual mutexes were
    // kept for backwards compatibility. If no mutex pointer is passed in, 
    // then the member mutexes will be used to lock the buffers. -- jbl
    pthread_mutex_t *externalDataMutexPtr_;
    float* m_pfAngles;
    float m_pfPitchAngle;
    float* m_pfReadings;
    int m_iType;
    int m_iLenExtra;
    int m_iNumAngles;
    int m_iNumReadings;
    int m_iIndex;
    char* m_pcExtra;

    static pthread_mutex_t localDataMutex_;

    static bool TryToLock(pthread_mutex_t* pMutex);
    static bool TryToUnlock(pthread_mutex_t* pMutex);

public:
    int getType(void){ return m_iType; };
    int getLenExtra(void){ return m_iLenExtra; };
    char* getExtra(void){ return m_pcExtra; };
    int getAngleCount(void);
    int getReadingCount(void);
    void getAngleInfo(float*& pfAngles, int& iCount);
    float getPitchAngle(void){ return m_pfPitchAngle; };
    void getReadingInfo(float*& pfReadings, int& iCount);
    void setAngleInfo(float* pfAngles, int iCount);
    void setPitchAngle(float pfPitchAngle);
    void setReadingInfo(float* pfReadings, int iCount);

    Sensor(
        int type,
        int len_extra,
        char *extra,
        int num_angles,
        float *angle,
        int num_readings,
        float *reading, 
        pthread_mutex_t *externalDataMutexPtr);

    ~Sensor(void);
};

//-----------------------------------------------------------------------
// Class Sensors
//-----------------------------------------------------------------------
class Sensors {

protected:
    int max_sensors;
    char *index;
    int num_sensors;
    WINDOW *window;
    PANEL *panel;
    int panelX,panelY,panelW,panelH;
    pthread_t updator;
    void capInt(int *i,int l,int u);
    void drawWindow(void);
    void movePanel(int xdif,int ydif);
    void movePanelKey(int c);
    int runKey(int c);
    void windowInit(void);

public:
    Sensor **latest_sensor;
    int numSensors(void){return num_sensors;};
    void updator_thread(void);
    void show(void);
    int add(Sensor *sensor);
    void remove(int idx);
    Sensors(int max);
    ~Sensors(void);
};

// this will try first to lock a passed-in mutex, but if there isn't one 
// a local member mutex will be used.
inline bool Sensor::TryToLock(pthread_mutex_t* pMutex)
{
    if (pMutex != NULL)
    {
        pthread_mutex_lock(pMutex);
    }

    return (pMutex != NULL);
}

// this will try first to unlock a passed-in mutex, but if there isn't one 
// a local member mutex will be used.
inline bool Sensor::TryToUnlock(pthread_mutex_t* pMutex)
{
    if (pMutex != NULL)
    {
        pthread_mutex_unlock(pMutex);
    }

    return (pMutex != NULL);
}

extern Sensors *sensors;

#endif

/**********************************************************************
# $Log: sensors.h,v $
# Revision 1.4  2008-11-20 03:51:46  ts2883
# Finished adding the pitch information for the laser.
#
# Revision 1.3  2008/11/04 04:53:58  zkira
# Added pitch for 3D laser readings
#
# Revision 1.2  2008/07/15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.3  2006/12/07 20:30:23  endo
# Regular backup.
#
# Revision 1.2  2006/11/16 18:20:26  endo
# The virtual bumper added.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2006/05/15 06:01:03  endo
# gcc-3.4 upgrade
#
# Revision 1.1.1.1  2005/02/06 22:59:42  endo
# AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
#
# Revision 1.4  2000/12/12 23:08:59  blee
# Changed member variable names in Sensor. Changed "private" to "protected".
# Added mutex pointer member variables. Added TryToLock() and TryToUnlock().
# Changed/added accessor members. Changed sensor constructor prototype.
#
# Revision 1.3  2000/09/19 03:49:46  endo
# RCS log added.
#
#
#**********************************************************************/
