/**********************************************************************
 **                                                                  **
 **                            statusbar.c                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Status bar updates for HServer                                  **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: statusbar.c,v 1.2 2008-07-15 22:33:53 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <string.h>

#include "hserver.h"
#include "cognachrome.h"
#include "nomad.h"
#include "statusbar.h"

//-----------------------------------------------------------------------
// Extern
//-----------------------------------------------------------------------
Statusbar *statusbar = NULL;
//char srstr[100];

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------
pthread_mutex_t Statusbar::numLinesMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constructor
//-----------------------------------------------------------------------
Statusbar::Statusbar(int size) :
    window_(NULL),
    panel_(NULL),
    shNumLines_(0),
    ORIGINAL_BAR_SIZE_(size)
{
    pthread_mutex_init(&numLinesMutex_, NULL);

    window_ = newwin(
        ORIGINAL_BAR_SIZE_ + shNumLines_,
        screenX - 2,
        HS_STATUS_Y,
        HS_STATUS_X);

    panel_ = new_panel(window_);

    memset(lines, 0x0, sizeof(char*)*STATUSBAR_MAX_LINES);
    memset(index, -1, sizeof(int)*STATUSBAR_MAX_LINES);
    empty_line = new char[300];
    memset(empty_line,' ',screenX-2);
    draw();
    show_panel(panel_);
}

void Statusbar::draw(void)
{
    int i, numLines;

    if (SilentMode)
    {
        return;
    }

    pthread_mutex_lock(&numLinesMutex_);
    numLines = shNumLines_;
    pthread_mutex_unlock(&numLinesMutex_);

    pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);

    werase(window_);


    for(i = 0; i < numLines; i++)
    {
        wmove(window_,i,0);
        waddstr(window_,lines[i]);
    }


    pthread_cleanup_pop(1);
    redrawWindows();
}

int Statusbar::add(char *line)
{
    int i;
    int idx = -1;

    pthread_mutex_lock(&numLinesMutex_);

    if (shNumLines_ == STATUSBAR_MAX_LINES)
    {
        pthread_mutex_unlock(&numLinesMutex_);
        return (-1);
    }

    lines[shNumLines_] = line;

    for(i = 0; i < STATUSBAR_MAX_LINES; i++)
    {
        if (index[i] == -1)
        {
            idx = i;
            break;
        }
    }

    if (idx == -1)
    {
        pthread_mutex_unlock(&numLinesMutex_);
        return (-1);
    }

    index[idx] = shNumLines_;
    shNumLines_++;

    pthread_mutex_unlock(&numLinesMutex_);

    statusbarDevideLine++;
    resizeWindows(false);
    resize();

    return(idx);
}

void Statusbar::update(int line_idx)
{
    if (SilentMode)
    {
       return;
    }

    pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock, (void *) &screenMutex);
    pthread_mutex_lock(&screenMutex);

    wmove(window_,index[line_idx],0);
    waddnstr(window_,empty_line,screenX-2);
    wmove(window_,index[line_idx],0);
    waddnstr(window_,lines[index[line_idx]],screenX-2);

    pthread_cleanup_pop(1);

    redrawWindows();
}

void Statusbar::remove(int line_idx)
{
    int i,low;

    low = index[line_idx];
    index[line_idx] = -1;

    for(i = 0; i < STATUSBAR_MAX_LINES; i++)
    {
        if (index[i] > low)
        {
            index[i]--;
        }

        if ((i >= low) && (i < (STATUSBAR_MAX_LINES-1)))
        {
            lines[i] = lines[i+1];
        }
    }

    pthread_mutex_lock(&numLinesMutex_);
    shNumLines_--;
    pthread_mutex_unlock(&numLinesMutex_);

    statusbarDevideLine--;
    resizeWindows(false);
    draw();
}

void Statusbar::resize(void)
{
    pthread_mutex_lock(&numLinesMutex_);
    wresize(window_, ORIGINAL_BAR_SIZE_ + shNumLines_, screenX-2);
    pthread_mutex_unlock(&numLinesMutex_);
}

StatusbarSpinner::StatusbarSpinner(void) :
    counter_(0)
{
}

StatusbarSpinner::~StatusbarSpinner(void)
{
}

char StatusbarSpinner::getStatus(void)
{
    char status = '\0';

    switch(counter_) {

    case 0:
        status = '-';
        break;
    case 1:
        status = '\\';
        break;
    case 2:
        status = '|';
        break;
    case 3:
        status = '/';
        break;
    }

    counter_ = (counter_ + 1) % 4;

    return status;
}

/**********************************************************************
# $Log: statusbar.c,v $
# Revision 1.2  2008-07-15 22:33:53  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.2  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.3  2004/09/10 19:41:08  endo
# New PoseCalc integrated.
#
# Revision 1.2  2004/04/06 02:07:40  pulam
# Added code for Silent Mode
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.2  2000/09/19 03:49:46  endo
# RCS log added.
#
#
#**********************************************************************/
