/**********************************************************************
 **                                                                  **
 **                         usarsim-robot.c                          **
 **                                                                  **
 **                                                                  **
 **  Written by: Yoichiro Endo                                       **
 **                                                                  **
 **  Interface for USARSim robot.                                    **
 **                                                                  **
 **  Copyright 2008, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: usarsim-robot.c,v 1.15 2009-08-28 19:46:40 zkira Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/un.h>
#include <netdb.h>
#include <pthread.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "usarsim-robot.h"
#include "Watchdog.h"
#include "LogManager.h"

//-----------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t USARSimRobot::latestRawPoseMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t USARSimRobot::sensorDataTableMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t USARSimRobot::socketMutex_ = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t USARSimRobot::usarsimPoseOffsetMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const USARSimRobot::RobotSpec_t USARSimRobot::ROBOT_SPECS_[NUM_USARSIM_ROBOT_CLASSES] =
{
    // Note: The robot's Z-axis in USARSim points opposite of Z-axis in MissionLab.
    {
        USARSIM_ROBOT_CLASS_AIRROBOT,
        "USARBot.AirRobot", // className
        0, // numSonars
        {}, // sonarAngles
        {}, // sonarCoordinates
        0.5, // robotRadius
        0.0 // wheelRadius
    },
    {   USARSIM_ROBOT_CLASS_DRAGONFLYROBOT,
        "USARBot.DragonflyRobot", // className
        0, // numSonars
        {}, // sonarAngles
        {}, // sonarCoordinates
        0.05, // robotRadius
        0.0 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_ATRVJR,
        "USARBot.ATRVJr", // className
        17, // numSonars
        {30,15,0,-15,-30,-45,-60,-75,-90,-90,-180,-180,90,90,75,60,45}, // sonarAngles
        {0.335,0.104,0.34,0.05,0.347,0,0.34,-0.05,0.335,-0.104,0.23,-0.175,0.173,-0.179,0.117,-0.181,0.072,-0.181,-0.295,-0.181,-0.347,-0.15,-0.347,0.15,-0.295,0.181,0.072,0.181,0.117,0.181,0.173,0.179,0.23,0.175}, // sonarCoordinates
        0.25, // robotRadius
        0.11 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_ERS,
        "USARBot.ERS", // className
        0, // numSonars
        {}, // sonarAngles
        {}, // sonarCoordinates
        0.25, // robotRadius
        0.15 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_HMMWV,
        "USARBot.Hummer", // className
        0, // numSonars
        {}, // sonarAngles
        {}, // sonarCoordinates
        0.9, // robotRadius
        0.373 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_P2AT, // robotClass
        "USARBot.P2AT", // className
        16, // numSonars
        {90,50,30,10,-10,-30,-50,-90,-90,-130,-150,-170,170,150,130,90}, // sonarAngles
        {0.145,0.13,0.185,0.115,0.22,0.08,0.24,0.025,0.24,0.-025,0.22,-0.08,0.185,-0.115,0.145,-0.13,-0.145,-0.13,-0.185,-0.115,-0.22,-0.08,-0.24,-0.025,-0.24,0.025,-0.22,0.08,-0.185,0.115,-0.145,0.13}, // sonarCoordinates
        0.283, // robotRadius
        0.11 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_P2DX,
        "USARBot.P2DX", // className
        8, // numSonars
        {90,50,30,10,-10,-30,-50,-90}, // sonarAngles
        {0.115,0.13,0.155,0.115,0.19,0.08,0.21,0.025,0.21,-0.025,0.19,-0.08,0.155,-0.115,0.115,-0.13}, // sonarCoordinates
        0.209, // robotRadius
        0.083 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_TALON,
        "USARBot.Talon", // className
        0, // numSonars
        {}, // sonarAngles
        {}, // sonarCoordinates
        0.75, // robotRadius
        0.28 // wheelRadius
    },
    {
        USARSIM_ROBOT_CLASS_TARANTULA,
        "USARBot.Tarantula", // className
        0, // numSonars
        {}, // sonarAngles
        {}, // sonarCoordinates
        0.5, // robotRadius
        0.07 // wheelRadius
    }
};
const string USARSimRobot::CMD_KEY_DRIVE_ = "DRIVE";
const string USARSimRobot::CMD_KEY_INIT_ = "INIT";
const string USARSimRobot::CMD_KEY_MULTIDRIVE_ = "MULTIDRIVE";
const string USARSimRobot::CMD_STRING_ALTITUDE_VELOCITY_ = "AltitudeVelocity";
const string USARSimRobot::CMD_STRING_CLASS_NAME_ = "ClassName";
const string USARSimRobot::CMD_STRING_FALSE_ = "False";
const string USARSimRobot::CMD_STRING_LATERAL_VELOCITY_ = "LateralVelocity";
const string USARSimRobot::CMD_STRING_LEFT_ = "LEFT";
const string USARSimRobot::CMD_STRING_LINEAR_VELOCITY_ = "LinearVelocity";
const string USARSimRobot::CMD_STRING_LOCATION_ = "Location";
const string USARSimRobot::CMD_STRING_NAME_ = "Name";
const string USARSimRobot::CMD_STRING_NORMALIZED_ = "Normalzed";
const string USARSimRobot::CMD_STRING_RIGHT_ = "RIGHT";
const string USARSimRobot::CMD_STRING_ROTATION_ = "Rotation";
const string USARSimRobot::CMD_STRING_ROTATIONAL_VELOCITY_ = "RotationalVelocity";
const string USARSimRobot::CMD_STRING_TRUE_ = "True";
const string USARSimRobot::DATA_HEADER_SENSOR_ = "SEN";
const string USARSimRobot::DATA_KEY_FOV_ = "FOV";
const string USARSimRobot::DATA_KEY_LOCATION_ = "Location";
const string USARSimRobot::DATA_KEY_ORIENTATION_ = "Orientation";
const string USARSimRobot::DATA_KEY_POSE_ = "Pose";
const string USARSimRobot::DATA_KEY_RANGE_ = "Range";
const string USARSimRobot::DATA_KEY_RFID_ID_ = "ID";
const string USARSimRobot::DATA_KEY_RFID_LOCATION_ = "Location";

const string USARSimRobot::DATA_KEY_TIME_ = "Time";

const string USARSimRobot::DATA_KEY_ENEMY_ID_ = "ID";
const string USARSimRobot::DATA_KEY_ENEMY_DISTANCE_ = "Distance";
const string USARSimRobot::DATA_KEY_ENEMY_LOCATION_ = "Location";
const string USARSimRobot::DATA_KEY_ENEMY_PODSELF_ = "PODSelf";
const string USARSimRobot::DATA_KEY_ENEMY_PODENEMY_ = "PODEnemy";

const string USARSimRobot::DATA_KEY_CHEMICAL_ID_ = "ID";
const string USARSimRobot::DATA_KEY_CHEMICAL_DISTANCE_ = "Distance";
const string USARSimRobot::DATA_KEY_CHEMICAL_LOCATION_ = "Location";
const string USARSimRobot::DATA_KEY_CHEMICAL_PODCHEM_ = "PODChemical";

const string USARSimRobot::DATA_KEY_COMM_BUGNUM_ = "BugNum";
const string USARSimRobot::DATA_KEY_COMM_PD_ = "Pd";


const string USARSimRobot::DATA_KEY_TYPE_ = "Type";
// Do the subsequent Values follow Type?
const string USARSimRobot::DATA_VALUE_INS_ = "INS";
const string USARSimRobot::DATA_VALUE_LASER_ = "RangeScanner";
const string USARSimRobot::DATA_VALUE_ODOMETRY_ = "Odometry";
const string USARSimRobot::DATA_VALUE_GROUND_TRUTH_ = "GroundTruth";
const string USARSimRobot::DATA_VALUE_SONAR_ = "Sonar";
const string USARSimRobot::DATA_VALUE_RFID_ = "RFIDTag";   // NOTE: In the manual, sometimes it says RFIDTag sometimes RFID (have to verify by testing it)

const string USARSimRobot::DATA_VALUE_CHEMICAL_ = "Chemical";
const string USARSimRobot::DATA_VALUE_ENEMY_ = "Enemy";
const string USARSimRobot::DATA_VALUE_COMM_ = "Communication";

const string USARSimRobot::DATA_VALUE_TOUCH_ = "Touch";
const string USARSimRobot::DEFAULT_USARSIM_ROBOT_NAME_ = "us_bot";
const string USARSimRobot::DEFAULT_USARSIM_ROBOT_CLASS_ = "USARBot.P2AT";
const string USARSimRobot::DEFAULT_USARSIM_SERVER_HOST_MACHINE_ = "127.0.0.1";
const string USARSimRobot::EMPTY_STRING_ = "";
const int USARSimRobot::DEFAULT_USARSIM_SERVER_PORT_ = 3000;
const int USARSimRobot::INVALID_SOCKET_FD_ = -1;
const int USARSimRobot::MAIN_UT_SLOMO_ = 10.0;  // For speeding up UT faster than real-time (1.0 is realtime)
const int USARSimRobot::MAIN_THREAD_USLEEP_ = (2000/MAIN_UT_SLOMO_);
const int USARSimRobot::MAX_CMD_LENGTH_ = 4096;
const int USARSimRobot::SKIP_STATUSBAR_UPDATE_ = 10;
const int USARSimRobot::WATCHDOG_CHECK_INTERVAL_SEC_ = 2;
const char USARSimRobot::CHAR_BRACE_END_ = '}';
const char USARSimRobot::CHAR_BRACE_START_ = '{';
const char USARSimRobot::CHAR_CARRIAGE_RETURN_ = '\r';
const char USARSimRobot::CHAR_COMMA_ = ',';
const char USARSimRobot::CHAR_NEW_LINE_ = '\n';
const char USARSimRobot::CHAR_NULL_ = '\0';
const char USARSimRobot::CHAR_WHITE_SPACE_ = ' ';

// The following hardcode the sensors as active--they're not parsed in the .rc file -?
const bool USARSimRobot::REPORT_SONARS_ = true;
const bool USARSimRobot::REPORT_RFID_ = false;
const bool USARSimRobot::REPORT_ENEMY_ = true;
const bool USARSimRobot::REPORT_CHEMICAL_ = true;
const bool USARSimRobot::REPORT_COMM_ = true;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
USARSimRobot::USARSimRobot(
    Robot** a,
    char *serverHostMachine,
    char *usarsimRobotName,
    char *usarsimRobotClass,
    int serverPort,
    float initX,
    float initY,
    float initZ,
    float initYaw,
    float initPitch,
    float initRoll) :
    Robot(a, HS_ROBOT_TYPE_USARSIM, HS_MODULE_NAME_ROBOT_USARSIM),
    partXyt_(NULL),
    sensorSonar_(NULL),
    sensorRFID_(NULL),
    sensorEnemy_(NULL),
    sensorChemical_(NULL),
    sensorComm_(NULL),
    initX_(initX),
    initY_(initY),
    initZ_(initZ),
    initYaw_(initYaw),
    initPitch_(initPitch),
    initRoll_(initRoll),
    sonarMountDist_(NULL),
    sonarReadings_(NULL),
    numSonars_(0),
    RFIDData_(NULL),
    numRFID_(0),

    enemyData_(NULL),
    numEnemy_(0),
    chemicalData_(NULL),
    numChemical_(0),
    commData_(NULL),
    numComm_(0),

    robotSpecIndex_(-1),
    shSocketFd_(INVALID_SOCKET_FD_),
    serverPort_(DEFAULT_USARSIM_SERVER_PORT_),
    mainThreadIsUp_(false),
    readerThreadIsUp_(false)
{
    serverHostMachine_ = (serverHostMachine != NULL)?
        strdup(serverHostMachine) : DEFAULT_USARSIM_SERVER_HOST_MACHINE_;

    usarsimRobotName_ = (usarsimRobotName != NULL)?
        strdup(usarsimRobotName) : DEFAULT_USARSIM_ROBOT_NAME_;

    usarsimRobotClass_ = (usarsimRobotClass != NULL)?
        strdup(usarsimRobotClass) : DEFAULT_USARSIM_ROBOT_CLASS_;

    serverPort_ = (serverPort_ >= 0)?
        serverPort : DEFAULT_USARSIM_SERVER_PORT_;

    pthread_mutex_init(&sensorDataTableMutex_, NULL);
    pthread_mutex_init(&socketMutex_, NULL);
    pthread_mutex_init(&usarsimPoseOffsetMutex_, NULL);
    pthread_mutex_init(&latestRawPoseMutex_, NULL);

    resetPose_();
    updateStatusBar_();

    if (!initializeUSARSim_())
    {
        printfTextWindow("Error: USARSimRobot could not be initialized. Exiting USARSim.\n");
        delete this;
        return;
    }

    addSensors_();

    pthread_create(&readerThread_, NULL, &startReaderThread_, (void*)this);
    pthread_create(&mainThread_, NULL, &startMainThread_, (void*)this);

    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(NAME_, WATCHDOG_CHECK_INTERVAL_SEC_, getpid(), mainThread_);
    }

    refreshScreen();
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
USARSimRobot::~USARSimRobot(void)
{
    if (mainThreadIsUp_)
    {
        pthread_cancel(mainThread_);
        pthread_join(mainThread_, NULL);
    }

    if (readerThreadIsUp_)
    {
        pthread_cancel(readerThread_);
        pthread_join(readerThread_, NULL);
    }

    if (sonarMountDist_ != NULL)
    {
        delete [] sonarMountDist_;
        sonarMountDist_ = NULL;
    }

    if (sonarReadings_ != NULL)
    {
        delete [] sonarReadings_;
        sonarReadings_ = NULL;
    }

    if (RFIDData_ != NULL)
    {
        delete [] RFIDData_;
        RFIDData_ = NULL;
    }

    if (chemicalData_ != NULL)
    {
        delete [] chemicalData_;
        chemicalData_ = NULL;
    }

    if (enemyData_ != NULL)
    {
        delete [] enemyData_;
        enemyData_ = NULL;
    }

    if (commData_ != NULL)
    {
        delete [] commData_;
        commData_ = NULL;
    }

    delete partXyt_;
}

//-----------------------------------------------------------------------
// This function connects to the server through a socket.
//-----------------------------------------------------------------------
int USARSimRobot::connectToServer_(string hostMacine, int port)
{
    struct sockaddr_in sin;
    struct hostent *hent = NULL;
    int sockfd;
    int status = INIT_CONNECTION_STATUS_FAILURE;

    sin.sin_family = PF_INET;
    sin.sin_port = htons(port);

    if (hostMacine == EMPTY_STRING_)
    {
        printfTextWindow(
            "Error: USARSimRobot::connectToServer_(). Host name is empty.\n");
        return INIT_CONNECTION_STATUS_FAILURE;
    }

    if (isdigit(hostMacine[0]))
    {
        sin.sin_addr.s_addr = inet_addr(hostMacine.c_str());
    }
    else
    {
        // Get the IP address.
        if((hent = gethostbyname(hostMacine.c_str())) == NULL)
        {
            printfTextWindow(
                "Error: USARSimRobot::connectToServer_(). Unknown host: %s\n",
                hostMacine.c_str());
            return INIT_CONNECTION_STATUS_FAILURE;
        }

        memcpy((char*) &sin.sin_addr, hent->h_addr, sizeof(sin.sin_addr));
    }

    bzero(&(sin.sin_zero), 8);

    // Open the socket.
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        printfTextWindow(
            "Error: USARSimRobot::connectToServer_(). Socket could not be opened.\n");
        return INIT_CONNECTION_STATUS_FAILURE;
    }

    // Connect
    if (connect(sockfd, (sockaddr*)&sin, sizeof(sin)))
    {
        close(sockfd);

        switch(errno) {

        case ECONNREFUSED:
            status = INIT_CONNECTION_STATUS_NOT_READY;
            break;

        case ENETUNREACH:
            printfTextWindow(
                "Error: USARSimRobot::connectToServer_(). Network unreachable from this host.");
            status = INIT_CONNECTION_STATUS_FAILURE;
            break;

        default:
            printfTextWindow(
                "Error: USARSimRobot::connectToServer_(). Could not connect to socket.");
            status = INIT_CONNECTION_STATUS_FAILURE;
            break;
        }

        return status;
    }

    shSocketFd_ = sockfd;

    return INIT_CONNECTION_STATUS_SUCCESS;
}

//-----------------------------------------------------------------------
// This function reads n bytes from the file descriptor.
//-----------------------------------------------------------------------
bool USARSimRobot::readNBytes_(const int fd, unsigned char *buf, const int n)
{
    int msgLength = -1, index = 0, remaining = 0;

    remaining = n;

    while (remaining > 0)
    {
        msgLength = read(fd, &buf[index], remaining);

        if (msgLength == 0)
        {
            printfTextWindow(
                "USARSimRobot: WARNING! Server may have died. Reconnection invoked.\n");
            initializeUSARSim_();
            return false;
        }
        else if (msgLength <= 0)
        {
            printfTextWindow(
                "USARSimRobot: WARNING! Read error %d occured. Reconnection invoked.\n",
                errno);
            initializeUSARSim_();
            return false;
        }

        index += msgLength;
        remaining -= msgLength;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function reads a string from the server.
//-----------------------------------------------------------------------
string USARSimRobot::readString_(void)
{
    string value;
    char c;

    pthread_mutex_lock(&socketMutex_);

    while (true)
    {
        // Read one byte at the time.
        readNBytes_(
            shSocketFd_,
            (unsigned char *)&c,
            sizeof(char));

        if ((c == CHAR_WHITE_SPACE_) ||
            (c == CHAR_CARRIAGE_RETURN_) ||
            (c == CHAR_NEW_LINE_))
        {
            break;
        }

        value += c;
    }

    pthread_mutex_unlock(&socketMutex_);

    return value;
}

//-----------------------------------------------------------------------
// This function reads a line from the server and saved in a vector array.
// white space and braces in the data are omitted.
//-----------------------------------------------------------------------
USARSimRobot::DataLine_t USARSimRobot::readDataLine_(void)
{
    DataLine_t dataLine;
    string value = EMPTY_STRING_;
    char c;
    char *buf = NULL;
    int i;
    const int SHOW_DATA_CONTENTS_LENGTH = 40;
    const bool SHOW_DATA_CONTENTS = false;

    pthread_mutex_lock(&socketMutex_);

    while (true)
    {
        // Read one byte at the time.
        readNBytes_(
            shSocketFd_,
            (unsigned char *)&c,
            sizeof(char));

        if ((c == CHAR_CARRIAGE_RETURN_) || (c == CHAR_NEW_LINE_))
        {
            if (value.size() > 0)
            {
                dataLine.push_back(value);
            }

            break;
        }
        else if ((c == CHAR_BRACE_START_) || (c == CHAR_BRACE_END_))
        {
            continue;
        }
        else if (c == CHAR_WHITE_SPACE_)
        {
            if (value.size() > 0)
            {
                dataLine.push_back(value);
                value = EMPTY_STRING_;
            }

            continue;
        }

        value += c;
    }

    pthread_mutex_unlock(&socketMutex_);

    if ((report_level >= HS_REP_LEV_DEBUG) || SHOW_DATA_CONTENTS)
    {
        buf = new char[SHOW_DATA_CONTENTS_LENGTH+1];

        for (i = 0; i < (int)(dataLine.size()); i++)
        {
            if ((int)(dataLine[i].size()) > SHOW_DATA_CONTENTS_LENGTH)
            {
                memcpy(buf, dataLine[i].c_str(), SHOW_DATA_CONTENTS_LENGTH);
                buf[SHOW_DATA_CONTENTS_LENGTH] = CHAR_NULL_;
                sprintf(buf, "%s...", buf);
            }
            else
            {
                sprintf(buf, "%s]", dataLine[i].c_str());
            }

            printfTextWindow(
                "%s %s(): data[%s\n",
                __FILE__,
                __FUNCTION__,
                buf);

            sleep(1);
        }

        printfTextWindow("\n");

        delete [] buf;
        buf = NULL;
    }

    return dataLine;
}

//-----------------------------------------------------------------------
// This function reads the sensor data from the server.
//-----------------------------------------------------------------------
void USARSimRobot::readSensorData_(void)
{
    DataLine_t dataLine;
    string sensorType;
    int i;

    while (true)
    {
        dataLine = readDataLine_();

        if ((dataLine.size() >= 3) && (dataLine[0] == DATA_HEADER_SENSOR_))
        {
            // Remove the header.
            dataLine.erase(dataLine.begin());

            for (i = 0; i < ((int)(dataLine.size()) - 1); i++)
            {
              if (dataLine[i] == DATA_KEY_TYPE_)
              {
                
                // Use the sensor type as the key of the hash
                // table.
                if (dataLine[i+1] == DATA_VALUE_LASER_) {
                  // To support multiple lasers, we will store
                  // sensor readings by actual name of the range
                  // scanner, not just the sensor type
                  sensorType = dataLine[i+3];
                } else {
                  sensorType = dataLine[i+1];
                }
                
                // Add the data line in the hash table (sensor type as the key).
                pthread_mutex_lock(&sensorDataTableMutex_);
                shSensorDataTable_[sensorType] = dataLine;
                pthread_mutex_unlock(&sensorDataTableMutex_);
                
                break;
              }
            }
        }
    }
}

//-----------------------------------------------------------------------
// This function writes an integer number.
//-----------------------------------------------------------------------
void USARSimRobot::writeString_(string value)
{
    pthread_mutex_lock(&socketMutex_);

    if (write(shSocketFd_, (unsigned char *)(value.c_str()), sizeof(value.size())) == -1)
    {
        printfTextWindow("Error: USARSimRobot::writeString_(). write() failed.\n");
    }

    pthread_mutex_unlock(&socketMutex_);
}

//-----------------------------------------------------------------------
// This function starts the robot.
//-----------------------------------------------------------------------
bool USARSimRobot::startRobot_(void)
{
    string bufString;
    char cmd[MAX_CMD_LENGTH_ + 1];
    int i, cmdLength;

    for (i = 0; i < NUM_USARSIM_ROBOT_CLASSES; i++)
    {
        if (ROBOT_SPECS_[i].className == usarsimRobotClass_)
        {
            robotSpecIndex_ = i;
            break;
        }
    }

    if (robotSpecIndex_ < 0)
    {
        robotSpecIndex_ = 0;

        printfTextWindow(
            "Warning: USARSimRobot::%s(). Robot spect not found. Using %s.\n",
            __FUNCTION__,
            ROBOT_SPECS_[robotSpecIndex_].className.c_str());
    }

    memset(cmd, 0x0, MAX_CMD_LENGTH_);

    // Note: USARSim's Y and Yaw are opposite of MissionLab's.
    sprintf(
        cmd,
        "%s {%s %s} {%s %s} {%s %.2f,%.2f,%.2f} {%s %.2f,%.2f,%.2f}",
        CMD_KEY_INIT_.c_str(),
        CMD_STRING_CLASS_NAME_.c_str(),
        usarsimRobotClass_.c_str(),
        CMD_STRING_NAME_.c_str(),
        usarsimRobotName_.c_str(),
        CMD_STRING_LOCATION_.c_str(),
        initX_,
        -initY_,
        initZ_,
        CMD_STRING_ROTATION_.c_str(),
        USARSIM_ROBOT_DEG2RAD(initRoll_),
        USARSIM_ROBOT_DEG2RAD(initPitch_),
        -USARSIM_ROBOT_DEG2RAD(initYaw_));

    bufString = cmd;

    cmdLength = bufString.size();

    sprintf(cmd, "%s\r\n", cmd);
    cmdLength += 2; // '\r' + '\n'

    if (cmdLength > MAX_CMD_LENGTH_)
    {
        printfTextWindow("Error: USARSimRobot::startRobot_(). Command too long.\n");
        return false;
    }

    cmd[cmdLength] = '\0';

    write(shSocketFd_, (unsigned char *)cmd, cmdLength);

    return true;
}

//-----------------------------------------------------------------------
// This function initialize usarsim
//-----------------------------------------------------------------------
bool USARSimRobot::initializeUSARSim_(void)
{
    const struct timespec RETRY_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep
    int connStatus;
    bool done = false, robotStarted = false;

    while (!done)
    {
        // Connect to server.
        connStatus = connectToServer_(serverHostMachine_, serverPort_);

        switch (connStatus) {

        case INIT_CONNECTION_STATUS_NOT_READY:
            nanosleep(&RETRY_SLEEP_TIME_NSEC, NULL);
            break;

        case INIT_CONNECTION_STATUS_SUCCESS:
            printfTextWindow("USARSimRobot: Server connected.\n");
            done = true;
            break;

        default:
            printfTextWindow(
                "Error: USARSimRobot::initializeUSARSim(). Failed to connect to the server.\n");
            return false;
        }
    }

    robotStarted = startRobot_();

    return true;
}

//-----------------------------------------------------------------------
// This function updates the status bar.
//-----------------------------------------------------------------------
void USARSimRobot::updateStatusBar_(void)
{
    HSPose_t pose;
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = 0;

    if (mainThreadIsUp_)
    {
        strcat(statusStr, "USARSimRobot [");

        getPose(pose);

        sprintf(
            buf,
            "%s]: %c x: %.2f  y: %.2f z: %.2f t: %.2f pitch:%.2f",
            usarsimRobotName_.c_str(),
            statusbarSpinner_->getStatus(),
            pose.loc.x,
            pose.loc.y,
            pose.loc.z,
            pose.rot.yaw,
            pose.rot.pitch);

        if (DISPLAY_FREQUENCY)
        {
            getFrequency_(frequency);
            sprintf(buf, "%s : %.2f Hz", buf, frequency);
        }
    }
    else
    {
        sprintf(
            buf,
            "USARSimRobot: Connecting to the server on %s [port %d]...",
            serverHostMachine_.c_str(),
            serverPort_);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

//-----------------------------------------------------------------------
// This function starts the main thread.
//-----------------------------------------------------------------------
void *USARSimRobot::startMainThread_(void *usarsimInstance)
{
    ((USARSimRobot*) usarsimInstance)->mainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function starts the reader thread.
//-----------------------------------------------------------------------
void *USARSimRobot::startReaderThread_(void *usarsimInstance)
{
    ((USARSimRobot*) usarsimInstance)->readerLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to move USARSimRobot.
//-----------------------------------------------------------------------
void USARSimRobot::mainLoop_(void)
{
    int count = 0;
    
    mainThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the pose data from the usarsim.
        updatePose_();

        // Check for the pthread termination again.
        pthread_testcancel();

        if (REPORT_SONARS_)
        {
            // Read the sonar data from the usarsim.
            parseSonarData_();

            // Check for the pthread termination again.
            pthread_testcancel();
        }

        if (REPORT_RFID_)
        {
            // Read the RFID data from the usarsim.
            parseRFIDData_();

            // Check for the pthread termination again.
            pthread_testcancel();
    	}

        if (REPORT_CHEMICAL_)
        {
            // Read the Chemical data from the usarsim.
            parseChemicalData_();

            // Check for the pthread termination again.
            pthread_testcancel();
    	}

        if (REPORT_ENEMY_)
        {
            // Read the Enemy data from the usarsim.
            parseEnemyData_();

            // Check for the pthread termination again.
            pthread_testcancel();
    	}

        if (REPORT_COMM_)
        {
            // Read the Enemy data from the usarsim.
            parseCommData_();

            // Check for the pthread termination again.
            pthread_testcancel();
    	}

        // Compute the control speed.
        compControlSpeed_();

        // Control the robot.
        controlRobot_();

        // Compute the frequency.
        compFrequency_();

        // Compute the velocity.
        updateVelocity_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (gLogManager != NULL)
        {
            manageLog_(readerThread_);
        }

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        // Ease the computational load.
        usleep(MAIN_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function loops around to receive the sensor from the server.
//-----------------------------------------------------------------------
void USARSimRobot::readerLoop_(void)
{
    readerThreadIsUp_ = true;

    while(true)
    {
        // Make sure termination was not requested.
        pthread_testcancel();

        // Read the sensor data.
        readSensorData_();

        // Check for the pthread termination again.
        pthread_testcancel();

        // Ease the computational load.
        usleep(MAIN_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function extracts the values in the string sperated by commas.
//-----------------------------------------------------------------------
vector<string> USARSimRobot::extractCommaSepratedValues_(string values)
{
    vector<string> valueList;
    string value; 
    string::size_type index;
    
    while (true)
    {
        index = values.find(CHAR_COMMA_);

        if (index == string::npos)
        {
            valueList.push_back(values);
            break;
        }

        value = values.substr(0, index);
        valueList.push_back(value);

        values = values.substr(index+1, string::npos);
    }

    return valueList;
}

//-----------------------------------------------------------------------
// This function waits for the sensor data to arrive.
//-----------------------------------------------------------------------
void USARSimRobot::waitForSensorData_(void)
{
    bool arrived = false;
    const struct timespec WAITER_SLEEP_TIME_NSEC = {1, 0}; // 1 second sleep

    while (true)
    {
        pthread_mutex_lock(&sensorDataTableMutex_);

        if (shSensorDataTable_.size() > 0)
        {
            arrived = true;
        }

        pthread_mutex_unlock(&sensorDataTableMutex_);

        if (arrived)
        {
            break;
        }

        nanosleep(&WAITER_SLEEP_TIME_NSEC, NULL);
    }
}

//-----------------------------------------------------------------------
// This function updates the current pose.
//-----------------------------------------------------------------------
void USARSimRobot::updatePose_(void)
{
    HSPose_t pose, lastPose;
    HSSpeed_t controlSpeed;
    DataLine_t odoDataLine;
    map<string, DataLine_t>::iterator itr;
    string poseString = EMPTY_STRING_;
    vector<string> poseList;
    double dist;
    double dt, curTime;
    static double lastTime = 0;

    memset((HSPose_t *)&pose, 0x0, sizeof(HSPose_t));
    getLatestRawPose_(lastPose);

    switch (robotSpecIndex_) {

    case USARSIM_ROBOT_CLASS_P2AT:
    case USARSIM_ROBOT_CLASS_P2DX:
    case USARSIM_ROBOT_CLASS_ATRVJR:
    case USARSIM_ROBOT_CLASS_HMMWV:
    case USARSIM_ROBOT_CLASS_TALON:
        parsePositionData_(pose);
        break;

    case USARSIM_ROBOT_CLASS_AIRROBOT:
    case USARSIM_ROBOT_CLASS_DRAGONFLYROBOT:
        parsePositionData_(pose);

        // Odometry & GroundTruth supply only x,y, & theta (yaw); have to maintain state from last time for others
        //pose.loc.z = lastPose.loc.z;
        //pose.rot.pitch = lastPose.rot.pitch;
        //pose.rot.roll = lastPose.rot.roll;
/*
        // No altitude (z) odometry info available. Simulate it.
        curTime = getCurrentEpochTime();

        if (lastTime > 0)
        {
            dt = curTime - lastTime;
            Robot::getControlSpeed_(controlSpeed);
            Robot::advancePose2D_(controlSpeed, dt);
            //Robot::advancePoseAltitude_(pose, controlSpeed, dt);
            Robot::getPose(pose);
        }
        lastTime = curTime;
*/
        break;

    case USARSIM_ROBOT_CLASS_TARANTULA:
        break;

    case USARSIM_ROBOT_CLASS_ERS:
        break;
    }

    dist = sqrt(pow((pose.loc.x - lastPose.loc.x), 2.0) + pow((pose.loc.y - lastPose.loc.y), 2.0) + pow((pose.loc.z - lastPose.loc.z), 2.0));
    addTravelDist_(dist);
    setLatestRawPose_(pose);

    updatePoseFromUSARSimPoseOffset_();
}

//-----------------------------------------------------------------------
// This function parses the position data from the sensor buffer.
//-----------------------------------------------------------------------
void USARSimRobot::parsePositionData_(HSPose_t &pose)
{
    DataLine_t odoDataLine;
    map<string, DataLine_t>::iterator itr;
    string poseString = EMPTY_STRING_;
    vector<string> poseList;
    double curTime;
    int i;
    const bool DEBUG_PARSE_POSITION_DATA = false;
    bool useGroundTruth = true;

    memset((HSPose_t *)&pose, 0x0, sizeof(HSPose_t));
    pthread_mutex_lock(&sensorDataTableMutex_);

    // Extract the location from the odometry or ground truth
    if(robotSpecIndex_ == USARSIM_ROBOT_CLASS_AIRROBOT || robotSpecIndex_ == USARSIM_ROBOT_CLASS_DRAGONFLYROBOT) {
      useGroundTruth = true;
      itr = shSensorDataTable_.find(DATA_VALUE_GROUND_TRUTH_);
    }
    else
      itr = shSensorDataTable_.find(DATA_VALUE_ODOMETRY_);

    if (itr != shSensorDataTable_.end())
    {
        odoDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);

    if (odoDataLine.size() == 0)
    {
        if (DEBUG_PARSE_POSITION_DATA)
        {
            printfTextWindow(
                "Error: USARSimRobot::parsePositionData_(). Sensor data not available.\n");
        }

        // Data not available. Abort the computation.
        return;
    }

    // Extract the x, y, and theta.
    for (i = 0; i < ((int)(odoDataLine.size()) - 1); i++)
    {
        if (odoDataLine[i] == DATA_KEY_POSE_ || odoDataLine[i] == DATA_KEY_LOCATION_)
        {
            poseString = odoDataLine[i+1];
        }
    }

    if (poseString == EMPTY_STRING_)
    {
        printfTextWindow(
            "Error: USARSimRobot::parsePositionData_(). Position data not available.\n");
        return;
    }

    poseList = extractCommaSepratedValues_(poseString);

    if (poseList.size() != 3)
    {
        printfTextWindow(
            "Error: USARSimRobot::parsePositionData_(). Position data corrupted.\n");
        return;
    }

    if (useGroundTruth == false) {

      // Odometry provides x, y, and theta (yaw)

      // Note: USARSim's Y and Yaw are opposite of MissionLab's.
      pose.loc.x = atof(poseList[0].c_str());
      pose.loc.y = -atof(poseList[1].c_str());
      pose.rot.yaw = -atof(poseList[2].c_str());
      pose.rot.yaw = ROBOT_RAD2DEG(pose.rot.yaw);
      ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);
    } else {
  
      // Ground truth provides x,y,z & roll,pitch,yaw
      pose.loc.x = atof(poseList[0].c_str());
      pose.loc.y = -atof(poseList[1].c_str());
      pose.loc.z = -atof(poseList[2].c_str());

      // Extract roll,pitch,yaw
      poseString = EMPTY_STRING_;
      for (i = 0; i < ((int)(odoDataLine.size()) - 1); i++)
      {
        if (odoDataLine[i] == DATA_KEY_POSE_ || odoDataLine[i] == DATA_KEY_ORIENTATION_)
        {
            poseString = odoDataLine[i+1];
        }
      }

      if (poseString == EMPTY_STRING_)
      {
        printfTextWindow(
            "Error: USARSimRobot::parsePositionData_(). Ground truth orientation data not available.\n");
        return;
      }

      poseList = extractCommaSepratedValues_(poseString);

      if (poseList.size() != 3)
      {
        printfTextWindow(
            "Error: USARSimRobot::parsePositionData_(). Ground truth orientation data corrupted.\n");
        return;
      }
      pose.rot.roll = atof(poseList[0].c_str());
      pose.rot.roll = ROBOT_RAD2DEG(pose.rot.roll);
      ROBOT_CRAMPDEG(pose.rot.roll, -180.0, 180.0);

      pose.rot.pitch = atof(poseList[1].c_str());
      pose.rot.pitch = ROBOT_RAD2DEG(pose.rot.pitch);
      ROBOT_CRAMPDEG(pose.rot.pitch, -180.0, 180.0);

      pose.rot.yaw = -atof(poseList[2].c_str());
      pose.rot.yaw = ROBOT_RAD2DEG(pose.rot.yaw);
      ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);

    }

    curTime = getCurrentEpochTime();
    pose.extra.time = curTime;
    pose.loc.extra.time = curTime;
    pose.rot.extra.time = curTime;

    return;
}

//-----------------------------------------------------------------------
// This function parses the sonar data from the sensor buffer.
//-----------------------------------------------------------------------
void USARSimRobot::parseSonarData_(void)
{
    DataLine_t sonarDataLine;
    HSPose_t pose;
    string readingString;
    map<string, DataLine_t>::iterator itr;
    int i, sonarCount;
    const bool DEBUG_PARSE_SONAR_DATA = false;

    memset(sonarReadings_, 0x0, sizeof(float)*numSonars_);

    pthread_mutex_lock(&sensorDataTableMutex_);

    itr = shSensorDataTable_.find(DATA_VALUE_SONAR_);

    if (itr != shSensorDataTable_.end())
    {
        sonarDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);
    
    getPose(pose);
    sonarCount = 0;

    // Extract the sonar readings.
    for (i = 0; i < ((int)(sonarDataLine.size()) - 1); i++)
    {
        if (sonarDataLine[i] == DATA_KEY_RANGE_)
        {
            readingString = sonarDataLine[i+1];

            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data.\n",
                    __FUNCTION__);
                return;
            }

            sonarReadings_[sonarCount] = 
                atof(readingString.c_str()) +
                sonarMountDist_[sonarCount];
            
            sonarLocs_[sonarCount][0] = pose.loc.x;
            sonarLocs_[sonarCount][1] = pose.loc.y;
            sonarLocs_[sonarCount][2] = pose.rot.yaw;

            sonarCount++;

            if (sonarCount == numSonars_)
            {
                break;
            }
        }
    }

    if (sonarCount != numSonars_)
    {
        if (DEBUG_PARSE_SONAR_DATA)
        {
            printfTextWindow(
                "Error: USARSimRobot::%s(). Incorrect number of sonars.\n",
                __FUNCTION__);
        }

        return;
    }
}

//-----------------------------------------------------------------------
// This function parses the RFID data from the sensor buffer.
//-----------------------------------------------------------------------
void USARSimRobot::parseRFIDData_(void)
{
    DataLine_t rfidDataLine;
    HSPose_t pose;
    string readingString;
    map<string, DataLine_t>::iterator itr;
    //why is rfidCount here?    
    int i, rfidCount, rfidID = -1, rfidIndex = -1;
    const bool DEBUG_PARSE_RFID_DATA = false;

    // Zero out RFID data (only currently detected ones should have values)
    memset(RFIDData_, 0x0, sizeof(float)*numRFID_*4);

    pthread_mutex_lock(&sensorDataTableMutex_);

    // Get all of the sensor data lines containing RFID data
    itr = shSensorDataTable_.find(DATA_VALUE_RFID_);

    if (itr != shSensorDataTable_.end())
    {
        rfidDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);
    
    rfidCount = 0;

    // Extract the RFID readings.
    for (i = 0; i < ((int)(rfidDataLine.size()) - 1); i++)
    {
//        rfidID = -1;  // -1 for not available
        rfidIndex = -1;  // -1 for not available

        // Get the RFID ID
        if (rfidDataLine[i] == DATA_KEY_RFID_ID_)
        {
            readingString = rfidDataLine[i+1];

            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data.\n",
                    __FUNCTION__);
                return;
            }
 
            rfidID = atoi(readingString.c_str());
        } else if (rfidDataLine[i] == DATA_KEY_RFID_LOCATION_)
        { 
            if (rfidID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (RFID ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // RFID ID #s have to be less than the MAX_NUM_RFID constant
            if (rfidID >= USARSIM_ROBOT_MAX_NUM_RFID) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). RFID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_RFID);
                return;
            }

            // Starting index of the float array (since each RFID element packs four data items
            // (ID, x, y, z)
            rfidIndex = rfidID * 4;

            RFIDData_[rfidIndex] = 1; // Signify that it is being detected

            // X,Y,Z coordinates of RFID  (one by one)
            for (int count = 1; count < 4; count++) { 
              readingString = rfidDataLine[i+count];

              // Make sure it's a number
              if ((readingString.size() == 0) || (!isdigit(readingString[0])))
              {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted RFID coordinate data.\n",
                    __FUNCTION__);
                return;
              }

              RFIDData_[rfidIndex+count] = atof(readingString.c_str());
            }
        }
    }

   // <REMOVE ME!>
   // Made-up data! The code that actually attempts to parse RFID data is above, but UNTESTED
//   for (int i = 0; i < numRFID_; i++) {
//     rfidIndex = i*4;
//     RFIDData_[rfidIndex]   = i;  // ID
//     RFIDData_[rfidIndex+1] = i*1;  // X
//     RFIDData_[rfidIndex+2] = i*2;  // Y
//     RFIDData_[rfidIndex+3] = i*3;  // Z
//   }

}

//-----------------------------------------------------------------------
// This function parses the Enemy data from the sensor buffer.
//-----------------------------------------------------------------------
void USARSimRobot::parseEnemyData_(void)
{
    DataLine_t enemyDataLine;
    HSPose_t pose;
    string readingString, valueString;
    vector<string> valueList;
    map<string, DataLine_t>::iterator itr;
    int i, numReadings, enemyID = -1, enemyIndex = -1;
    const bool DEBUG_PARSE_ENEMY_DATA = false;

    // Zero out enemy data (only currently detected ones should have values)
    memset(enemyData_, 0x0, sizeof(float)*(numEnemy_*7 + 1));

    pthread_mutex_lock(&sensorDataTableMutex_);

    // Get all of the sensor data lines containing enemy data
    itr = shSensorDataTable_.find(DATA_VALUE_ENEMY_);

    if (itr != shSensorDataTable_.end())
    {
        enemyDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);
    static float lastUTTime = -1.0;
    // Extract the enemy readings.
    for (i = 0; i < ((int)(enemyDataLine.size()) - 1); i++)
    {
        //printfTextWindow("%s\n", enemyDataLine[i].c_str());
        // Get the enemy ID
        if (enemyDataLine[i] == DATA_KEY_ENEMY_ID_)
        {
            readingString = enemyDataLine[i+1];

            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted id data.\n",
                    __FUNCTION__);
                return;
            }
 
            enemyID = atoi(readingString.c_str());
            assert(enemyID < numEnemy_);
            // Starting index of the float array (each ID has 7 elements stored with it)
            enemyIndex = enemyID * 7 + 1; // + 1 to account for timestamp
            enemyData_[enemyIndex] = 1; // Signify that it is being detected
        } else if (enemyDataLine[i] == DATA_KEY_ENEMY_LOCATION_)
        { 
            if (enemyID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Enemy ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Enemy ID #s have to be less than the MAX_NUM_ENEMY constant
            if (enemyID >= USARSIM_ROBOT_MAX_NUM_ENEMY) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Enemy ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_ENEMY);
                return;
            }

            // get the actual data line
            readingString = enemyDataLine[i+1];

            // create a new array with the x, y, z values by extracting comma separated values
            valueList = extractCommaSepratedValues_(readingString);
            numReadings = valueList.size();

            if (numReadings <= 0)
            {
                // No location readings.
                return;
            }
            
            // Copy the data.
            for (int count = 0; count < numReadings; count++)
            {
                valueString = valueList[count];

                if ((valueString.size() == 0))
                {
                    printfTextWindow(
                        "Error: USARSimRobot::%s(). Corrupted location data.\n",
                        __FUNCTION__);
                    return;
                }
                // copy the locations into the enemyData array (float) from the temporary valueList array (string)
                // + 2 used since x is the 3rd element in the array (id dist x ...)
                enemyData_[enemyIndex+count+2] = atof(valueString.c_str());
            }
            double tx = enemyData_[enemyIndex+2], ty = enemyData_[enemyIndex+2+1];
            double c,s;
            double tx1, ty1;
            HSPose_t pose;
            getLatestRawPose_(pose);
            tx1 = tx - pose.loc.x;
            ty1 = -ty - pose.loc.y;
            s = sin(-1 * pose.rot.yaw*M_PI/180.0);
            c = cos(-1 * pose.rot.yaw*M_PI/180.0);
            enemyData_[enemyIndex+2] = tx1*c - ty1*s;
            enemyData_[enemyIndex+2+1] = tx1*s + ty1*c;

            //printfTextWindow("%f,%f == (%f,%f,%f) == %f,%f ==> %f,%f", tx, ty, pose.loc.x, pose.loc.y, pose.rot.yaw, tx1, ty1, enemyData_[enemyIndex+2], enemyData_[enemyIndex+2+1]);
        } else if (enemyDataLine[i] == DATA_KEY_ENEMY_DISTANCE_)
        {
            if (enemyID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Enemy ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Enemy ID #s have to be less than the MAX_NUM_ENEMY constant
            if (enemyID >= USARSIM_ROBOT_MAX_NUM_ENEMY) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Enemy ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_ENEMY);
                return;
            }

            readingString = enemyDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted distance data.\n",
                    __FUNCTION__);
                return;
            }
            // add the sensed distance to the data array in the proper location
            enemyData_[enemyIndex+1] = atof(readingString.c_str());

        } else if (enemyDataLine[i] == DATA_KEY_ENEMY_PODENEMY_)
        {
            if (enemyID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Enemy ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Enemy ID #s have to be less than the MAX_NUM_ENEMY constant
            if (enemyID >= USARSIM_ROBOT_MAX_NUM_ENEMY) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Enemy ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_ENEMY);
                return;
            }

            readingString = enemyDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted PODenemy data.\n",
                    __FUNCTION__);
                return;
            }
            // add the sensed PODE to the data array in the proper location
            enemyData_[enemyIndex+5] = atof(readingString.c_str());

        } else if (enemyDataLine[i] == DATA_KEY_ENEMY_PODSELF_)
        {
            if (enemyID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Enemy ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Enemy ID #s have to be less than the MAX_NUM_ENEMY constant
            if (enemyID >= USARSIM_ROBOT_MAX_NUM_ENEMY) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Enemy ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_ENEMY);
                return;
            }

            readingString = enemyDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted PODself data.\n",
                    __FUNCTION__);
                return;
            }
            // add the sensed PODSelf to the data array in the proper location
            enemyData_[enemyIndex+6] = atof(readingString.c_str());

        } else if (enemyDataLine[i] == DATA_KEY_TIME_)
        {
            readingString = enemyDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted time data.\n",
                    __FUNCTION__);
                return;
            }
            // add the timestamp to the data array in the proper location
            enemyData_[0] = atof(readingString.c_str());

            // TODO : REMOVE : There is a bug in USARSim that send the same data over and over if 
            // nothing is detected.  This buffers against it, but it should be fixed on the USARSim
            // side
            if (enemyData_[0] < lastUTTime) {
              //memset(enemyData_, 0x0, sizeof(float)*(numEnemy_*7 + 1));
              break;
            } else {
              lastUTTime = enemyData_[0];
            }
        } //end if
    }


//    for (int z=1; z < numEnemy_*7+1; z+=7)
//      if (enemyData_[z] == 1)
//       printfTextWindow("%f: enemy %d found at dist=%f loc=(%f,%f,%f) pode=%f pods=%f", enemyData_[0], (z-1)/7, enemyData_[z+1], enemyData_[z+2], enemyData_[z+3], enemyData_[z+4], enemyData_[z+5], enemyData_[z+6]);

}

//-----------------------------------------------------------------------
// This function parses the Chemical data from the sensor buffer.
//-----------------------------------------------------------------------
void USARSimRobot::parseChemicalData_(void)
{
    DataLine_t chemicalDataLine;
    HSPose_t pose;
    string readingString, valueString;
    vector<string> valueList;
    map<string, DataLine_t>::iterator itr;
    int i, numReadings, chemicalID = -1, chemicalIndex = -1;
    const bool DEBUG_PARSE_CHEMICAL_DATA = false;

    // Zero out chemical data (only currently detected ones should have values)
    memset(chemicalData_, 0x0, sizeof(float)*(numChemical_*6 + 1));

    pthread_mutex_lock(&sensorDataTableMutex_);

    // Get all of the sensor data lines containing chemical data
    itr = shSensorDataTable_.find(DATA_VALUE_CHEMICAL_);

    if (itr != shSensorDataTable_.end())
    {
        chemicalDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);

    // Extract the chemical readings.
    static float lastUTTime = -1.0;
    for (i = 0; i < ((int)(chemicalDataLine.size()) - 1); i++)
    {
        // Get the chemical ID
        if (chemicalDataLine[i] == DATA_KEY_CHEMICAL_ID_)
        {
            readingString = chemicalDataLine[i+1];

            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted id data.\n",
                    __FUNCTION__);
                return;
            }
 
            chemicalID = atoi(readingString.c_str());
            assert(chemicalID < numChemical_);
            // Starting index of the float array (each ID has 6 elements stored with it)
            chemicalIndex = chemicalID * 6 + 1; // + 1 to account for timestamp
            chemicalData_[chemicalIndex] = 1; // Signify that it is being detected
        } else if (chemicalDataLine[i] == DATA_KEY_CHEMICAL_LOCATION_)
        { 
            if (chemicalID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Chemical ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Chemical ID #s have to be less than the MAX_NUM_CHEMICAL constant
            if (chemicalID >= USARSIM_ROBOT_MAX_NUM_CHEMICAL) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Chemical ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_CHEMICAL);
                return;
            }

            // get the actual data line
            readingString = chemicalDataLine[i+1];

            // create a new array with the x, y, z values by extracting comma separated values
            valueList = extractCommaSepratedValues_(readingString);
            numReadings = valueList.size();

            if (numReadings <= 0)
            {
                // No location readings.
                return;
            }
            
            // Copy the data.
            for (int count = 0; count < numReadings; count++)
            {
                valueString = valueList[count];

                if ((valueString.size() == 0))
                {
                    printfTextWindow(
                        "Error: USARSimRobot::%s(). Corrupted location data.\n",
                        __FUNCTION__);
                    return;
                }
                // copy the locations into the chemicalData array (float) from the temporary valueList array (string)
                // + 2 used since x is the third element in the array 
                chemicalData_[chemicalIndex+count+2] = atof(valueString.c_str());
            }
            double tx = chemicalData_[chemicalIndex+2], ty = chemicalData_[chemicalIndex+2+1];
            double c,s;
            double tx1, ty1;
            HSPose_t pose;
            getLatestRawPose_(pose);
            tx1 = tx - pose.loc.x;
            ty1 = -ty - pose.loc.y;
            s = sin(-1 * pose.rot.yaw*M_PI/180.0);
            c = cos(-1 * pose.rot.yaw*M_PI/180.0);
            chemicalData_[chemicalIndex+2] = tx1*c - ty1*s;
            chemicalData_[chemicalIndex+2+1] = tx1*s + ty1*c;


        } else if (chemicalDataLine[i] == DATA_KEY_CHEMICAL_DISTANCE_)
        {
            if (chemicalID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Chemical ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Chemical ID #s have to be less than the MAX_NUM_CHEMICAL constant
            if (chemicalID >= USARSIM_ROBOT_MAX_NUM_CHEMICAL) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Chemical ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_CHEMICAL);
                return;
            }

            readingString = chemicalDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted distance data.\n",
                    __FUNCTION__);
                return;
            }
            // add the sensed distance to the data array in the proper location
            chemicalData_[chemicalIndex+1] = atof(readingString.c_str());

        } else if (chemicalDataLine[i] == DATA_KEY_CHEMICAL_PODCHEM_)
        {
            if (chemicalID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Chemical ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Chemical ID #s have to be less than the MAX_NUM_CHEMICAL constant
            if (chemicalID >= USARSIM_ROBOT_MAX_NUM_CHEMICAL) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Chemical ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_CHEMICAL);
                return;
            }

            readingString = chemicalDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted PODchem data.\n",
                    __FUNCTION__);
                return;
            }
            // add the sensed PODC to the data array in the proper location
            chemicalData_[chemicalIndex+5] = atof(readingString.c_str());

        } else if (chemicalDataLine[i] == DATA_KEY_TIME_)
        {
            readingString = chemicalDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted time data.\n",
                    __FUNCTION__);
                return;
            }
            // add the timestamp to the data array in the proper location
            chemicalData_[0] = atof(readingString.c_str());

            // TODO : REMOVE : There is a bug in USARSim that send the same data over and over if
            // nothing is detected.  This buffers against it, but it should be fixed on the USARSim
            // side
            if (chemicalData_[0] < lastUTTime) {
              //memset(chemicalData_+1, 0x0, sizeof(float)*(numChemical_*6 + 1));
              break;
            } else {
              lastUTTime = chemicalData_[0];
            }

        } //end if
    }

//    for (int z=1; z < numChemical_*6+1; z+=6)
//      if (chemicalData_[z] == 1)
//       printfTextWindow("%f: chem %d found at dist=%f loc=(%f,%f,%f) pode=%f pods=%f", chemicalData_[0], (z-1)/7, chemicalData_[z+1], chemicalData_[z+2], chemicalData_[z+3], chemicalData_[z+4], chemicalData_[z+5], chemicalData_[z+6]);

}


//-----------------------------------------------------------------------
// This function parses the Comm data from the sensor buffer.
//-----------------------------------------------------------------------
void USARSimRobot::parseCommData_(void)
{
    DataLine_t commDataLine;
    HSPose_t pose;
    string readingString, valueString;
    vector<string> valueList;
    map<string, DataLine_t>::iterator itr;
    int i, commID = -1, commIndex = -1;
    const bool DEBUG_PARSE_COMM_DATA = false;

    // Zero out comm data (only currently detected ones should have values)
    memset(commData_, 0x0, sizeof(float)*(numComm_*2 + 1));

    pthread_mutex_lock(&sensorDataTableMutex_);

    // Get all of the sensor data lines containing comm data
    itr = shSensorDataTable_.find(DATA_VALUE_COMM_);

    if (itr != shSensorDataTable_.end())
    {
        commDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);
    
    // Extract the comm readings.
    for (i = 0; i < ((int)(commDataLine.size()) - 1); i++)
    {
        // Get the comm ID (BUGNUM)
        if (commDataLine[i] == DATA_KEY_COMM_BUGNUM_)
        {
            readingString = commDataLine[i+1];

            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted id data.\n",
                    __FUNCTION__);
                return;
            }
 
            commID = atoi(readingString.c_str());
            // Starting index of the float array (each ID has 2 elements stored with it)
            commIndex = commID * 2 + 1; // + 1 to account for timestamp
            commData_[commIndex] = 1; // Signify that it is being detected

        } else if (commDataLine[i] == DATA_KEY_COMM_PD_)
        {
            if (commID == -1) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data (Comm ID did not come first).\n",
                    __FUNCTION__);
                return;
            }

            // Comm ID #s have to be less than the MAX_NUM_BUG constant
            if (commID >= USARSIM_ROBOT_MAX_NUM_BUG) {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Comm ID is greater than the max allowed (%d)\n",
                    __FUNCTION__, USARSIM_ROBOT_MAX_NUM_BUG);
                return;
            }

            readingString = commDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted pathloss data.\n",
                    __FUNCTION__);
                return;
            }
            // add the sensed pd (pathloss) to the data array in the proper location
            commData_[commIndex+1] = atof(readingString.c_str());

        } else if (commDataLine[i] == DATA_KEY_TIME_)
        {
            readingString = commDataLine[i+1];

            if ((readingString.size() == 0))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted time data.\n",
                    __FUNCTION__);
                return;
            }
            // add the timestamp to the data array in the proper location
            commData_[0] = atof(readingString.c_str());
        } //end if
    } // end loop

}


//-----------------------------------------------------------------------
// This function parses the laser data from the sensor buffer, and return
// it.
//-----------------------------------------------------------------------
int USARSimRobot::getLaserData(float **laserReadings, float **laserAngles, const char *laserName)
{
    DataLine_t laserDataLine;
    string readingString, valueString;
    vector<string> valueList;
    map<string, DataLine_t>::iterator itr;
    float fov, inc;
    int i, numReadings;

    pthread_mutex_lock(&sensorDataTableMutex_);

    itr = shSensorDataTable_.find(laserName);

    if (itr != shSensorDataTable_.end())
    {
        laserDataLine = itr->second;
    }

    pthread_mutex_unlock(&sensorDataTableMutex_);
    
    // Extract the laser readings.
    for (i = 0; i < ((int)(laserDataLine.size()) - 1); i++)
    {
        if (laserDataLine[i] == DATA_KEY_RANGE_)
        {
            readingString = laserDataLine[i+1];

            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data.\n",
                    __FUNCTION__);
                return 0;
            }

            valueList = extractCommaSepratedValues_(readingString);
        }
        else if (laserDataLine[i] == DATA_KEY_FOV_)
        {
            readingString = laserDataLine[i+1];
            
            if ((readingString.size() == 0) || (!isdigit(readingString[0])))
            {
                printfTextWindow(
                    "Error: USARSimRobot::%s(). Corrupted data.\n",
                    __FUNCTION__);
                return 0;
            }

            fov = atof(readingString.c_str());
            fov = USARSIM_ROBOT_RAD2DEG(fov);
        }
    }

    numReadings = valueList.size();

    if (numReadings <= 0)
    {
        // No reading.
        return 0;
    }

    inc = (numReadings <= 1)? 0 : (fov/(float)(numReadings - 1));

    if ((*laserReadings) == NULL)
    {
        (*laserReadings) = new float[numReadings];
        memset((*laserReadings), 0x0, sizeof(float)*numReadings);
    }

    if ((*laserAngles) == NULL)
    {
        (*laserAngles) = new float[numReadings];
        memset((*laserAngles), 0x0, sizeof(float)*numReadings);
    }

    // Copy the data.
    for (i = 0; i < numReadings; i++)
    {
        valueString = valueList[i];

        if ((valueString.size() == 0) || (!isdigit(valueString[0])))
        {
            printfTextWindow(
                "Error: USARSimRobot::%s(). Corrupted data.\n",
                __FUNCTION__);
            return 0;
        }

        (*laserReadings)[i] = atof(valueString.c_str());
        (*laserAngles)[i] = -fov/2.0+(inc*(float)i);
    }

    return numReadings;
}

//-----------------------------------------------------------------------
// This function sets the offset between the raw usarsim data and the
// current robot pose.
//-----------------------------------------------------------------------
void USARSimRobot::setUSARSimPoseOffset_(HSPose_t poseOffset)
{
    pthread_mutex_lock(&usarsimPoseOffsetMutex_);
    shUSARSimPoseOffset_ = poseOffset;
    pthread_mutex_unlock(&usarsimPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function returns the offset between the raw usarsim data and the
// current robot pose.
//-----------------------------------------------------------------------
void USARSimRobot::getUSARSimPoseOffset_(HSPose_t &poseOffset)
{
    pthread_mutex_lock(&usarsimPoseOffsetMutex_);
    poseOffset = shUSARSimPoseOffset_;
    pthread_mutex_unlock(&usarsimPoseOffsetMutex_);
}

//-----------------------------------------------------------------------
// This function saves the latest raw pose data.
//-----------------------------------------------------------------------
void USARSimRobot::setLatestRawPose_(HSPose_t pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    shLatestRawPose_ = pose;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function returns the saved latest raw pose data.
//-----------------------------------------------------------------------
void USARSimRobot::getLatestRawPose_(HSPose_t &pose)
{
    pthread_mutex_lock(&latestRawPoseMutex_);
    pose = shLatestRawPose_;
    pthread_mutex_unlock(&latestRawPoseMutex_);
}

//-----------------------------------------------------------------------
// This function updates the robot pose from the offset.
//-----------------------------------------------------------------------
void USARSimRobot::updatePoseFromUSARSimPoseOffset_(void)
{
    HSPose_t pose;
    HSPose_t poseOffset;
    double x, y, z, angle, altAngle;

    getLatestRawPose_(pose);
    getUSARSimPoseOffset_(poseOffset);

    x = pose.loc.x;
    y = pose.loc.y;
    z = pose.loc.z;

    angle = poseOffset.rot.yaw;
    ROBOT_CRAMPDEG(angle, -180.0, 180.0);
    angle = ROBOT_DEG2RAD(angle);

    pose.loc.x = (x*cos(angle)) - (y*sin(angle)) + poseOffset.loc.x;
    pose.loc.y = (x*sin(angle)) + (y*cos(angle)) + poseOffset.loc.y;
    pose.rot.yaw += poseOffset.rot.yaw;
    ROBOT_CRAMPDEG(pose.rot.yaw, -180.0, 180.0);
    pose.rot.pitch += poseOffset.rot.pitch;
    ROBOT_CRAMPDEG(pose.rot.pitch, -180.0, 180.0);

    setPose_(pose);
}

//-----------------------------------------------------------------------
// This function generates ERS gate control command.
//-----------------------------------------------------------------------
string USARSimRobot::generateERSGateCommand_(HSSpeed_t controlSpeed)
{
    string cmdString;
    char cmd[MAX_CMD_LENGTH_ + 1];
    double rfa, rfc, lfa, lfc, rra, rrc, lra, lrc;
    double curTime, phaseInterval;
    double driveSpeed = 0;
    static int ersGatePhase = -1;
    static double lastTime = 0;
    const int NUM_PHASES_PER_SEC = 40;
    double ERS_MAX_DRIVE_SPEED_ = 0.07;


    curTime = getCurrentEpochTime();

    if (fabs(controlSpeed.driveSpeed.value) == 0)
    {
        rfa = 0;
        rfc = 0;
        lfa = 0;
        lfc = 0;
        rra = 0;
        rrc = 0;
        lra = 0;
        lrc = 0;
    }
    else
    {
        driveSpeed = controlSpeed.driveSpeed.value;

        printfTextWindow("driveSpeed = %f\n", driveSpeed);

        driveSpeed = (driveSpeed <= ERS_MAX_DRIVE_SPEED_)? driveSpeed : ERS_MAX_DRIVE_SPEED_;

        phaseInterval = 1.0/(driveSpeed*NUM_PHASES_PER_SEC);

        switch (ersGatePhase) {

        case ERS_LOWSPEED_GATE_PHASE_0:
            rfa = 0;
            rfc = 0;
            lfa = 0;
            lfc = 0;
            rra = -60;
            rrc = 60;
            lra = -60;
            lrc = 60;

            if ((curTime - lastTime) > phaseInterval)
            {
                ersGatePhase++;
                lastTime = curTime;
            }
            break;

        case ERS_LOWSPEED_GATE_PHASE_1:
            rfa = 90;
            rfc = 0;
            lfa = 90;
            lfc = 0;
            rra = 0;
            rrc = 0;
            lra = 0;
            lrc = 0;

            if ((curTime - lastTime) > phaseInterval)
            {
                ersGatePhase++;
                lastTime = curTime;
            }
            break;

        case ERS_LOWSPEED_GATE_PHASE_2:
            rfa = 0;
            rfc = 0;
            lfa = 0;
            lfc = 0;
            rra = 0;
            rrc = 0;
            lra = 0;
            lrc = 0;

            if ((curTime - lastTime) > phaseInterval)
            {
                ersGatePhase++;
                //ersGatePhase = ERS_LOWSPEED_GATE_PHASE_0;
                lastTime = curTime;
            }
            break;

        case ERS_LOWSPEED_GATE_PHASE_3:
            rfa = 0;
            rfc = 0;
            lfa = 0;
            lfc = 0;
            rra = -60;
            rrc = 60;
            lra = 0;
            lrc = 0;

            if ((curTime - lastTime) > phaseInterval)
            {
                ersGatePhase = ERS_LOWSPEED_GATE_PHASE_0;
                lastTime = curTime;
            }
            break;

        default:
            ersGatePhase = ERS_LOWSPEED_GATE_PHASE_0;
            lastTime = curTime;
            break;
        }

    }

    sprintf(
        cmd,
        "%s {RFA %.2f} {RFC %.2f} {LFA %.2f} {LFC %.2f} {RRA %.2f} {RRC %.2f} {LRA %.2f} {LRC %.2f}",
        CMD_KEY_MULTIDRIVE_.c_str(),
        ROBOT_DEG2RAD(rfa),
        ROBOT_DEG2RAD(rfc),
        ROBOT_DEG2RAD(lfa),
        ROBOT_DEG2RAD(lfc),
        ROBOT_DEG2RAD(rra),
        ROBOT_DEG2RAD(rrc),
        ROBOT_DEG2RAD(lra),
        ROBOT_DEG2RAD(lrc));

    cmdString = cmd;

    return cmdString;
}

//-----------------------------------------------------------------------
// This function controls the robot with the control speed.
//-----------------------------------------------------------------------
void USARSimRobot::controlRobot_(void)
{
    HSSpeed_t controlSpeed;
    string bufString, cmdString;
    float vt, va, rr, rt, leftSpeed, rightSpeed;
    float altVelocity, linearVelocity, latVelocity, rotVelocity;
    char cmd[MAX_CMD_LENGTH_ + 1];
    int cmdLength;
    bool normalized = false;

    getControlSpeed_(controlSpeed);

    memset(cmd, 0x0, MAX_CMD_LENGTH_);

    switch (robotSpecIndex_) {
   
    case USARSIM_ROBOT_CLASS_P2AT:
    case USARSIM_ROBOT_CLASS_P2DX:
    case USARSIM_ROBOT_CLASS_ATRVJR:
    case USARSIM_ROBOT_CLASS_HMMWV:
    case USARSIM_ROBOT_CLASS_TALON:
    case USARSIM_ROBOT_CLASS_TARANTULA:
        vt = controlSpeed.driveSpeed.value;
        va = ROBOT_DEG2RAD(controlSpeed.steerSpeed.value);
        rr = ROBOT_SPECS_[robotSpecIndex_].robotRadius;
        rt = ROBOT_SPECS_[robotSpecIndex_].wheelRadius;

        if ((rr <= 0) || (rt <= 0))
        {
            printfTextWindow("Error:USARSimRobot::controlRobot_(). Invalid radius.\n");
            return;
        }
        leftSpeed = (vt - rr*va)/rt;
        rightSpeed = (vt + rr*va)/rt;

        sprintf(
            cmd,
            "%s {%s %.2f} {%s %.2f}",
            CMD_KEY_DRIVE_.c_str(),
            CMD_STRING_LEFT_.c_str(),
            leftSpeed,
            CMD_STRING_RIGHT_.c_str(),
            rightSpeed);
        break;

    case USARSIM_ROBOT_CLASS_AIRROBOT:
    case USARSIM_ROBOT_CLASS_DRAGONFLYROBOT:
        altVelocity = controlSpeed.altitudeSpeed.value;
        linearVelocity = controlSpeed.driveSpeed.value;
        latVelocity = 0;
        rotVelocity = -USARSIM_ROBOT_DEG2RAD(controlSpeed.steerSpeed.value);
        sprintf(
            cmd,
            "%s {%s %.2f} {%s %.2f} {%s %.2f} {%s %.2f} {%s %s}",
            CMD_KEY_DRIVE_.c_str(),
            CMD_STRING_ALTITUDE_VELOCITY_.c_str(),
            altVelocity,
            CMD_STRING_LINEAR_VELOCITY_.c_str(),
            linearVelocity,
            CMD_STRING_LATERAL_VELOCITY_.c_str(),
            latVelocity,
            CMD_STRING_ROTATIONAL_VELOCITY_.c_str(),
            rotVelocity,
            CMD_STRING_NORMALIZED_.c_str(),
            ((normalized)? CMD_STRING_TRUE_.c_str() : CMD_STRING_FALSE_.c_str()));
        break;

    case USARSIM_ROBOT_CLASS_ERS:
        cmdString = generateERSGateCommand_(controlSpeed);
        sprintf(cmd, "%s", cmdString.c_str());
        break;
    }

    bufString = cmd;
    cmdLength = bufString.size();

    sprintf(cmd, "%s\r\n", cmd);
    cmdLength += 2; // '\r' + '\n'

    if (cmdLength > MAX_CMD_LENGTH_)
    {
        printfTextWindow("Error: USARSimRobot::controlRobot_(). Command too long.\n");
        return;
    }

    cmd[cmdLength] = '\0';

    write(shSocketFd_, (unsigned char *)cmd, cmdLength);
}

//-----------------------------------------------------------------------
// This function adds the sensors.
//-----------------------------------------------------------------------
void USARSimRobot::addSensors_(void)
{
    float x, y, dist;
    int i, sonarLocSize;

    partXyt_ = new RobotPartXyt();

    if (REPORT_SONARS_)
    {
        numSonars_ = ROBOT_SPECS_[robotSpecIndex_].numSonars;
        sonarMountDist_ = new float[numSonars_];
        sonarLocSize = sizeof(float)*numSonars_*USARSIM_ROBOT_SONAR_LOC_DIMENSIONS;
        sonarLocs_ = new float[numSonars_][USARSIM_ROBOT_SONAR_LOC_DIMENSIONS];

        for (i = 0; i < numSonars_; i++)
        {
            x = ROBOT_SPECS_[robotSpecIndex_].sonarCoordinates[i*2];
            y = ROBOT_SPECS_[robotSpecIndex_].sonarCoordinates[(i*2)+1];
            dist = sqrt(pow(x, 2) + pow(y, 2));

            sonarMountDist_[i] = dist;
        }

        memset(sonarLocs_, 0x0, sonarLocSize);
        sonarReadings_ = new float[numSonars_];
        memset(sonarReadings_, 0x0, sizeof(float)*numSonars_);

        partSonar_ = new RobotPartSonar(
            numSonars_,
            (float *)(ROBOT_SPECS_[robotSpecIndex_].sonarAngles));

        sensorSonar_ = new Sensor(
            SENSOR_SONAR,
            //sonarLocSize,
            //(char*)sonarLocs_, 
            0,
            NULL, 
            numSonars_,
            (float *)(ROBOT_SPECS_[robotSpecIndex_].sonarAngles),
            numSonars_,
            sonarReadings_,
            NULL);
    }

    if (REPORT_RFID_)
    {
        numRFID_ = USARSIM_ROBOT_MAX_NUM_RFID;
        RFIDData_ = new float[numRFID_ * 4]; // four elements: ID, x, y, z

        // Zero out the data
        memset(RFIDData_, 0x0, sizeof(float)*numRFID_*4);

        // Create a part for it.  This is mainly for sending configuration information to the robot executable and beyond, I believe
        partRFID_ = new RobotPartRFID(numRFID_*4);

        // Create the actual sensor data.  We ignore parts of it such as angle information, since the RFID position is incorporated
        // into the sensor data itself
        sensorRFID_ = new Sensor(
            SENSOR_RFID,
            0,
            NULL, 
            0,
	    NULL,
            numRFID_*4,    // 4 parts to the reading: ID, x, y, z
            RFIDData_,
            NULL);
    }

    if (REPORT_ENEMY_)
    {
        numEnemy_ = USARSIM_ROBOT_MAX_NUM_ENEMY;
        enemyData_ = new float[numEnemy_ * 7 + 1]; // seven elements: ID, x, y, z, dist, podE, podS + timestamp

        // Zero out the data
        memset(enemyData_, 0x0, sizeof(float)*(numEnemy_*7 + 1));

        // Create a part for it.  This is mainly for sending configuration information to the robot executable and beyond, I believe
        partEnemy_ = new RobotPartEnemy(numEnemy_*7 + 1);

        // Create the actual sensor data.  
        sensorEnemy_ = new Sensor(
            SENSOR_ENEMY,
            0,
            NULL, 
            0,
	    NULL,
            numEnemy_*7 + 1,    // 7 parts to the reading + timestamp
            enemyData_,
            NULL);
    }

    if (REPORT_CHEMICAL_)
    {
        numChemical_ = USARSIM_ROBOT_MAX_NUM_CHEMICAL;
        chemicalData_ = new float[numChemical_ * 6 + 1]; // six elements: ID, x, y, z, dist, podC + timestamp

        // Zero out the data
        memset(chemicalData_, 0x0, sizeof(float)*(numChemical_*6 + 1));

        // Create a part for it.  This is mainly for sending configuration information to the robot executable and beyond, I believe
        partChemical_ = new RobotPartChemical(numChemical_*6 + 1);

        // Create the actual sensor data.  
        sensorChemical_ = new Sensor(
            SENSOR_CHEMICAL,
            0,
            NULL, 
            0,
	    NULL,
            numChemical_*6 + 1,    // 6 parts to the reading + timestamp
            chemicalData_,
            NULL);
    }

    if (REPORT_COMM_)
    {
        numComm_ = USARSIM_ROBOT_MAX_NUM_BUG;
        commData_ = new float[numComm_ * 2 + 1]; // two elements: BugNum, pd + timestamp

        // Zero out the data
        memset(commData_, 0x0, sizeof(float)*(numComm_*2 + 1));

        // Create a part for it.  This is mainly for sending configuration information to the robot executable and beyond, I believe
        partComm_ = new RobotPartComm(numComm_*2 + 1);

        // Create the actual sensor data.  
        sensorComm_ = new Sensor(
            SENSOR_COMM,
            0,
            NULL, 
            0,
	    NULL,
            numComm_*2 + 1,    // 2 parts to the reading + timestamp
            commData_,
            NULL);
    }
}

//-----------------------------------------------------------------------
// This function sets up the user interface for the robot option.
//-----------------------------------------------------------------------
void USARSimRobot::control(void)
{
    bool done = false;
    int msgData[2];

    msgData[0] = 0; // "speed_factor" is obsolete.
    msgData[1] = 0; // "angular_speed_factor" is obsolete.

    /*
      messageDrawWindow(
      EnMessageType_USARSIM_ROBOT_CONTROL,
      EnMessageErrType_NONE,
      msgData);
    */

    do
    {
        int c = getch();

        switch(c) {

        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;

        }
    } while (!done);

    messageHide();
}

/**********************************************************************
# $Log: usarsim-robot.c,v $
# Revision 1.15  2009-08-28 19:46:40  zkira
# Rovio fixes and added header includes to fix compilation on other machines
#
# Revision 1.14  2009/04/17 17:51:49  zkira
# Added maxSpeed in robot (clamps speed to this max value), and ability to speed up usarsim data acquisition so that speeding up the USARSim simulator becomes possible
#
# Revision 1.13  2009/04/13 22:24:34  zkira
# Added assert include
#
# Revision 1.12  2009/04/13 03:57:39  zkira
# Enemy, chemical, and comm sensors (USARSim-DOE-MAST)
#
# Revision 1.10  2009/03/02 18:55:04  zkira
# Full 3D information
#
# Revision 1.9  2008/10/30 22:50:51  zkira
# Added support for up to 4 laser sensors
#
# Revision 1.8  2008/10/28 23:12:18  ts2883
# Use ground truth always (even for ground vehicles)
#
# Revision 1.7  2008/10/28 23:05:17  zkira
# Fixed LaserPart problem
#
# Revision 1.6  2008/10/27 21:39:25  zkira
# Added separate dragonfly model
#
# Revision 1.5  2008/09/18 07:36:01  zkira
# Added ground truth
# BUG fix: Changed MAIN_LOOP_SLEEP because it was too slow (only every 10th or so message
# is an odometry reading!).  It may have to be changed back if it causes network troubles though.
#
# Revision 1.4  2008/09/09 20:18:53  zkira
# Removed debug print
#
# Revision 1.3  2008/09/09 20:16:23  zkira
# Crude altitude control
#
# Revision 1.2  2008/08/26 19:40:50  endo
# Improving the USARSim module.
#
# Revision 1.5  2008/02/01 23:18:26  endo
# USARSimLaser added.
#
# Revision 1.4  2008/01/13 00:35:40  endo
# Regular backup.
#
# Revision 1.3  2008/01/11 15:51:18  endo
# Regular backup.
#
# Revision 1.2  2008/01/10 23:02:54  endo
# Regular backup.
#
# Revision 1.1  2008/01/09 22:54:03  endo
# USARSimRobot added.
#
#**********************************************************************/
