/**********************************************************************
 **                                                                  **
 **                              xwindow.c                           **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  X11 Interface for HServer                                       **
 **                                                                  **
 **  Copyright 2000 - 2006, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: xwindow.c,v 1.2 2008-07-15 22:33:55 endo Exp $ */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <pthread.h>

#include "HServerTypes.h"
#include "hserver.h"
#include "xwindow.h"
#include "message.h"
#include "Watchdog.h"

//-----------------------------------------------------------------------
// Externs
//-----------------------------------------------------------------------
Xwindow *gXwindow = NULL;

//-----------------------------------------------------------------------
// Mutexes
//-----------------------------------------------------------------------
pthread_mutex_t Xwindow::currentImageDataMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
//const string Xwindow::DEFAULT_WINDOW_NAME_ = "HServer X";
const string Xwindow::DEFAULT_WINDOW_NAME_ = "";
const string Xwindow::DEFAULT_COLOR_FILE_NAME_ = "./hserver_colors.txt";
const int Xwindow::READER_THREAD_USLEEP_ = 20000;
const int Xwindow::SKIP_STATUSBAR_UPDATE_ = 10;
const int Xwindow::WATCHDOG_CHECK_INTERVAL_SEC_ = 10;

//-----------------------------------------------------------------------
// Contructor
//-----------------------------------------------------------------------
Xwindow::Xwindow(
    Xwindow **a,
    char *displayName,
    int displayX,
    int displayY,
    int displayWidth,
    int displayHeight) :
    Module((Module**)a, HS_MODULE_NAME_XWINDOW),
    display_(NULL),
    lastImageDataTimeStamp_(0),
    screenNum_(0),
    screenDepth_(0),
    displayX_(displayX),
    displayY_(displayY),
    displayWidth_(displayWidth),
    displayHeight_(displayHeight),
    displayThreadIsUp_(false)
{
    // Assign the display name.
    displayName_ = (displayName != NULL)? displayName : DEFAULT_WINDOW_NAME_;

    // Initialize the shared data image.
    memset(&shCurrentImageData_, 0x0, sizeof(ImageData_t));
    shCurrentImageData_.rgbImage = NULL;

    // Initialize mutexes.
    pthread_mutex_init(&currentImageDataMutex_, NULL);

    // Initialize the window.
    if (!initializeWindow_())
    {
        a = NULL;
        return;
    }

    // Start the trhead.
    pthread_create(&displayThread_, NULL, &startDisplayThread_, (void*)this);

    // Start the watchdog.
    if (gWatchdogEnabled)
    {
        watchdog_ = new Watchdog(
            NAME_,
            WATCHDOG_CHECK_INTERVAL_SEC_,
            getpid(),
            displayThread_);
    }
}

//-----------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------
Xwindow::~Xwindow(void)
{
    if (displayThreadIsUp_)
    {
        pthread_cancel(displayThread_);
        pthread_join(displayThread_, NULL);
    }

    if (display_ != NULL)
    {
        XCloseDisplay(display_);    
        XFreeGC(display_, gc_);
    }

    printfTextWindow("Xwindow disconnected");
}

//-----------------------------------------------------------------------
// This function starts the display thread.
//-----------------------------------------------------------------------
void *Xwindow::startDisplayThread_(void *xwindowInstance)
{
    ((Xwindow*)xwindowInstance)->displayLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to update display.
//-----------------------------------------------------------------------
void Xwindow::displayLoop_(void)
{
    int count = 0;

    displayThreadIsUp_ = true;

    while(true)
    {
        // Check for the pthread termination.
        pthread_testcancel();

        // Update the display.
        updateDisplay_();

        // Check for the pthread termination again.
        pthread_testcancel();

        // Check for the key press etc.
        checkXEvent_();

        // Compute the frequency.
        compFrequency_();

        if (count > SKIP_STATUSBAR_UPDATE_)
        {
            updateStatusBar_();
            count = 0;
        }
        count++;

        if (watchdog_ != NULL)
        {
            watchdog_->notifyUpdate();
        }

        usleep(READER_THREAD_USLEEP_);
    }
}

//-----------------------------------------------------------------------
// This function updates the display.
//-----------------------------------------------------------------------
void Xwindow::updateDisplay_(void)
{
    XImage xImage;
    unsigned char *rgbImage = NULL;
    int x, y, width, height, imageSize;

    pthread_mutex_lock(&currentImageDataMutex_);

    if (shCurrentImageData_.timeStamp != lastImageDataTimeStamp_)
    {
        // Calcuate the image size.
        x = shCurrentImageData_.x;
        y = shCurrentImageData_.y;
        width = shCurrentImageData_.width;
        height = shCurrentImageData_.height;
        imageSize = width * height * 3;

        // Copy the image.
        rgbImage = new unsigned char[imageSize];
        memcpy(rgbImage, shCurrentImageData_.rgbImage, imageSize);

        // Save the time stamp.
        lastImageDataTimeStamp_ = shCurrentImageData_.timeStamp;
    }

    pthread_mutex_unlock(&currentImageDataMutex_);

    if (rgbImage != NULL)
    {
        if (rgb2xImage_(width, height, rgbImage, &xImage))
        {
            XPutImage(
                display_,
                window_,
                gc_,
                &xImage,
                0,
                0,
                x,
                y,
                width,
                height);
        }

        delete [] rgbImage;
        rgbImage = NULL;
    }
}

//-----------------------------------------------------------------------
// This function checks the current key press.
//-----------------------------------------------------------------------
void Xwindow::checkXEvent_(void)
{
    XEvent report;

    if (XPending(display_))
    {
        XNextEvent(display_, &report);

        switch(report.type) {

        case KeyPress:
            handleKeyPress_((XKeyEvent*)&report);
            break;

        default:
            if (report_level >= HS_REP_LEV_DEBUG)
            {
                printfTextWindow("Xwindow::checkXEvent_(): Event = %d", report.type);
            }
            break;
        }
    }
}

//-----------------------------------------------------------------------
// This handles the current key press.
//-----------------------------------------------------------------------
void Xwindow::handleKeyPress_(XKeyEvent* report)
{
    char buf[100];
    KeySym keysym;
    XComposeStatus compose;
    int count;

    count = XLookupString(report, buf, 100, &keysym, &compose);

    switch(keysym){

    default:
        printfTextWindow("KeyPress on Xwindow.");
       // Implement handler.
        break;
    }
}

//-----------------------------------------------------------------------
// This function initializes the window.
//-----------------------------------------------------------------------
bool Xwindow::initializeWindow_(void)
{
    char buf[100];

    if (!loadColors_())
    {
        return false;
    }

    display_ = XOpenDisplay(NULL);

    if (display_ == NULL)
    {
        printfTextWindow("Error: Xwindow::initializeWindow_() XOpenDisplay failed.");
        return false;
    }

    screenNum_ = DefaultScreen(display_);
    createWindow_();
    mapWindow_();

    if (report_level>=HS_REP_LEV_DEBUG)
    {
        sprintf(
            buf,
            "Screen Num %d  width %d  height %d  x %d  y %d  depth %d",
            screenNum_,
            displayWidth_,
            displayHeight_,
            displayX_,
            displayY_,
            screenDepth_);

        printTextWindow(buf);
    }

    return true;
}

//-----------------------------------------------------------------------
// This function loads the colors.
//-----------------------------------------------------------------------
bool Xwindow::loadColors_(void)
{
    FILE *fp;
    int i;
    unsigned int r,g,b;
    const bool SET_COLORS_MANUALLY = true;

    if (!SET_COLORS_MANUALLY)
    {
        // color_file should contain:
        //
        // 5
        // 0 0 0 
        // 255 255 255
        // 255 0 0
        // 0 255 0
        // 0 0 255
        //
        fp = fopen(DEFAULT_COLOR_FILE_NAME_.c_str(),"r");
        if (fp == NULL)
        {
            printfTextWindow(
                "Error: Xwindow::loadColors_(). Can't open \"%s\".",
                DEFAULT_COLOR_FILE_NAME_.c_str());
            return false;
        }
        fscanf(fp,"%d",&i);
        numColors_ = i;

        for(i = 0; i < numColors_; i++)
        {
            fscanf(fp,"%d%d%d",&r,&g,&b);
            colors_[i] = b|(g<<8)|(r<<16);
        }

        fclose(fp);
    }
    else
    {
        r = 0;
        g = 0;
        b = 0;
        colors_[0] = b|(g<<8)|(r<<16);

        r = 255;
        g = 255;
        b = 255;
        colors_[1] = b|(g<<8)|(r<<16);

        r = 255;
        g = 0;
        b = 0;
        colors_[2] = b|(g<<8)|(r<<16);

        r = 0;
        g = 255;
        b = 0;
        colors_[3] = b|(g<<8)|(r<<16);

        r = 0;
        g = 0;
        b = 255;
        colors_[4] = b|(g<<8)|(r<<16);
    }

    return true;
}

//-----------------------------------------------------------------------
// This function converts R, G, & B values to a unsigned int color.
//-----------------------------------------------------------------------
unsigned int Xwindow::rgb2Color_(int r, int g, int b)
{
    return (b | (g << 8) | (r << 16));
}

//-----------------------------------------------------------------------
// This function displays the RGB image.
//-----------------------------------------------------------------------
void Xwindow::displayRGBImage(
    int x,
    int y,
    int w,
    int h,
    unsigned char *rgbImage)
{
    if (!displayThreadIsUp_)
    {
        return;
    }

    // Save, so that display handler will display it.
    saveImage_(x, y, w, h, rgbImage, IMAGE_COLOR_ORDER_RGB);
}

//-----------------------------------------------------------------------
// This function displays the BGR image.
//-----------------------------------------------------------------------
void Xwindow::displayBGRImage(
    int x,
    int y,
    int w,
    int h,
    unsigned char *bgrImage)
{
    if (!displayThreadIsUp_)
    {
        return;
    }

    // Save, so that display handler will display it.
    saveImage_(x, y, w, h, bgrImage, IMAGE_COLOR_ORDER_BGR);
}

//-----------------------------------------------------------------------
// This function displays the RGB/BGR image.
//-----------------------------------------------------------------------
void Xwindow::saveImage_(
    int x,
    int y,
    int w,
    int h,
    const unsigned char* image,
    int imageColorOrder)
{
    double timeStamp;
    int imageSize;

    if (!displayThreadIsUp_)
    {
        return;
    }

    // Get the time stamp.
    timeStamp = getCurrentEpochTime();

    // Compute the image size.
    imageSize = w*h*3;

    pthread_mutex_lock(&currentImageDataMutex_);

    // Check the memory space for the image.
    if (shCurrentImageData_.rgbImage != NULL)
    {
        if (shCurrentImageData_.imageSize != imageSize)
        {
            delete [] shCurrentImageData_.rgbImage;
            shCurrentImageData_.rgbImage = new unsigned char[imageSize];
        }
    }
    else
    {
        shCurrentImageData_.rgbImage = new unsigned char[imageSize];
    }

    // Copy the image.
    switch (imageColorOrder) {

    case IMAGE_COLOR_ORDER_RGB:
        memcpy(shCurrentImageData_.rgbImage, image, imageSize);
        break;

    case IMAGE_COLOR_ORDER_BGR:
        bgr2rgb_(w, h, image, &(shCurrentImageData_.rgbImage));
        break;

    default:
        memset(shCurrentImageData_.rgbImage, 0x0, imageSize);
        break;
    }

    // Copy other data.
    shCurrentImageData_.timeStamp = timeStamp;
    shCurrentImageData_.x = x;
    shCurrentImageData_.y = y;
    shCurrentImageData_.width = w;
    shCurrentImageData_.height = h;
    shCurrentImageData_.imageSize = imageSize;

    pthread_mutex_unlock(&currentImageDataMutex_);
}

//-----------------------------------------------------------------------
// This function converts RGB data into an XImage.
//-----------------------------------------------------------------------
bool Xwindow::rgb2xImage_(
    int w,
    int h,
    const unsigned char* rgbImage,
    XImage *xImage)
{
    if (xImage == NULL)
    {
        return false;
    }

    xImage->width = w;
    xImage->height = h;
    xImage->xoffset = 0;
    xImage->format = ZPixmap;
    xImage->data = (char*)rgbImage;
  

    xImage->byte_order = LSBFirst;
    xImage->bitmap_unit = 8;
    xImage->bitmap_bit_order = LSBFirst;

    xImage->bitmap_pad = 8;

    xImage->depth = screenDepth_;
    xImage->bytes_per_line = w * (24/8);
    xImage->bits_per_pixel = 24;

    /*
    xImage->red_mask   = 0xFF << 16; //magic markers
    xImage->green_mask = 0xFF << 8;
    xImage->blue_mask  = 0xFF << 0;  
    */

    XInitImage(xImage);

    return true;
}

//-----------------------------------------------------------------------
// This function convert a BGR base image into a RGB image. 
//-----------------------------------------------------------------------
bool Xwindow::bgr2rgb_(
    int w,
    int h,
    const unsigned char *bgrImage,
    unsigned char **rgbImage)
{
    int x, y, idx;
    unsigned char *rgbImagePtr = NULL, *rgbImagePtrOrig = NULL;
    unsigned char *bgrImagePtr = NULL;

    rgbImagePtrOrig = *rgbImage;

    for (y = 0; y < h; y++)
    {
        for (x = 0; x < w; x++)
        {
            idx = ((y*w)+x)*3;

            rgbImagePtr = (unsigned char *)(rgbImagePtrOrig + idx + 2);
            bgrImagePtr = (unsigned char *)(bgrImage + idx);

            if (rgbImagePtr != NULL)
            {
                // Copy blue.
                *rgbImagePtr = *bgrImagePtr;

                // Copy green.
                rgbImagePtr--;
                bgrImagePtr++;
                *rgbImagePtr = *bgrImagePtr;

                // Copy red.
                rgbImagePtr--;
                bgrImagePtr++;
                *rgbImagePtr = *bgrImagePtr;
            }
            else
            {
                printfTextWindow("Xwindow::bgr2rgb_(): Insufficient memory allocation.");
                return false;
            }
        }
    }

    *rgbImage = rgbImagePtrOrig;
    
    return true;
}

//-----------------------------------------------------------------------
// This function maps the window.
//-----------------------------------------------------------------------
void Xwindow::mapWindow_(void)
{
    XMapWindow(display_, window_);
}


//-----------------------------------------------------------------------
// This function creates the window.
//-----------------------------------------------------------------------
void Xwindow::createWindow_(void)
{
    XGCValues values;
    XWindowAttributes wa;
    unsigned long valuemask;

    // make our window
    window_ = XCreateSimpleWindow(
        display_, 
        RootWindow(display_, screenNum_), 
        displayX_,
        displayY_,
        displayWidth_,
        displayHeight_,
        0, 
        BlackPixel(display_, screenNum_), 
        BlackPixel(display_, screenNum_));

    sizeHints_ = XAllocSizeHints();
    sizeHints_->min_width = 300;
    sizeHints_->min_height = 200;

    XmbSetWMProperties(
        display_,
        window_,
        displayName_.c_str(),
        displayName_.c_str(),
        NULL,
        0,
        sizeHints_,
        NULL,
        NULL);

    XSelectInput(
        display_,
        window_,
        ExposureMask | KeyPressMask | ButtonPressMask | StructureNotifyMask);

    // make the pesky, GC
    valuemask = 0; // ignore XGCvalues and use defaults 
    gc_ = XCreateGC(display_, window_, valuemask, &values);

    // get the screen depth
    XGetWindowAttributes(display_, window_, &wa);
    screenDepth_ = wa.depth;
}  

void Xwindow::drawRect(int x, int y, int w, int h, int c)
{
    XSetForeground(display_, gc_, colors_[c]);
    XDrawRectangle(display_, window_, gc_, x, y, w, h);  
}

void Xwindow::drawPoint(int x,int y,int c)
{
    XSetForeground(display_,gc_,colors_[c]);
    XDrawPoint(display_, window_, gc_, x, y);
}

void Xwindow::control()
{
    bool bDone = 0;
    
    messageDrawWindow(EnMessageType_X_CONTROL, EnMessageErrType_NONE);
    do
    {
        int c = getch();
        switch(c){
        case 'd':
            messageHide();
            delete this;
            return;
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            bDone = true;
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey(c);
            break;
        }
    } while (!bDone);
    messageHide();
}

void Xwindow::updateStatusBar_(void)
{
    char buf[100];
    double frequency = 0;
    const bool DISPLAY_FREQUENCY = true;

    statusStr[0] = '\0';

    strcat(statusStr, "Xwindow [");

    sprintf(
        buf,
        "%s]: %c",
        displayName_.c_str(),
        statusbarSpinner_->getStatus());

    if (DISPLAY_FREQUENCY)
    {
        getFrequency_(frequency);
        sprintf(buf, "%s : %.2f Hz", buf, frequency);
    }

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}


/**********************************************************************
# $Log: xwindow.c,v $
# Revision 1.2  2008-07-15 22:33:55  endo
# OpenCV, Gazebo and USARSim added.
#
# Revision 1.9  2008/01/07 21:44:42  endo
# xwindow.c
#
# Revision 1.8  2006/11/30 17:51:55  endo
# Regular update.
#
# Revision 1.7  2006/11/29 23:13:28  endo
# OpenCV improved.
#
# Revision 1.6  2006/11/29 00:01:45  endo
# OpenCV BlobTrack improved.
#
# Revision 1.5  2006/11/21 23:13:38  endo
# FrameGrabber separated from Video class.
#
# Revision 1.4  2006/11/20 22:55:52  endo
# OpenCV added.
#
# Revision 1.3  2006/11/20 16:43:37  endo
# Xwindow improved.
#
# Revision 1.2  2006/11/17 23:35:53  endo
# GazeboVideo implemented.
#
# Revision 1.1.1.1  2006/07/21 17:42:11  endo
# HServer for Gazebo (based on MissionLab-7.0.20060712.tar.gz)
#
# Revision 1.1.1.1  2006/07/12 13:37:56  endo
# MissionLab 7.0
#
# Revision 1.3  2006/06/19 18:51:19  endo
# color_file no longer necessary.
#
# Revision 1.2  2006/06/05 21:35:52  endo
# PoseCalculator from MARA 2020 migrated
#
# Revision 1.2  2004/09/10 19:41:08  endo
# New PoseCalc integrated.
#
# Revision 1.1.1.1  2004/02/27 22:30:51  endo
# Initial MARS 2020 revision
#
# Revision 1.4  2000/12/12 23:23:05  blee
# Changed #defines for the user interface to enumerated types.
#
# Revision 1.3  2000/09/19 03:51:29  endo
# RCS log added.
#
#
#**********************************************************************/
